/*
 * Copyright (C) 2019 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
/*

  This loader uses the decoder from LibreSprite. In order to do so it
  implements the API used by ase_format.cc and provides the files it
  inludes:

    app/context.h
    app/document.h
    app/file/file.h
    app/file/file_format.h
    app/file/format_options.h
    base/cfile.h
    base/exception.h
    base/file_handle.h
    base/path.h
    doc/doc.h
    ui/alert.h
    zlib.h

*/
extern "C"{
#define new new_
#include <abydos-plugin.h>
#include <debug.h>
#undef new
}
#include <array>
#include <functional>
#include "app/file/file.h"
#include "app/file/file_format.h"
#include "base/exception.h"
#include "base/file_handle.h"
#include "doc/cel.h"
#include "doc/image_traits.h"
#include "doc/image_impl.h"
#include "doc/layer.h"
#include "doc/mask.h"

#ifndef MIN
# define MIN(a,b) ((a)<(b)?(a):(b))
#endif
#ifndef MAX
# define MAX(a,b) ((a)>(b)?(a):(b))
#endif

namespace app{
    void FileOp::setError(char const* a,...){
        DEBUG_p("error: %s",a);
    }

    void FileOp::setProgress(double a){
        //DEBUG_p("%f",a);
    }

    FileFormat* CreateAseFormat();
}

namespace base{
    Exception::Exception(const char* a,...)throw(){
        DEBUG_p("!!NOT implemented");
    }
    Exception::~Exception(){
        DEBUG_p("!!NOT implemented");
    }
    const char* Exception::what()const throw(){
        DEBUG_p("!!NOT implemented");
        return "";
    }
    std::string get_file_name(const std::string& a){
        DEBUG_p("!!NOT implemented");
        return a;
    }
}

namespace doc{
    /* --- Cel --- */
    Cel* Cel::createCopy(Cel const* a){
        DEBUG_p("!!NOT implemented (Cel)");
        return NULL;
    }
    Cel* Cel::createLink(Cel const* a){
        DEBUG_p("!!NOT implemented (Cel)");
        return NULL;
    }
    Cel* Cel::link()const{
        DEBUG_p("!!NOT implemented (Cel)");
        return NULL;
    }

    /* --- FrameTags --- */
    void FrameTags::add(FrameTag* a){
        DEBUG_p("!!NOT implemented");
    }

    /* --- Image --- */
    Image* Image::create(PixelFormat f,int w,int h,base::SharedPtr<ImageBuffer> const &b){
        switch(f){
        case IMAGE_RGB:return new ImageImpl<RgbTraits>(w,h,b);
        case IMAGE_GRAYSCALE:return new ImageImpl<GrayscaleTraits>(w,h,b);
        case IMAGE_INDEXED:return new ImageImpl<IndexedTraits>(w,h,b);
        case IMAGE_BITMAP:return new ImageImpl<BitmapTraits>(w,h,b);
        }
        return NULL;
    }

    /* --- Layer --- */
    void LayerImage::addCel(Cel* cel){
        auto iter = _cels.begin();
        auto end = _cels.end();

        iter = std::lower_bound(iter,end,nullptr,[cel](Cel *c,Cel*)->bool{
            return c->frame() <= cel->frame();
        });
        _cels.insert(iter, cel);
    }

    Cel* LayerImage::cel(int frame)const{
        for(auto cel: _cels){
            if(cel->frame() == frame)
                return cel;
        }
        return NULL;
    }

    void LayerImage::displaceFrames(int a,int b){
        DEBUG_p("!!NOT implemented (LayerImage)");
    }
    void LayerImage::getCels(CelList& a)const{
        DEBUG_p("!!NOT implemented (LayerImage)");
    }

     LayerFolder::~LayerFolder(){
        DEBUG_p("!!NOT implemented (LayerFolder)");
     }

    void LayerFolder::addLayer(Layer* l){
        DEBUG_p("begin (LayerFolder)");
        _layers.push_back(l);
        l->setParent(this);
    }

    void LayerFolder::displaceFrames(int a,int b){
        DEBUG_p("!NOT implemented (LayerFolder)");
    }

    void LayerFolder::getCels(CelList& a)const{
        DEBUG_p("!!NOT implemented (LayerFolder)");
    }

    void LayerFolder::removeLayer(Layer* a){
        DEBUG_p("!!NOT implemented (LayerFolder)");
    }

    /* --- Mask --- */
    void Mask::replace(const gfx::Rect& a){
        DEBUG_p("!!NOT implemented");
    }

    /* --- Palette --- */
    Palette::Palette(int f, int c):
        _frame(f)
    {
        resize(c);
    }

    Palette::Palette(const Palette &p){
        _frame = p._frame;
        _colors = p._colors;
    }

    void Palette::resize(int c){
        _colors.resize(c, rgba(0,0,0,255));
    }
    void Palette::setEntry(int i,color_t c){
        _colors[i] = c;
    }

    void Palette::copyColorsTo(Palette* dst) const{
        dst->_colors = _colors;
    }

    int Palette::countDiff(Palette const* p,int* from,int* to) const{
        int diff = 0;
        int min = MIN(_colors.size(), p->_colors.size());
        int max = MAX(_colors.size(), p->_colors.size());
        int i;

        if(from)
            *from = -1;
        if(to)
            *to = -1;

        for(i=0; i<min; ++i){
            if(_colors[i] != p->_colors[i]){
                if(from && *from<0)
                    *from = i;
                if(to)
                    *to = i;
                ++diff;
            }
        }
        if(min != max){
            diff += max - min;
            if(from && *from<0)
                *from = min;
            if(to)
                *to = max-1;
        }
        return diff;
    }

    /* --- Sprite --- */
    Sprite::Sprite(PixelFormat f,int w,int h,int c):
        _format(f),
        _width(w),
        _height(h),
        _colors(c),
        _frame_tags(this),
        _default_duration(100)
    {
        DEBUG_p("begin");
        _folder = new LayerFolder(this);

        DEBUG_p("colors: %d",c);
        Palette p(0,c);
        switch(f){
        case IMAGE_GRAYSCALE:
        case IMAGE_BITMAP:
            for(int i=0; i<c; ++i){
                int v = 255 * i / (c-1);
                p.setEntry(i, rgba(v,v,v,255));
            }
            break;
        default:
            break;
        }
        setPalette(&p,true);
    }

    Layer* Sprite::indexToLayer(int index)const{
        return folder()->getLayersList()[index];
    }

    int Sprite::layerToIndex(Layer const* l)const{
        int i = 0;
        for(auto iter=folder()->getLayerBegin();iter!=folder()->getLayerEnd();++iter,++i)
            if(*iter == l)
                return i;
        return -1;

    }

    Palette* Sprite::palette(int f)const{
        Palette *best = NULL;
        PalettesConstIter end = _palettes.end();
        PalettesConstIter iter = _palettes.begin();
        for(; iter!=end; ++iter){
            Palette *p = *iter;
            if(p->frame() == f)
                return p;
            if(f < p->frame())
                return NULL;
            best = p;
        }
        return best;
    }

    void Sprite::setTotalFrames(int c){
        _frame_count = c;
        _duration.resize(c,_default_duration);
    }

    void Sprite::setFrameDuration(int i,int msec){
        _duration[i] = msec;
    }

    void Sprite::setDurationForAllFrames(int msec){
        _default_duration = msec;
        std::fill(_duration.begin(),_duration.end(),msec);
    }


    void Sprite::setPalette(Palette const* src,bool truncate){
        if(truncate){
            PalettesIter end = _palettes.end();
            PalettesIter iter = _palettes.begin();
            for(; iter!=end; ++iter){
                Palette *dst = *iter;
                if(dst->frame() == src->frame()){
                    src->copyColorsTo(dst);
                    return;
                }
                if(dst->frame() > src->frame())
                    break;
            }
            _palettes.insert(iter, new Palette(*src));
        }else{
            DEBUG_p("!!NOT implemented (truncate == false) (Sprite)");
        }
    }

    void copy_bitmaps(Image *dst,const Image *src, gfx::Clip area){
        DEBUG_p("!!NOT implemented");
    }
    void put_pixel(Image* a,int b,int c,unsigned int d){
        DEBUG_p("!!NOT implemented");
    }
    color_t rgba_blender_normal(unsigned int, unsigned int, int){
        DEBUG_p("!!NOT implemented");
        return 0;
    }
}

namespace gfx{
    bool Clip::clip(int dst_w, int dst_h,int src_w,int src_h){
        DEBUG_p("!!NOT implemented");
        return false;
    }
}


struct layer_t{
    int x;
    int y;
    cairo_surface_t *surface;
};


struct frame_t{
    layer_t *layer;
    double duration;
};

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    app::FileFormat* format;
    cairo_surface_t *cel;
    int cel_count;
    frame_t *frame;
    int frame_count;
};


static cairo_surface_t *
_surface_from_cel(doc::Cel *cel)
{
    DEBUG_p("processing cel of frame %d",cel->frame());
    doc::Image *img = cel->image();
    DEBUG_p("size: %dx%d",img->width(),img->height());
    //auto bounds = img->bounds();
    //DEBUG_p("bounds: %dx%d",bounds.width(),bounds.height());

    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,img->width(),img->height());
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_skip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t)-img->width();
    for(int y=0;y<img->height();++y){
        for(int x=0;x<img->width();++x){
            doc::color_t c = img->getPixel(x,y);
            int a = doc::rgba_geta(c);
            int r = a*doc::rgba_getr(c)/255;
            int g = a*doc::rgba_getg(c)/255;
            int b = a*doc::rgba_getb(c)/255;
            *dst++ = r<<16|g<<8|b<<0|a<<24;
        }
        dst += dst_skip;
    }

    cairo_surface_mark_dirty(surface);
    return surface;
}


static void
_foreach_layer(const doc::LayerFolder *lf,std::function<void(const doc::LayerImage*)> func)
{
    for(auto iter=lf->getLayerBegin(); iter!=lf->getLayerEnd(); ++iter){
        doc::Layer *l = *iter;
        if(l->isImage()){
            func((const doc::LayerImage*)l);
        }else if(l->isFolder()){
            _foreach_layer((const doc::LayerFolder*)l,func);
        }else{
            DEBUG_p("unknown layer type");
        }
    }
}


static int
_aseprite_create_from_stream(abydos_plugin_handle_t *h, FILE *f)
{
    app::FileOp op(f);
    h->format->load(&op);
    h->format->postLoad(&op);

    const doc::Sprite *s = op.document()->sprite();
    h->info->width = s->width();
    h->info->height = s->height();
    h->info->frame_count = s->totalFrames();
    DEBUG_p("frames: %d",h->info->frame_count);

    std::vector<const doc::LayerImage*> layers;
    _foreach_layer(s->folder(),[&](const doc::LayerImage *li){
        DEBUG_p("image layer");
        ++h->info->layer_count;
        layers.push_back(li);
    });
    h->info->layer_count = layers.size();
    DEBUG_p("layers: %d",h->info->layer_count);

    h->frame = new frame_t[h->info->frame_count];
    for(int i=0;i<h->info->frame_count;++i){
        auto frame = &h->frame[i];
        frame->layer = new layer_t[h->info->layer_count];
        for(int j=0;j<h->info->layer_count;++j){
            auto layer = layers[j];
            auto cel = layer->cel(i);
            frame->layer[j].x = cel->x();
            frame->layer[j].y = cel->y();
            frame->layer[j].surface = _surface_from_cel(cel);
        }
        frame->duration = (double)s->frameDuration(i)/1000;
    }

    return 0;
}


static int
_aseprite_create_from_data(abydos_plugin_handle_t *h,const char *bytes, size_t len)
{
    FILE *f = fmemopen((void*)bytes,len,"r");
    if(!f)
        return -1;
    int res = _aseprite_create_from_stream(h,f);
    fclose(f);
    return res;
}


static int
_aseprite_create_from_file(abydos_plugin_handle_t *h,const char *filename)
{
    FILE *f = fopen(filename,"r");
    if(!f)
        return -1;
    int res = _aseprite_create_from_stream(h,f);
    fclose(f);
    return res;
}


static void
_aseprite_render(abydos_plugin_handle_t *h,
                 cairo_t *cr,
                 int page,
                 cairo_bool_t *layer_visibility,
                 int variant,
                 double time,
                 int frame)
{
    cairo_save(cr);
    for(int i=0;i<h->info->layer_count;++i){
        if(!layer_visibility[i])
            continue;
        DEBUG_p("drawing layer %d",i);
        auto layer = &h->frame[frame].layer[i];
        cairo_set_source_surface(cr,layer->surface,layer->x,layer->y);
        cairo_paint(cr);
        cairo_set_operator(cr,CAIRO_OPERATOR_OVER);
    }
    cairo_restore(cr);
}


extern "C" const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = [](){
            static const char *mime_types[] = {
                "image/x-aseprite",
                NULL
            };
            return mime_types;
        },
        .new_ = [](const char *mime_type, abydos_plugin_info_t *info){
            auto *h = new abydos_plugin_handle_t;
            h->info = info;
            h->info->threadsafe = 1;
            h->format = app::CreateAseFormat();
            return h;
        },
        .free = [](abydos_plugin_handle_t *h){
            delete h;
        },
        .create_from_data = _aseprite_create_from_data,
        .create_from_file = _aseprite_create_from_file,
        .render = _aseprite_render,
        .get_duration = [](abydos_plugin_handle_t *h, int page, double *duration){
            for(int i=0;i<h->info->frame_count;++i)
                duration[i] = h->frame[i].duration;
        }
    };
    return &p;
}
