/*
 * Copyright (C) 2019 - 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <endian.h>
#include <nil-cairo.h>
#include <nil-color.h>
#include <stdlib.h>                     /* malloc(), free() */

#include "avif/avif.h"

#if __BYTE_ORDER == __BIG_ENDIAN
# define AVIF_CAIRO_FORMAT_ARGB32 AVIF_RGB_FORMAT_ARGB
/* FIXME!! the format should ideally be xRGB */
# define AVIF_CAIRO_FORMAT_RGB24  AVIF_RGB_FORMAT_RGB
#else
# define AVIF_CAIRO_FORMAT_ARGB32 AVIF_RGB_FORMAT_BGRA
/* FIXME!! the format should ideally be BGRx */
# define AVIF_CAIRO_FORMAT_RGB24  AVIF_RGB_FORMAT_BGR
#endif


typedef struct{
    cairo_surface_t *surface;
    double duration;
}frame_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    frame_t *frame;
};


static const char **
_avif_get_mime_types()
{
    static const char *mime_types[] = {
        "image/avif",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_avif_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    return h;
}


static void
_avif_free(abydos_plugin_handle_t *h)
{
    free(h);
}


static nile_t *
_ni_from_rgb_image(nile_t *ni, avifRGBImage *image)
{
#if 1
    /* this is the only format used */
    ni->bpc = 4;
    ni->format = NIL_CAIRO_FORMAT_ARGB32;
#else
    switch(image->format){
    case AVIF_CAIRO_FORMAT_RGB24:
        ni->bpc = 3;
        ni->format = NIL_CAIRO_FORMAT_RGB24;
        break;
    case AVIF_CAIRO_FORMAT_ARGB32:
        ni->bpc = 4;
        ni->format = NIL_CAIRO_FORMAT_ARGB32;
        break;
    default:
        return NULL;
    }
#endif
    ni->type = NILE_TYPE_PLANAR;
    ni->width = image->width;
    ni->height = image->height;
    ni->pixels = image->pixels;
    return ni;
}


static cairo_surface_t *
_surface_from_image(avifImage *image)
{
    cairo_surface_t *surface;
    avifRGBImage rgb;
    avifRGBImageSetDefaults(&rgb, image);
#if 1
    /* FIXME!! is this the correct way to check for an alpha channel? */
    int has_alpha = image->alphaRowBytes > 0;
    surface = cairo_image_surface_create(has_alpha ? CAIRO_FORMAT_ARGB32 : CAIRO_FORMAT_RGB24, image->width, image->height);
    /* Always convert to the 32-bit cairo format since cairo uses the extra
     * byte for padding then there is no alpha channel.
     */
    rgb.format = AVIF_CAIRO_FORMAT_ARGB32;
    rgb.pixels = cairo_image_surface_get_data(surface);
    rgb.rowBytes = cairo_image_surface_get_stride(surface);
    if(avifImageYUVToRGB(image, &rgb) != AVIF_RESULT_OK){
        cairo_surface_destroy(surface);
        return NULL;
    }
    if(has_alpha){
        nile_t ni;
        nile_init(&ni);
        _ni_from_rgb_image(&ni, &rgb);
        nil_color_premultiply(&ni);
        nile_done(&ni);
    }
#else
    avifRGBImageAllocatePixels(&rgb);
    avifImageYUVToRGB(image, &rgb);
    surface = _surface_from_rgb_image(&rgb);
    avifRGBImageFreePixels(&rgb);
#endif
    return surface;
}


static int
_avif_create_from_data(abydos_plugin_handle_t *h,const char *bytes, size_t len)
{
    avifDecoder *dec;
    int i;

    dec = avifDecoderCreate();
    if(avifDecoderSetIOMemory(dec, (uint8_t*)bytes, len) != AVIF_RESULT_OK)
        goto error;
    if(avifDecoderParse(dec) != AVIF_RESULT_OK)
        goto error;

    h->info->frame_count = dec->imageCount;
    h->frame = (frame_t*)calloc(h->info->frame_count,sizeof(frame_t));
    for(i=0;i<h->info->frame_count;++i){
        if(avifDecoderNextImage(dec) != AVIF_RESULT_OK)
            goto error;
        if(dec->image->width > h->info->width)
            h->info->width = dec->image->width;
        if(dec->image->height > h->info->height)
            h->info->height = dec->image->height;

        h->frame[i].surface = _surface_from_image(dec->image);
    }
    avifDecoderDestroy(dec);
    return 0;
error:
    avifDecoderDestroy(dec);
    return -1;
}


static cairo_surface_t *
_avif_get_image_surface(abydos_plugin_handle_t *h,
                         int page,
                         cairo_bool_t *layer_visibility,
                         int variant,
                         double time,
                         int frame)
{
    return cairo_surface_reference(h->frame[frame].surface);
}


static void
_avif_get_duration(abydos_plugin_handle_t *h,int page,double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->frame[i].duration;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _avif_get_mime_types,
        .new = _avif_new,
        .free = _avif_free,
        .create_from_data = _avif_create_from_data,
        .get_image_surface = _avif_get_image_surface,
        .get_duration = _avif_get_duration
    };
    return &p;
}
