/*
 * Copyright (C) 2019 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <nil-cairo.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */

#include "cons_buf.h"
#include "nil-c64.h"
#include "petscii_dec.h"

#define SCREEN_WIDTH 40
#define SCREEN_HEIGHT 25
#define CHAR_WIDTH 8
#define CHAR_HEIGHT 8

static
#include "c64font.c"

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
_petscii_get_mime_types(){
    static const char *mime_types[] = {
        "text/x-petscii-sequence",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_petscii_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    info->threadsafe = 1;
    h->surface = NULL;
    return h;
}


static void
_petscii_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


static void
_draw_character(uint32_t *dst,int stride,int ch, uint32_t bg, uint32_t fg)
{
    uint8_t *fontchar = &c64font[ch*CHAR_HEIGHT];
    int x;
    int y;
    //DEBUG_p("drawing %d, bg: %x, fg: %x",ch,bg,fg);
    for(y=0;y<CHAR_HEIGHT;++y){
        int pixels = *fontchar++;
        for(x=0;x<CHAR_WIDTH;++x){
            pixels <<= 1;
            dst[x] = pixels & (1<<CHAR_WIDTH) ? fg : bg;
        }
        dst += stride;
    }
}


static cairo_surface_t *
_surface_from_console(Console *cons)
{
    const ConsoleCell *src = cons_buf_buffer(cons);
    int width = cons_buf_width(cons);
    int height = cons_buf_height(cons);
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width*CHAR_WIDTH,height*CHAR_HEIGHT);
    nile_t ni;
    uint32_t palette[16];
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    int x;
    int y;
    nile_init(&ni);
    nil_cairo_from_surface(&ni, surface);
    nil_c64_copy_palette_colors((uint8_t*)palette, ni.bpc, ni.format, NIL_C64_COLODORE, 16);
    for(y=0;y<height;++y){
        for(x=0;x<width;++x){
            union{
                PetsciiState state;
                int attr:CONSOLE_BITS_ATTR;
            }s;
            ConsoleCell cell = *src++;
            uint32_t bg;
            uint32_t fg;
            DEBUG_p("cell attr: %x",cell.attr);
            s.attr = cell.attr;
            if(s.state.reverse){
                bg = palette[s.state.color];
                fg = 0;
            }else{
                bg = 0;
                fg = palette[s.state.color];
            }
            _draw_character(dst+x*CHAR_WIDTH,stride,cell.ch,bg,fg);
        }
        dst += stride*CHAR_HEIGHT;
    }
    nile_done(&ni);
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_petscii_create_from_data(abydos_plugin_handle_t *h, const char *data, size_t len)
{
    Console *cons = cons_buf_new();
    PetsciiDecoder *dec;
    cons_buf_set_size(cons, SCREEN_WIDTH, SCREEN_HEIGHT, 0);
    dec = petscii_decoder_new(&cons_buf_if,cons);
    h->info->width = SCREEN_WIDTH * CHAR_WIDTH;
    h->info->height = SCREEN_HEIGHT * CHAR_HEIGHT;
    h->info->pixel_ratio = 5.0/6;
    while(len-->0)
        petscii_decoder_write(dec,*(unsigned char*)data++);
    h->surface = _surface_from_console(cons);
    petscii_decoder_free(dec);
    cons_buf_free(cons);
    return 0;
}


static cairo_surface_t *
_petscii_get_image_surface(abydos_plugin_handle_t *h,int page,cairo_bool_t *layer_visibility,int variant,double time,int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _petscii_get_mime_types,
        .new = _petscii_new,
        .free = _petscii_free,
        .create_from_data = _petscii_create_from_data,
        .get_image_surface = _petscii_get_image_surface
    };
    return &p;
}
