/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <debug.h>
#include <nil-cairo.h>
#include <stdlib.h>
#include <string.h>

#include "cgmdraw.h"
#include "cgmdraw-cairo.h"
#include "cgmparse.h"


typedef struct{
    cgm_draw_page_func_t pagefunc;
    void *userdata;
    cairo_t *cr;
    double cursor_x;
    double cursor_y;
}context_t;


static void
_cairo_set_color(cairo_t *cr, color_t color){
    int r = (color >> 16) & 0xff;
    int g = (color >> 8 ) & 0xff;
    int b = (color >> 0 ) & 0xff;
    cairo_set_source_rgb(cr, (double)r/255, (double)g/255, (double)b/255);
}


static void
_cairo_set_line_type(cairo_t *cr, line_type_t t, double w)
{
    double dash[8];
    switch(t){
    case LT_SOLID:
        cairo_set_dash(cr, NULL, 0, 0);
        break;
    case LT_DASH:
        dash[0] = 3*w;
        dash[1] = w;
        cairo_set_dash(cr, dash, 2, 0);
        break;
    case LT_DOT:
        dash[0] = w;
        cairo_set_dash(cr, dash, 1, 0);
        break;
    case LT_DASHDOT:
        dash[0] = 3*w;
        dash[1] = w;
        dash[2] = w;
        dash[3] = w;
        cairo_set_dash(cr, dash, 4, 0);
        break;
    case LT_DASHDOTDOT:
        dash[0] = 3*w;
        dash[1] = w;
        dash[2] = w;
        dash[3] = w;
        dash[4] = w;
        dash[5] = w;
        cairo_set_dash(cr, dash, 6, 0);
        break;
    case LT_NONE:
        dash[0] = 0;
        dash[1] = w;
        cairo_set_dash(cr, dash, 2, 0);
        break;
    case LT_DOTDOTSPACE:
        dash[0] = w;
        dash[1] = w;
        dash[2] = w;
        dash[3] = w*3;
        cairo_set_dash(cr, dash, 4, 0);
        break;
    case LT_LONGDASH:
        dash[0] = w*6;
        dash[1] = w;
        cairo_set_dash(cr, dash, 2, 0);
        break;
    case LT_DASHDASHDOT:
        dash[0] = 3*w;
        dash[1] = w;
        dash[2] = 2*w;
        dash[3] = w;
        dash[4] = w;
        dash[5] = w;
        cairo_set_dash(cr, dash, 6, 0);
        break;
    }
}


static void
_cairo_set_line(cairo_t *cr, cgm_attributes_t *a)
{
    double w = a_get_line_width(a);
    _cairo_set_color(cr, a_get_line_color(a));
    _cairo_set_line_type(cr, a_get_line_type(a), w);
    cairo_set_line_width(cr, w);
}


static void
_cairo_set_edge(cairo_t *cr, cgm_attributes_t *a)
{
    double w = a_get_edge_width(a);
    _cairo_set_color(cr, a_get_edge_color(a));
    _cairo_set_line_type(cr, a_get_edge_type(a), w);
    cairo_set_line_width(cr, a_get_edge_width(a));
}


static void
_cairo_set_fill(cairo_t *cr, cgm_attributes_t *a)
{
    _cairo_set_color(cr, a_get_fill_color(a));
}


static void
_cairo_draw_rect(cairo_t *cr, const rect_t *rect)
{
    cairo_rectangle(cr, rect->pos.x, rect->pos.y, rect->size.x, rect->size.y);
}


static void
_cairo_draw_path(cairo_t *cr, const path_t *path)
{
    const loa_segment_t *seg = path->segment;
    int i;
    for(i=0;i<path->segment_count;++i){
        const loa_point_t *p = path->point + seg->point;
        switch(seg->type){
        case LOA_GOTO:
            cairo_move_to(cr, p[0].x, p[0].y);
            break;
        case LOA_LINE:
            cairo_line_to(cr, p[0].x, p[0].y);
            break;
        case LOA_CURVE:
            cairo_curve_to(cr, p[0].x, p[0].y, p[1].x, p[1].y, p[2].x, p[2].y);
            break;
        case LOA_CLOSE:
            cairo_close_path(cr);
            break;
        }
        ++seg;
    }
}


static void
_cairo_draw_arc(cairo_t *cr, const arc_t *arc)
{
    cairo_save(cr);
    cairo_translate(cr, arc->pos.x, arc->pos.y);
    cairo_rotate(cr, arc->orientation);
    cairo_scale(cr, arc->size.x, arc->size.y);
    if(arc->negative){
        cairo_arc_negative(cr, 0, 0, 1, arc->startangle, arc->endangle);
    }else{
        cairo_arc(cr, 0, 0, 1, arc->startangle, arc->endangle);
    }
    if(CT_PIE == arc->close_type){
        cairo_line_to(cr, 0, 0);
        cairo_close_path(cr);
    }
    cairo_restore(cr);
}


static void
_cairo_draw_object(cairo_t *cr, const loa_object_t *shape)
{
    switch(shape->type){
    case OT_RECT:
        _cairo_draw_rect(cr, &shape->rect);
        break;
    case OT_PATH:
        _cairo_draw_path(cr, &shape->path);
        break;
    case OT_ARC:
        _cairo_draw_arc(cr, &shape->arc);
        break;
    case OT_IMAGE:
        break;
    }
}


static void
_cairo_set_font(cairo_t *cr, FontEntry *font)
{
    const char *family = font->font_name;
    cairo_font_face_t *ff;
    DEBUG_p("setting font family '%s'", family);
    ff = cairo_toy_font_face_create(family,
                                    font->type & 1 ? CAIRO_FONT_SLANT_ITALIC : CAIRO_FONT_SLANT_NORMAL,
                                    font->type & 2 ? CAIRO_FONT_WEIGHT_BOLD : CAIRO_FONT_WEIGHT_NORMAL);
    cairo_set_font_face(cr, ff);
    cairo_font_face_destroy(ff);
}


static void
_on_insert_page(void *userdata, cgm_attributes_t *a)
{
    //context_t *c = (context_t*)userdata;
    DEBUG_p("begin");
}


static void
_on_begin_group(void *userdata)
{
    DEBUG_p("begin");
}


void _on_end_group(void *userdata)
{
    DEBUG_p("begin");
}


static void
_on_end_grouping(void *userdata)
{
    DEBUG_p("begin");
}


static void
_on_stroke_shape(void *userdata, cgm_attributes_t *a, const loa_object_t *shape)
{
    context_t *c = (context_t*)userdata;
    _cairo_draw_object(c->cr, shape);
    _cairo_set_line(c->cr, a);
    cairo_stroke(c->cr);
}


static void
_on_fill_shape(void *userdata, cgm_attributes_t *a, const loa_object_t *shape)
{
    context_t *c = (context_t*)userdata;
    _cairo_draw_object(c->cr, shape);
    _cairo_set_fill(c->cr, a);
    if(EV_ON == a->eEdgeVisibility){
        cairo_fill_preserve(c->cr);
        _cairo_set_edge(c->cr, a);
        cairo_stroke(c->cr);
    }else{
        cairo_fill(c->cr);
    }
}


static void
_on_draw_bitmap(void *userdata, const loa_point_t *p, nile_t *ni)
{
    context_t *c = (context_t*)userdata;
    cairo_surface_t *surface = nil_cairo_surface_create(ni);
    cairo_pattern_t *pattern = cairo_pattern_create_for_surface(surface);
    DEBUG_p("begin");
    cairo_surface_destroy(surface);
    cairo_pattern_set_filter(pattern, CAIRO_FILTER_NEAREST);
    cairo_save(c->cr);
    cairo_translate(c->cr, p[0].x, p[0].y);
    {
        cairo_matrix_t m;
        m.xx = 1;
        m.xy = (p[1].x-p[2].x)/(p[1].y-p[2].y);
        m.yx = (p[2].y-p[0].y)/(p[2].x-p[0].x);
        m.yy = 1;
        m.x0 = 0;
        m.y0 = 0;
        cairo_transform(c->cr, &m);
    }
    cairo_scale(c->cr, (p[2].x-p[0].x)/ni->width, (p[1].y-p[2].y)/ni->height);
    cairo_set_source(c->cr, pattern);
    cairo_paint(c->cr);
    cairo_restore(c->cr);
    cairo_pattern_destroy(pattern);
}


static void
_on_draw_text(void *userdata, rect_t *rect, cgm_attributes_t *a, const char *s, FinalFlag flag)
{
    context_t *c = (context_t*)userdata;
    cairo_text_extents_t ext;
    cairo_matrix_t matrix;
    DEBUG_p("[%s]",s);
    cairo_save(c->cr);
    _cairo_set_font(c->cr, a_get_font_entry(a));
    cairo_set_font_size(c->cr, a->nCharacterHeight);
    cairo_text_extents(c->cr, s, &ext);
    c->cursor_x = rect->pos.x;
    c->cursor_y = rect->pos.y;
    switch(a->eTextAlignmentH){
    case TAH_CENTER:
        c->cursor_x -= ext.width/2;
        break;
    case TAH_RIGHT:
        c->cursor_x -= ext.width;
        break;
    default:
        break;
    }
    switch(a->eTextAlignmentV){
    case TAV_TOP:
        c->cursor_y -= ext.height/2;
        break;
    case TAV_HALF:
        c->cursor_y -= ext.height;
        break;
    default:
        break;
    }
    cairo_translate(c->cr, c->cursor_x, c->cursor_y);
    c->cursor_x += ext.x_advance;
    c->cursor_y += ext.y_advance;
    cairo_get_matrix(c->cr, &matrix);
    if(matrix.xx < 0)
        cairo_scale(c->cr, -1, 1);
    if(matrix.yy < 0)
        cairo_scale(c->cr, 1, -1);
    cairo_move_to(c->cr, 0, 0);

#ifdef _USE_DEBUG
    _cairo_set_color(c->cr, 0xff00ff);
    cairo_arc(c->cr, 0, 0, 25, 0, 2*M_PI);
    cairo_fill(c->cr);
#endif

    switch(a->eTextPath){
    case TPR_RIGHT:
        break;
    case TPR_LEFT:
        cairo_rotate(c->cr, 1.0*M_PI);
        break;
    case TPR_UP:
        cairo_rotate(c->cr, 1.5*M_PI);
        break;
    case TPR_DOWN:
        cairo_rotate(c->cr, 0.5*M_PI);
        break;
    }
    //cairo_rotate(c->cr, 0.25*M_PI);
    _cairo_set_color(c->cr, a_get_text_color(a));
#if 1
    cairo_show_text(c->cr, s);
#else
    cairo_text_path(c->cr, s);
    cairo_fill(c->cr);
#endif
    cairo_restore(c->cr);
}


static void
_on_append_text(void *userdata, cgm_attributes_t *a, const char *s, FinalFlag f)
{
    context_t *c = (context_t*)userdata;
    rect_t rect;
    rect.pos.x = c->cursor_x;
    rect.pos.y = c->cursor_y;
    rect.size.x = 0;
    rect.size.y = 0;
    _on_draw_text(userdata, &rect, a, s, f);
}


static void
_on_first_output(void *userdata, int width, int height, int xadd, int yadd)
{
    context_t *c = (context_t*)userdata;
    if(!c->cr){
        c->cr = c->pagefunc(c->userdata, width, height);
        cairo_scale(c->cr, 1, -1);
        cairo_translate(c->cr, xadd, yadd);
        //cairo_translate(c->cr, 0, -height);
    }
    DEBUG_p("size: %dx%d", width, height);
}


static void
_on_end_figure(void *userdata)
{
    DEBUG_p("begin");
}


static void
_on_set_gradient_offset(void *userdata, long nHorzOfs, long nVertOfs )
{
    DEBUG_p("begin");
}


int cgm_draw_cairo(const uint8_t *data, size_t len, cgm_draw_page_func_t pagefunc, void *userdata)
{
    static const cgm_interface_t cgm_vt = {
        .InsertPage = _on_insert_page,
        .BeginGroup = _on_begin_group,
        .EndGroup = _on_end_group,
        .EndGrouping = _on_end_grouping,
        .DrawBitmap = _on_draw_bitmap,
        .DrawText = _on_draw_text,
        .AppendText = _on_append_text,
        .stroke_shape = _on_stroke_shape,
        .fill_shape = _on_fill_shape,
        .FirstOutPut = _on_first_output,
        .EndFigure = _on_end_figure,
        .SetGradientOffset = _on_set_gradient_offset
    };
    context_t c;
    int res;
    c.pagefunc = pagefunc;
    c.userdata = userdata;
    c.cr = NULL;

    res = cgm_interprete((uint8_t*)data, len, &cgm_vt, &c);

    if(c.cr)
        cairo_destroy(c.cr);

    return res;
}
