/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 * Copyright (C) 2017  Bastien Nocera <hadess@hadess.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */
#include <debug.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* memcpy() */

#include "abydos-archive.h"

#define BUFFER_SIZE (64 * 1024)

#ifndef MIN
# define MIN(a,b) ((a)<(b)?(a):(b))
#endif


abydos_archive_t *
abydos_archive_init(abydos_archive_t *archive, abydos_archive_type_t type, const char *filename)
{
    archive->type = type;
    switch(type){
#if defined(USE_LIBARCHIVE) || defined(USE_LIBUNARR)
    case AT_7Z:
    case AT_TAR:
    case AT_ZIP:
# ifdef USE_LIBARCHIVE
        archive->libar.archive = archive_read_new();
        if(AT_7Z == type){
            archive_read_support_format_7zip (archive->libar.archive);
        }else if(AT_TAR == type){
            archive_read_support_format_tar (archive->libar.archive);
        }else if(AT_ZIP == type){
            archive_read_support_format_zip (archive->libar.archive);
        }
        if(archive_read_open_filename(archive->libar.archive, filename, BUFFER_SIZE) != ARCHIVE_OK)
            return NULL;
# else
        archive->unarr.stream = ar_open_file(filename);
        if(!archive->unarr.stream)
            return NULL;
        if(AT_7Z == type){
            archive->unarr.archive = ar_open_7z_archive(archive->unarr.stream);
        }else if(AT_TAR == type){
            archive->unarr.archive = ar_open_tar_archive(archive->unarr.stream);
        }else if(AT_ZIP == type){
            archive->unarr.archive = ar_open_zip_archive(archive->unarr.stream, false);
        }
        if(!archive->unarr.archive)
            return NULL;
# endif
        break;
#endif
#ifdef USE_ACEUNPACK
    case AT_ACE:
        archive->ace = aceunpack_new(filename);
        if(!archive->ace)
            return NULL;
        break;
#endif
#if 1
    case AT_RAR:
        if(dmc_unrar_archive_init(&archive->dmc.archive) != DMC_UNRAR_OK)
            return NULL;
        if(dmc_unrar_archive_open_path(&archive->dmc.archive, filename) != DMC_UNRAR_OK)
            return NULL;
        archive->dmc.entry = NULL;
        archive->dmc.buffer = NULL;
        archive->dmc.buffer_size = 0;
        archive->dmc.offset = -1;
        break;
#elif defined(USE_LIBUNARR)
    case AT_RAR:
        archive->unarr.stream = ar_open_file(filename);
        if(!archive->unarr.stream)
            return NULL;
        archive->unarr.archive = ar_open_rar_archive(archive->unarr.stream);
        if(!archive->unarr.archive)
            return NULL;
        break;
#endif
    }
    return archive;
}


void
abydos_archive_done(abydos_archive_t *archive)
{
    switch(archive->type) {
#if defined(USE_LIBARCHIVE) || defined(USE_LIBUNARR)
    case AT_7Z:
    case AT_TAR:
    case AT_ZIP:
# ifdef USE_LIBARCHIVE
        archive_free(archive->libar.archive);
# else
        ar_close_archive(archive->unarr.archive);
        ar_close(archive->unarr.stream);
# endif
        break;
#endif
#ifdef USE_ACEUNPACK
    case AT_ACE:
        aceunpack_free(archive->ace);
        break;
#endif
#if 1
    case AT_RAR:
        dmc_unrar_archive_close(&archive->dmc.archive);
        if(archive->dmc.buffer)
            free(archive->dmc.buffer);
        break;
#elif defined(USE_LIBUNARR)
    case AT_RAR:
        ar_close_archive(archive->unarr.archive);
        ar_close(archive->unarr.stream);
        break;
#endif
    }
}


int
abydos_archive_next(abydos_archive_t *archive)
{
    switch (archive->type) {
#if defined(USE_LIBARCHIVE) || defined(USE_LIBUNARR)
    case AT_7Z:
    case AT_TAR:
    case AT_ZIP:
# ifdef USE_LIBARCHIVE
        for(;;){
            if (archive_read_next_header (archive->libar.archive, &archive->libar.entry) != ARCHIVE_OK)
                return -1;

            if (archive_entry_filetype (archive->libar.entry) != AE_IFREG)
                continue;

            DEBUG_p("At header for file '%s'", archive_entry_pathname (archive->libar.entry));
            return 0;
        }
# else
        return ar_parse_entry(archive->unarr.archive);
# endif
#endif
#ifdef USE_ACEUNPACK
    case AT_ACE:
        return aceunpack_next(archive->ace);
#endif
#if 1
    case AT_RAR:
        if(archive->dmc.entry){
            archive->dmc.entry = archive->dmc.entry->solid_next;
        }else{
            archive->dmc.entry = archive->dmc.archive.internal_state->files;
        }
        archive->dmc.offset = -1;
        return archive->dmc.entry ? 0 : -1;
#elif defined(USE_LIBUNARR)
    case AT_RAR:
        return ar_parse_entry(archive->unarr.archive);
#endif
    }
    return -1;
}

const char *
abydos_archive_entry_name(abydos_archive_t *archive)
{
    switch(archive->type){
#if defined(USE_LIBARCHIVE) || defined(USE_LIBUNARR)
    case AT_7Z:
    case AT_TAR:
    case AT_ZIP:
# ifdef USE_LIBARCHIVE
        if(!archive->libar.entry)
            return NULL;
        return archive_entry_pathname(archive->libar.entry);
# else
        if(!archive->unarr.archive)
            return NULL;
        return ar_entry_get_name (archive->unarr.archive);
# endif
#endif
#ifdef USE_ACEUNPACK
    case AT_ACE:
        return aceunpack_get_name(archive->ace);
#endif
#if 1
    case AT_RAR:
        if(!archive->dmc.entry)
            return NULL;
        if(!dmc_unrar_get_filename(&archive->dmc.archive, archive->dmc.entry, archive->dmc.filename, 4096) != DMC_UNRAR_OK)
            return NULL;
        return archive->dmc.filename;
#elif defined(USE_LIBUNARR)
    case AT_RAR:
        if(!archive->unarr.archive)
            return NULL;
        return ar_entry_get_name (archive->unarr.archive);
#endif
    }
    return NULL;
}


ssize_t
abydos_archive_get_entry_size(abydos_archive_t *archive)
{
    switch(archive->type) {
#if defined(USE_LIBARCHIVE) || defined(USE_LIBUNARR)
    case AT_7Z:
    case AT_TAR:
    case AT_ZIP:
# ifdef USE_LIBARCHIVE
        if(!archive->libar.entry)
            return -1;
        return archive_entry_size (archive->libar.entry);
# else
        if(!archive->unarr.archive)
            return -1;
        return ar_entry_get_size (archive->unarr.archive);
# endif
#endif
#ifdef USE_ACEUNPACK
    case AT_ACE:
        return aceunpack_get_size(archive->ace);
#endif
#if 1
    case AT_RAR:
        if(!archive->dmc.entry)
            return -1;
        return archive->dmc.entry->uncompressed_size;
#elif defined(USE_LIBUNARR)
    case AT_RAR:
        if(!archive->unarr.archive)
            return -1;
        return ar_entry_get_size (archive->unarr.archive);
#endif
    }
    return -1;
}


int
abydos_archive_entry_is_encrypted(abydos_archive_t *archive)
{
    switch (archive->type) {
#if defined(USE_LIBARCHIVE) || defined(USE_LIBUNARR)
    case AT_7Z:
    case AT_TAR:
    case AT_ZIP:
# ifdef USE_LIBARCHIVE
        if(!archive->libar.entry)
            return 0;
        return archive_entry_is_encrypted(archive->libar.entry);
# else
        return 0;
# endif
#endif
#ifdef USE_ACEUNPACK
    case AT_ACE:
        return 0;
#endif
#if 1
    case AT_RAR:
        if(!archive->dmc.entry)
            return -1;
        return archive->dmc.entry->is_encrypted;
#elif defined(USE_LIBUNARR)
    case AT_RAR:
        if(!archive->unarr.archive)
            return 0;
        /* password-protected RAR is not even detected right now */
        return 0;
#endif
    }
    return 0;
}


ssize_t
abydos_archive_entry_read(abydos_archive_t *archive, void *buf, size_t bytes)
{
    switch (archive->type) {
#if defined(USE_LIBARCHIVE) || defined(USE_LIBUNARR)
    case AT_7Z:
    case AT_TAR:
    case AT_ZIP:
# ifdef USE_LIBARCHIVE
        if(!archive->libar.entry)
            return -1;
        return archive_read_data (archive->libar.archive, buf, bytes);
# else
        if(!archive->unarr.archive)
            return -1;
        return ar_entry_uncompress(archive->unarr.archive, buf, bytes) ? bytes : -1;
# endif
#endif
#ifdef USE_ACEUNPACK
    case AT_ACE:
        return aceunpack_read(archive->ace, buf, bytes);
#endif
#if 1
    case AT_RAR:
        if(!archive->dmc.entry)
            return -1;
        if(archive->dmc.offset < 0){
            if(archive->dmc.buffer_size < archive->dmc.entry->uncompressed_size){
                if(archive->dmc.buffer){
                    archive->dmc.buffer = realloc(archive->dmc.buffer, archive->dmc.entry->uncompressed_size);
                }else{
                    archive->dmc.buffer = malloc(archive->dmc.entry->uncompressed_size);
                }
                archive->dmc.buffer_size = archive->dmc.entry->uncompressed_size;
            }
            if(dmc_unrar_extract_file_to_mem(&archive->dmc.archive, archive->dmc.entry, archive->dmc.buffer, archive->dmc.entry->uncompressed_size, NULL, true) != DMC_UNRAR_OK)
                return -1;
            archive->dmc.offset = 0;
        }
        {
            bytes = MIN(bytes, archive->dmc.entry->uncompressed_size - archive->dmc.offset);
            if(bytes > 0){
                memcpy(buf, archive->dmc.buffer + archive->dmc.offset, bytes);
                archive->dmc.offset += bytes;
            }
            return bytes;
        }
#elif defined(USE_LIBUNARR)
    case AT_RAR:
        if(!archive->unarr.archive)
            return -1;
        return ar_entry_uncompress(archive->unarr.archive, buf, bytes) ? bytes : -1;
#endif
    }
    return -1;
}
