/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <endian.h>                       /* be32toh() */
#include <stdlib.h>                       /* malloc() */
#include <string.h>                       /* strcmp() */

#include "buffer.h"


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;

    /* progressive loading */
    int stride;
    int x;
    int y;
    enum{
        S_HEAD,
        S_BODY,
        S_DONE
    }state;
    abydos_buffer_t inbuf;
    abydos_plugin_info_func_t info_func;
    abydos_plugin_progress_func_t progress_func;
    void *userdata;
};


static const char **
_farbfeld_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-farbfeld",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_farbfeld_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    info->threadsafe = 1;
    h->surface = NULL;
    return h;
}


static void
_farbfeld_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


static void
_farbfeld_progressive_begin(abydos_plugin_handle_t *h, abydos_plugin_info_func_t info_func, abydos_plugin_progress_func_t progress_func, void *userdata)
{
    h->state = S_HEAD;
    abydos_buffer_clear(&h->inbuf);
    h->info_func = info_func;
    h->progress_func = progress_func;
    h->userdata = userdata;
}


static int
_farbfeld_progressive_feed(abydos_plugin_handle_t *h ,const char *data, size_t len)
{
    abydos_buffer_add(&h->inbuf,(const uint8_t*)data,len);
    const uint8_t *bytes;
    uint32_t *pixels;
    cairo_rectangle_int_t rect;
    switch(h->state){
    case S_HEAD:
        if(abydos_buffer_size(&h->inbuf)<16)
            break;
        bytes = abydos_buffer_get(&h->inbuf,16);
        if(strcmp((char*)bytes,"farbfeld"))
            return -1;
        h->info->width = be32toh(*(uint32_t*)(bytes+8));
        h->info->height = be32toh(*(uint32_t*)(bytes+12));
        h->surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,h->info->width,h->info->height);
        h->stride = cairo_image_surface_get_stride(h->surface)/sizeof(uint32_t);
        if(h->info_func)
            h->info_func(h->userdata);
        h->state = S_BODY;
        h->x = 0;
        h->y = 0;
        /* fall through */
    case S_BODY:
        if(abydos_buffer_size(&h->inbuf)<8)
            return 0;
        pixels = (uint32_t*)cairo_image_surface_get_data(h->surface);
        rect.x = h->x;
        rect.y = h->y;
        while(abydos_buffer_size(&h->inbuf)>=8){
            int alpha;
            bytes = abydos_buffer_get(&h->inbuf,8);
            alpha = bytes[6];
            pixels[h->x+h->y*h->stride] = (alpha*bytes[0]/255)<<16 |
                                          (alpha*bytes[2]/255)<< 8 |
                                          (alpha*bytes[4]/255)<< 0 |
                                          alpha<<24;
            if(++h->x >= h->info->width){
                if(++h->y >= h->info->height){
                    h->state = S_DONE;
                    break;
                }
                h->x = 0;
            }
        }
        cairo_surface_mark_dirty(h->surface);
        if(h->progress_func){
            rect.height = rect.y - h->y + 1;
            if(rect.height > 1){
                rect.width = rect.x - h->x;
            }else{
                rect.x = 0;
                rect.width = h->info->width;
            }
            h->progress_func(h->userdata,&rect);
        }
        break;
    case S_DONE:
        /* silently igonore additional bytes */
        return 0;
    }
    abydos_buffer_privatize(&h->inbuf);
    return 0;
}


static int
_farbfeld_progressive_end(abydos_plugin_handle_t *h)
{
    return S_DONE == h->state ? 0 : -1;
}


static cairo_surface_t *
_farbfeld_get_image_surface(abydos_plugin_handle_t *h,int page,cairo_bool_t *layer_visibility,int variant, double time, int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0(){
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _farbfeld_get_mime_types,
        .new = _farbfeld_new,
        .free = _farbfeld_free,
        .progressive_begin = _farbfeld_progressive_begin,
        .progressive_feed = _farbfeld_progressive_feed,
        .progressive_end = _farbfeld_progressive_end,
        .get_image_surface = _farbfeld_get_image_surface
    };
    return &p;
}
