/*
 * Copyright (C) 2020 - 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <FreeImage.h>
#include <math.h>
#include <nil-cairo.h>
#include <stdlib.h>
#include <string.h>

/* #define ENABLE_EXR */
#define ENABLE_GIF
#define ENABLE_ICO
#define ENABLE_JNG
/* #define ENABLE_JPEG2000 */
/* #define ENABLE_JXR */
/* #define ENABLE_KOALA */
/* #define ENABLE_PFM */
/* #define ENABLE_PICT */
/* #define ENABLE_SUNRAS */
//#define ENABLE_SGI /* Not working */
/* #define ENABLE_TIFF */
/* #define ENABLE_WEBP */

/*

Supported formats

-------------v-----------------------------------v--------------------------
* FIF_BMP    | Windows or OS/2 Bitmap            | image/bmp
  FIF_CUT    | Dr. Halo                          | image/x-cut
* FIF_DDS    | DirectX Surface                   | image/x-direct-draw-surface
* FIF_EXR    | ILM OpenEXR                       | image/x-exr
  FIF_FAXG3  | Raw fax format CCITT G.3          | ???
* FIF_GIF    | Graphics Interchange Format       | image/gif
* FIF_HDR    | High Dynamic Range                | image/vnd.radiance
* FIF_ICO    | Windows Icon                      | image/vnd.microsoft.icon
* FIF_IFF    | IFF Interleaved Bitmap            | image/x-ilbm
  FIF_J2K    | JPEG-2000 codestream              | ???
* FIF_JNG    | JPEG Network Graphics             | image/x-jng
* FIF_JP2    | JPEG-2000 File Format             | image/jpeg2000
* FIF_JPEG   | JPEG - JFIF Compliant             | image/jpeg
* FIF_JXR    | JPEG XR image format              | image/vnd.ms-photo
* FIF_KOALA  | C64 Koala Graphics                | image/x-koa, image/x-koala
* FIF_MNG    | Multiple Network Graphics         | video/x-mng
  FIF_PBM    | Portable Bitmap (ASCII)           | image/x-portable-bitmap
  FIF_PBMRAW | Portable Bitmap (RAW)             | image/x-portable-bitmap
* FIF_PCD    | Kodak PhotoCD pcd                 | image/x-photo-cd
* FIF_PCX    | Zsoft Paintbrush                  | image/vnd.zbrush.pcx, image/x-pcx
* FIF_PFM    | Portable Floatmap                 | image/x-portable-floatmap
  FIF_PGM    | Portable Greymap (ASCII)          | image/x-portable-graymap, image/x-portable-greymap
  FIF_PGMRAW | Portable Greymap (RAW)            | image/x-portable-graymap, image/x-portable-greymap
* FIF_PICT   | Macintosh PICT                    | image/pict, image/x-pict
* FIF_PNG    | Portable Network Graphics         | image/png
  FIF_PPM    | Portable Pixelmap (ASCII)         | image/x-portable-pixmap
  FIF_PPMRAW | Portable Pixelmap (RAW)           | image/x-portable-pixmap
* FIF_PSD    | Adobe Photoshop                   | image/vnd.adobe.photoshop
* FIF_RAS    | Sun Raster Image                  | image/x-cmu-raster, image/x-sun-raster
  FIF_RAW    | RAW camera image                  | *
* FIF_SGI    | Silicon Graphics SGI image format | image/x-rgb, image/x-sgi
* FIF_TARGA  | Truevision Targa                  | image/x-tga
* FIF_TIFF   | Tagged Image File Format          | image/tiff
* FIF_WBMP   | Wireless Bitmap                   | image/vnd.wap.wbmp
* FIF_WEBP   | Google WebP image format          | image/webp
* FIF_XBM    | X11 Bitmap Format                 | image/x-xbitmap, image/x-xbmi
* FIF_XPM    | X11 Pixmap Format                 | image/x-xpixmap, image/x-xpmi
-------------^-----------------------------------^--------------------------

*/

typedef struct{
    cairo_surface_t *surface;
    union{
        abydos_size_t size;
        double duration;
    };
}multi_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    FREE_IMAGE_FORMAT fif;
    int flags;
    enum{
        MT_NONE = 0,
        MT_PAGE,
        MT_VARIANT,
        MT_FRAME
    }multi_type;
    multi_t *multi;
    int multi_count;
};


static void
_fi_done()
{
    FreeImage_DeInitialise();
}


static const char **
_fi_get_mime_types()
{
    static const char *mime_types[] = {
        "image/bmp",
        "image/x-direct-draw-surface", "image/x-dds",
#ifdef ENABLE_EXR
        "image/x-exr",
#endif
#ifdef ENABLE_GIF
        "image/gif",
#endif
        "image/vnd.radiance",
#ifdef ENABLE_ICO
        "image/vnd.microsoft.icon",
#endif
        "image/x-ilbm",
#ifdef ENABLE_JNG
        "image/x-jng", "video/x-jng",
#endif
#ifdef ENABLE_JPEG2000
        "image/jpeg2000",
#endif
        "image/jpeg",
#ifdef ENABLE_JXR
        "image/vnd.ms-photo",
#endif
#ifdef ENABLE_KOALA
        "image/x-koa", "image/x-koala",
#endif
        "video/x-mng", "image/x-mng",
        "image/x-photo-cd",
        "image/vnd.zbrush.pcx",
#ifdef ENABLE_PICT
        "image/pict", "image/x-pict",
#endif
#ifdef ENABLE_PFM
        "image/x-portable-floatmap",
#endif
        "image/png",
#ifdef ENABLE_PSD
        "image/vnd.adobe.photoshop",
#endif
#ifdef ENABLE_SUNRAS
        "image/x-cmu-raster", "image/x-sun-raster",
#endif
#ifdef ENABLE_SGI
        "image/x-rgb", "image/x-sgi",
#endif
        "image/x-tga",
#ifdef ENABLE_TIFF
        "image/tiff",
#endif
        "image/vnd.wap.wbmp", "image/x-wireless-bitmap",
#ifdef ENABLE_WEBP
        "image/webp",
#endif
        "image/x-xbitmap", "image/x-xbmi",
        "image/x-xpixmap", "image/x-xpmi",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_fi_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->flags = 0;
    h->multi_type = MT_NONE;
    if(!strcasecmp(mime_type,"image/bmp")){
        h->fif = FIF_BMP;
        h->flags = NILE_FLIPPED;
    }else
    if(!strcasecmp(mime_type,"image/x-direct-draw-surface")||
       !strcasecmp(mime_type,"image/x-dds"))
    {
        h->fif = FIF_DDS;
        h->flags = NILE_FLIPPED;
    }else
#ifdef ENABLE_EXR
    if(!strcasecmp(mime_type,"image/x-exr")){
        h->fif = FIF_EXR;
    }else
#endif
#ifdef ENABLE_GIF
    if(!strcasecmp(mime_type,"image/gif")){
        h->fif = FIF_GIF;
        h->multi_type = MT_FRAME;
    }else
#endif
    if(!strcasecmp(mime_type,"image/vnd.radiance")){
        h->fif = FIF_HDR;
    }else
#ifdef ENABLE_ICO
    if(!strcasecmp(mime_type,"image/vnd.microsoft.icon")){
        h->fif = FIF_ICO;
        h->flags = NILE_FLIPPED;
        h->multi_type = MT_VARIANT;
    }else
#endif
    if(!strcasecmp(mime_type,"image/x-ilbm")){
        h->fif = FIF_LBM;
        h->flags = NILE_FLIPPED;
    }else
#ifdef ENABLE_JNG
    if(!strcasecmp(mime_type,"image/x-jng")||
       !strcasecmp(mime_type,"video/x-jng"))
    {
        h->fif = FIF_JNG;
        h->flags = NILE_FLIPPED;
    }else
#endif
#ifdef ENABLE_JPEG2000
    if(!strcasecmp(mime_type,"image/jpeg2000")){
        h->fif = FIF_JP2;
    }else
#endif
    if(!strcasecmp(mime_type,"image/jpeg")){
        h->fif = FIF_JPEG;
        h->flags = NILE_FLIPPED;
    }else
#ifdef ENABLE_JXR
    if(!strcasecmp(mime_type,"image/vnd.ms-photo")){
        h->fif = FIF_JXR;
    }else
#endif
#ifdef ENABLE_KOALA
    if(!strcasecmp(mime_type,"image/x-koa")||
             !strcasecmp(mime_type,"image/x-koala"))
    {
        h->fif = FIF_KOALA;
    }else
#endif
    if(!strcasecmp(mime_type,"video/x-mng") ||
       !strcasecmp(mime_type,"image/x-mng"))
    {
        h->fif = FIF_MNG;
        h->flags = NILE_FLIPPED;
    }else
    if(!strcasecmp(mime_type,"image/x-photo-cd")){
        h->fif = FIF_PCD;
    }else
    if(!strcasecmp(mime_type,"image/vnd.zbrush.pcx")){
        h->fif = FIF_PCX;
        h->flags = NILE_FLIPPED;
    }else
#ifdef ENABLE_PICT
    if(!strcasecmp(mime_type,"image/pict")||
       !strcasecmp(mime_type,"image/x-pict"))
    {
        h->fif = FIF_PICT;
    }else
#endif
#ifdef ENABLE_PFM
    if(!strcasecmp(mime_type,"image/x-portable-floatmap")){
        h->fif = FIF_PFM;
    }else
#endif
    if(!strcasecmp(mime_type,"image/png")){
        h->fif = FIF_PNG;
        h->flags = NILE_FLIPPED;
    }else
#ifdef ENABLE_PSD
    if(!strcasecmp(mime_type,"image/vnd.adobe.photoshop")){
        h->fif = FIF_PSD;
    }else
#endif
#ifdef ENABLE_SUNRAS
    if(!strcasecmp(mime_type,"image/x-cmu-raster")||
       !strcasecmp(mime_type,"image/x-sun-raster"))
    {
        h->fif = FIF_RAS;
    }else
#endif
#ifdef ENABLE_SGI
    if(!strcasecmp(mime_type,"image/x-rgb")||
       !strcasecmp(mime_type,"image/x-sgi"))
    {
        h->fif = FIF_SGI;
    }else
#endif
    if(!strcasecmp(mime_type,"image/x-tga")){
        h->fif = FIF_TARGA;
        h->flags = NILE_FLIPPED;
    }else
#ifdef ENABLE_TIFF
    if(!strcasecmp(mime_type,"image/tiff")){
        h->fif = FIF_TIFF;
    }else
#endif
    if(!strcasecmp(mime_type,"image/vnd.wap.wbmp") ||
       !strcasecmp(mime_type,"image/x-wireless-bitmap"))
    {
        h->fif = FIF_WBMP;
    }else
#ifdef ENABLE_WEBP
    if(!strcasecmp(mime_type,"image/webp")){
        h->fif = FIF_WEBP;
    }else
#endif
    if(!strcasecmp(mime_type,"image/x-xbitmap")||
       !strcasecmp(mime_type,"image/x-xbmi"))
    {
        h->fif = FIF_XBM;
    }else
    if(!strcasecmp(mime_type,"image/x-xpixmap")||
       !strcasecmp(mime_type,"image/x-xpmi"))
    {
        h->fif = FIF_XPM;
        h->flags = NILE_FLIPPED;
    }else
    {
        free(h);
        return NULL;
    }
    h->info = info;
    h->info->threadsafe = 1;
    h->multi = NULL;
    h->multi_count = 0;
    return h;
}


static void
_fi_free(abydos_plugin_handle_t *h)
{
    if(h->multi){
        int i;
        for(i=0;i<h->multi_count;++i)
            if(h->multi[i].surface)
                cairo_surface_destroy(h->multi[i].surface);
        free(h->multi);
    }
    free(h);
}


static int
_palette_from_bitmap(uint32_t *dst,FIBITMAP *bm){
    RGBQUAD* src = FreeImage_GetPalette(bm);
    int count = FreeImage_GetColorsUsed(bm);
    int i;
    DEBUG_p("%d colors in the palette",count);
    for(i=0;i<count;++i)
        dst[i] = src[i].rgbRed<<16 |
                 src[i].rgbGreen<<8 |
                 src[i].rgbBlue<<0 |
                 255<<24;
    count = FreeImage_GetTransparencyCount(bm);
    DEBUG_p("%d transparent colors",count);
    if(count){
        const BYTE *table = FreeImage_GetTransparencyTable(bm);
        for(i=0;i<count;++i)
            dst[i] = (dst[i] & 0x00ffffff) | table[i]<<24;
    }
    return count;
}


static cairo_surface_t *
_surface_from_bitonal(const uint8_t *src, int width, int height, int stride)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB16_565,width,height);
    uint16_t *dst = (uint16_t*)cairo_image_surface_get_data(surface);
    int dst_skip = cairo_image_surface_get_stride(surface)/sizeof(uint16_t)-width;
    int src_skip;

    src += stride * (height-1);
    src_skip = -stride*2;

    while(height-->0){
        int x;
        for(x=0;x<width;++x){
            int bits;
            if(!(x&7))
                bits = *src++;
            bits<<=1;
            *dst++ = bits&0x100 ? 0xffff : 0;
        }
        src += src_skip;
        dst += dst_skip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static cairo_surface_t *
_surface_from_4bit(const uint8_t *src, int width, int height, int stride, const uint32_t *palette){
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_skip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t)-width;
    int src_skip;

    src += stride * (height-1);
    src_skip = -stride*2;

    while(height-->0){
        int x;
        for(x=0;x<width;++x){
            int bits;
            if(!(x&1))
                bits = *src++;
            bits<<=4;
            *dst++ = palette[bits>>8];
        }
        src += src_skip;
        dst += dst_skip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static cairo_surface_t *
_surface_from_8bit_transparent(const uint8_t *src, int width, int height, int stride, const uint32_t *palette){
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_skip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t)-width;
    int src_skip;

    src += stride * (height-1);
    src_skip = -stride*2;

    while(height-->0){
        int x;
        for(x=0;x<width;++x)
            *dst++ = palette[*src++];
        src += src_skip;
        dst += dst_skip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static cairo_surface_t *
_surface_from_rgbf(const float *src, int width, int height, int stride)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_skip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t)-width;
    int src_skip;

    stride /= sizeof(float);
    src += stride * (height-1);
    src_skip = -stride*2;

    while(height-->0){
        int x;
        for(x=width;x>0;--x){
            int r = round(src[0]*255);
            int g = round(src[1]*255);
            int b = round(src[2]*255);
            *dst++ = r<<16|g<<8|b<<0;
            src += 3;
        }
        src += src_skip;
        dst += dst_skip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static cairo_surface_t *
_surface_from_bitmap(FIBITMAP *bm, int width, int height, int flags)
{
    uint32_t palette[256];
    const uint8_t *src = FreeImage_GetBits(bm);
    int stride = FreeImage_GetPitch(bm);
    switch(FreeImage_GetImageType(bm)){
    case FIT_UNKNOWN:
        DEBUG_p("type unknown");
        return NULL;
    case FIT_BITMAP:
        switch(FreeImage_GetBPP(bm)){
        case 1:
            DEBUG_p("type 1bit bitmap");
            return _surface_from_bitonal(src,width,height,stride);
        case 4:
            DEBUG_p("type 4bit bitmap");
            _palette_from_bitmap(palette,bm);
            return _surface_from_4bit(src,width,height,stride,palette);
        case 8:
            DEBUG_p("type 8bit bitmap");
            if(_palette_from_bitmap(palette,bm)){
                return _surface_from_8bit_transparent(src,width,height,stride,palette);
            }else{
                return nil_cairo_surface_create_from_indexed(src,
                                                             width,height,stride,
                                                             (uint8_t*)palette,256,
                                                             4,NIL_CAIRO_FORMAT_RGB24,
                                                             flags);
            }
        case 16:
            DEBUG_p("type 16bit bitmap");
            return NULL;
        case 24:
            DEBUG_p("type 24bit bitmap");
            return nil_cairo_surface_create_from_planar((uint8_t*)FreeImage_GetBits(bm),
                                                        width,height,stride,
                                                        3,NILE_FORMAT_b8g8r8,
                                                        flags);
        case 32:
            DEBUG_p("type 32bit bitmap");
            return nil_cairo_surface_create_from_planar((uint8_t*)FreeImage_GetBits(bm),
                                                        width,height,
                                                        stride,
                                                        4,NILE_FORMAT_b8g8r8a8,
                                                        flags);
        default:
            DEBUG_p("unhandled bitmap type");
            return NULL;
        }
    case FIT_UINT16:
        DEBUG_p("type uin16");
        return NULL;
    case FIT_INT16:
        DEBUG_p("type sint16");
        return NULL;
    case FIT_UINT32:
        DEBUG_p("type uint32");
        return NULL;
    case FIT_INT32:
        DEBUG_p("type sint32");
        return NULL;
    case FIT_FLOAT:
        DEBUG_p("type float");
        return NULL;
    case FIT_DOUBLE:
        DEBUG_p("type double");
        return NULL;
    case FIT_COMPLEX:
        DEBUG_p("type complex");
        return NULL;
    case FIT_RGB16:
        DEBUG_p("type RGB16");
        return NULL;
    case FIT_RGBA16:
        DEBUG_p("type RGBA16");
        return NULL;
    case FIT_RGBF:
        DEBUG_p("type RGBF");
        return _surface_from_rgbf((float*)FreeImage_GetBits(bm),width,height,stride);
    case FIT_RGBAF:
        DEBUG_p("type RGBAF");
        return NULL;
    default:
        DEBUG_p("invalid type");
        return NULL;
    }
}


static int
_fi_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    int res = -1;
    FIMEMORY *stream = FreeImage_OpenMemory((BYTE*)data, len);
    if(h->multi_type){
        FIMULTIBITMAP *mbm = FreeImage_LoadMultiBitmapFromMemory(h->fif, stream, 0);
        int i;
        if(!mbm)
            goto out;
        h->multi_count = FreeImage_GetPageCount(mbm);
        DEBUG_p("multi count: %d",h->multi_count);
        switch(h->multi_type){
        case MT_NONE:
            /* not reached */
        case MT_PAGE:
            h->info->page_count = h->multi_count;
            break;
        case MT_VARIANT:
            h->info->variant_count = h->multi_count;
            break;
        case MT_FRAME:
            h->info->frame_count = h->multi_count;
            break;
        }
        h->multi = (multi_t*)malloc(h->multi_count * sizeof(multi_t));
        for(i=0;i<h->multi_count;++i){
            FIBITMAP *bm = FreeImage_LockPage(mbm, i);
            int width;
            int height;
            if(!bm)
                continue;
            width = FreeImage_GetWidth(bm);
            height = FreeImage_GetHeight(bm);
            if(width > h->info->width)
                h->info->width = width;
            if(height > h->info->height)
                h->info->height = height;
            h->multi[i].surface = _surface_from_bitmap(bm, width, height, h->flags);
            if(h->multi[i].surface)
                res = 0;
            switch(h->multi_type){
            case MT_NONE:
                /* not reached */
            case MT_PAGE:
            case MT_VARIANT:
                h->multi[i].size.width = width;
                h->multi[i].size.height = height;
                break;
            case MT_FRAME:
                {
                    FITAG *tag;
                    FreeImage_GetMetadata(FIMD_ANIMATION, bm, "FrameTime", &tag);
                    if(tag){
                        const DWORD *ms;
                        ms = (const DWORD*)FreeImage_GetTagValue(tag);
                        h->multi[i].duration = (double)*ms / 1000;
                    }else{
                        h->multi[i].duration = -1;
                    }
                }
                break;
            }
            FreeImage_UnlockPage(mbm, bm, FALSE);
        }
        FreeImage_CloseMultiBitmap(mbm, 0);
    }else{
        FIBITMAP *bm = FreeImage_LoadFromMemory(h->fif, stream, 0);
        if(!bm)
            goto out;
        h->multi = (multi_t*)malloc(sizeof(multi_t));
        h->info->width = FreeImage_GetWidth(bm);
        h->info->height = FreeImage_GetHeight(bm);
        h->multi[0].surface = _surface_from_bitmap(bm, h->info->width, h->info->height, h->flags);
        if(h->multi[0].surface)
            res = 0;
        FreeImage_Unload(bm);
    }

out:
    FreeImage_CloseMemory(stream);
    return res;
}


static cairo_surface_t *
_fi_get_image_surface(abydos_plugin_handle_t *h,
                      int page,
                      cairo_bool_t *layer_visibility,
                      int variant,
                      double time,
                      int frame)
{
    int index;
    switch(h->multi_type){
    case MT_NONE:
        index = 0;
        break;
    case MT_PAGE:
        index = page;
        break;
    case MT_VARIANT:
        index = variant;
        break;
    case MT_FRAME:
        index = frame;
        break;
    }
    return cairo_surface_reference(h->multi[index].surface);
}


static void
_fi_get_variant_size(abydos_plugin_handle_t *h, int page, abydos_size_t *size)
{
    int i;
    DEBUG_p("begin");
    for(i=0;i<h->info->variant_count;++i){
        size[i] = h->multi[i].size;
        DEBUG_p("size of variant %d is %dx%d",i,size[i].width,size[i].height);
    }
}


static void
_fi_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int i;
    DEBUG_p("begin");
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->multi[i].duration;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .done = _fi_done,
        .get_mime_types = _fi_get_mime_types,
        .new = _fi_new,
        .free = _fi_free,
        .create_from_data = _fi_create_from_data,
        .get_image_surface = _fi_get_image_surface,
        .get_variant_size = _fi_get_variant_size,
        .get_duration = _fi_get_duration
    };
    int count;
    int i;
    FreeImage_Initialise(FALSE);
    count = FreeImage_GetFIFCount();
    DEBUG_p("%d formats available",count);
    for(i=0;i<count;++i){
        FREE_IMAGE_FORMAT fif = (FREE_IMAGE_FORMAT)i;
        DEBUG_p("%d: %s: %s",i,FreeImage_GetFIFDescription(fif),FreeImage_GetFIFMimeType(fif));
    }
    return &p;
}
