/*
 * Copyright (C) 2019 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <icns.h>
#include <nil-cairo.h>
#include <stdlib.h>                     /* malloc(), free() */


#define MAX_VARIANTS 64


typedef struct{
    int width;
    int height;
    cairo_surface_t *surface;
}variant_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    variant_t *variant;
};


static const char **
_icns_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-icns",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_icns_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->variant = NULL;
    return h;
}


static void
_icns_free(abydos_plugin_handle_t *h)
{
    if(h->variant){
        int i;
        for(i=0;i<h->info->variant_count;++i)
            cairo_surface_destroy(h->variant[i].surface);
        free(h->variant);
    }
    free(h);
}


static cairo_surface_t *
_surface_from_image(icns_image_t *image)
{
    return nil_cairo_surface_create_from_planar(image->imageData,image->imageWidth,image->imageHeight,-1,4,NILE_FORMAT_r8g8b8a8,0);
    //return abydos_image_surface_create_from_pixels(ABYDOS_PIXEL_FORMAT_r8g8b8a8,0,image->imageData,image->imageWidth,image->imageHeight,-1,4,NULL);
}


static void
_variant_from_image(variant_t *variant,icns_image_t *image)
{
    variant->width = image->imageWidth;
    variant->height = image->imageHeight;
    variant->surface = _surface_from_image(image);
}


int _variants_from_family(abydos_plugin_handle_t *h,icns_family_t *family) {
    int error = ICNS_STATUS_OK;
    icns_byte_t *p = (icns_byte_t*)family;
    unsigned long offs = sizeof(icns_type_t) + sizeof(icns_size_t);
    variant_t variant[MAX_VARIANTS];

    h->info->variant_count = 0;

    while(((offs+8) < family->resourceSize) && (error == 0 || error == ICNS_STATUS_UNSUPPORTED) && h->info->variant_count<MAX_VARIANTS){
        icns_element_t element;
        icns_icon_info_t info;
        icns_size_t iconDataSize;
        char type_str[5];

        memcpy(&element,(p+offs),8);
        icns_type_str(element.elementType,type_str);
        iconDataSize = element.elementSize - 8;

        DEBUG_p("type: '%s'",type_str);

        switch(element.elementType) {
        case ICNS_TABLE_OF_CONTENTS:
            DEBUG_p("table of contents\n");
            break;
        case ICNS_ICON_VERSION:
#ifdef USE_DEBUG
            {
                icns_byte_t iconBytes[4];
                icns_uint32_t iconVersion = 0;
                float iconVersionNumber = 0;
                if(iconDataSize == 4) {
                    memcpy(&iconBytes[0],(p+offs+8),4);
                    iconVersion = iconBytes[3]|iconBytes[2]<<8|iconBytes[1]<<16|iconBytes[0]<<24;
                    iconVersionNumber = *((float *)(&iconVersion));
                }
                printf(" value: %f\n",iconVersionNumber);
            }
#endif
            break;
        case ICNS_TILE_VARIANT:
        case ICNS_ROLLOVER_VARIANT:
        case ICNS_DROP_VARIANT:
        case ICNS_OPEN_VARIANT:
        case ICNS_OPEN_DROP_VARIANT:
            {
                /* Load up the variant, over-write the type and size to make it into an icns. */
                icns_byte_t	*subdata = (icns_byte_t*)malloc(element.elementSize);
                icns_family_t *subfamily = NULL;
                icns_byte_t	b[4] = {0,0,0,0};
                memcpy(subdata,p+offs,element.elementSize);
                memcpy(subdata,"icns",4);
                b[0] = element.elementSize >> 24;
                b[1] = element.elementSize >> 16;
                b[2] = element.elementSize >> 8;
                b[3] = element.elementSize;
                memcpy(subdata+4, b, sizeof(icns_size_t));

                /* Display some info about the variant. */
                switch(element.elementType) {
                case ICNS_TILE_VARIANT:
                    DEBUG_p("icon variant: tile (%d bytes)\n",iconDataSize);
                    break;
                case ICNS_ROLLOVER_VARIANT:
                    DEBUG_p("icon variant: rollover (%d bytes)\n",iconDataSize);
                    break;
                case ICNS_DROP_VARIANT:
                    DEBUG_p("icon variant: drop (%d bytes)\n",iconDataSize);
                    break;
                case ICNS_OPEN_VARIANT:
                    DEBUG_p("icon variant: open (%d bytes)\n",iconDataSize);
                    break;
                case ICNS_OPEN_DROP_VARIANT:
                    DEBUG_p(" icon variant: open/drop (%d bytes)\n",iconDataSize);
                    break;
                }

                /* Try to parse out the variant. */
                error = icns_import_family_data(element.elementSize,subdata,&subfamily);

                if(error) {
                    DEBUG_p("unable to read icon variant type '%s' (error while parsing)",type_str);
                }else{
                    error = _variants_from_family(h,subfamily);
                }

                free(subfamily);

                if(subdata) {
                    free(subdata);
                    subdata = NULL;
                }
            }
            break;
        default:
            info = icns_get_image_info_for_type(element.elementType);
#ifdef USE_DEBUG
            DEBUG_p(" %dx%d",info.iconWidth,info.iconHeight);
            DEBUG_p(" %d-bit",info.iconBitDepth);
            if(info.isImage)
                DEBUG_p(" icon");
            if(info.isImage && info.isMask)
                DEBUG_p(" with");
            if(info.isMask)
                DEBUG_p(" mask");
            if((element.elementSize-8) < info.iconRawDataSize) {
                DEBUG_p(" (%d bytes compressed to %d)",(int)info.iconRawDataSize,iconDataSize);
            } else {
                DEBUG_p(" (%d bytes)",iconDataSize);
            }
            DEBUG_p("\n");
#endif
            if(info.isImage){
                icns_image_t image;
                memset(&image,0,sizeof(icns_image_t));
                error = icns_get_image32_with_mask_from_family(family,element.elementType,&image);
                switch(error){
                case ICNS_STATUS_OK:
                    DEBUG_p("storing image");
                    _variant_from_image(&variant[h->info->variant_count++],&image);
                    icns_free_image(&image);
                    break;
                case ICNS_STATUS_UNSUPPORTED:
                    /* just ignore this */
                    DEBUG_p("unable to convert '%s' element! (unsupported by this version of libicns)",type_str);
                    error = ICNS_STATUS_OK;
                    break;
                default:
                    DEBUG_p("unable to load 32-bit icon image with mask from icon family");
                    break;
                }
            }
            break;
        }
        offs += element.elementSize;
    }

    if(h->info->variant_count>0){
        int i;
        DEBUG_p("%d variants found",h->info->variant_count);
        h->variant = (variant_t*)malloc(h->info->variant_count*sizeof(variant_t));
        memcpy(h->variant,variant,h->info->variant_count*sizeof(variant_t));
        for(i=0;i<h->info->variant_count;++i){
            if(variant[i].width>h->info->width)
                h->info->width = variant[i].width;
            if(variant[i].height>h->info->height)
                h->info->height = variant[i].height;
        }
    }else{
        DEBUG_p("no images found");
    }

    return error;
}


static int
_icns_create_from_stream(abydos_plugin_handle_t *h,FILE *f)
{
    icns_family_t *family = NULL;
    int error = icns_read_family_from_file(f,&family);
    if(error != ICNS_STATUS_OK)
        goto out;
    error = _variants_from_family(h,family);
    free(family);

out:
    return ICNS_STATUS_OK == error ? 0 : -1;
}


static int
_icns_create_from_data(abydos_plugin_handle_t *h,const char *bytes, size_t len)
{
    FILE *f;
    int res;
    f = fmemopen((void*)bytes,len,"r");
    if(!f)
        return -1;
    res = _icns_create_from_stream(h,f);
    fclose(f);
    return res;
}


static int
_icns_create_from_file(abydos_plugin_handle_t *h,const char *filename)
{
    FILE *f;
    int res;
    f = fopen(filename,"r");
    if(!f)
        return -1;
    res = _icns_create_from_stream(h,f);
    fclose(f);
    return res;
}


static cairo_surface_t *
_icns_get_image_surface(abydos_plugin_handle_t *h,
                        int page,
                        cairo_bool_t *layer_visibility,
                        int variant,
                        double time,
                        int frame)
{
    return cairo_surface_reference(h->variant[variant].surface);
}


static void
_icns_get_variant_size(abydos_plugin_handle_t *h, int page, abydos_size_t *size)
{
    int i;
    for(i=0;i<h->info->variant_count;++i){
        size[i].width = h->variant[i].width;
        size[i].height = h->variant[i].height;
    }
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _icns_get_mime_types,
        .new = _icns_new,
        .free = _icns_free,
        .create_from_data = _icns_create_from_data,
        .create_from_file = _icns_create_from_file,
        .get_image_surface = _icns_get_image_surface,
        .get_variant_size = _icns_get_variant_size
    };
    return &p;
}
