/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <byteswap.h>
#include <debug.h>
#include <endian.h>
#include <nil-cairo.h>
#include <ni-ilbm.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc() */
#include <string.h>                     /* memset() */


typedef struct{
    cairo_surface_t *surface;
    double duration;
}frame_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    frame_t *frame;
};


static const char **
_ilbm_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-ilbm",
        "video/x-anim",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_ilbm_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->frame = NULL;
    info->threadsafe = 1;
    return h;
}


static void
_ilbm_free(abydos_plugin_handle_t *h)
{
    if(h->frame){
        int i;
        for(i=0;i<h->info->frame_count;++i)
            if(h->frame[i].surface)
                cairo_surface_destroy(h->frame[i].surface);
        free(h->frame);
    }
    free(h);
}


typedef struct{
    abydos_plugin_handle_t *h;
    nile_t *ni;
    int actual_frame;
}load_data_t;


static int
_on_request(void *userdata, int width, int height, double pixel_ratio, int frame_count, nile_t **r_ni)
{
    load_data_t *d = userdata;
    int i;
    d->h->info->width = width;
    d->h->info->height = height;
    d->h->info->pixel_ratio = pixel_ratio;
    d->h->info->frame_count = frame_count;
    d->h->frame = malloc(frame_count * sizeof(frame_t));
    d->ni = malloc(frame_count * sizeof(nile_t));
    for(i=0;i<frame_count;++i){
        d->h->frame[i].surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24, width, height);
        nile_init(&d->ni[i]);
        nil_cairo_from_surface(&d->ni[i], d->h->frame[i].surface);
    }
    *r_ni = d->ni;
    return 0;
}


static int
_on_frame(void *userdata, int frame, double duration)
{
    load_data_t *d = userdata;
    nile_t tmp;
    if(d->h->info->frame_count > 1 && duration<=0){
        DEBUG_p("frame %d has zero duration", frame);
        --d->h->info->frame_count;
        cairo_surface_destroy(d->h->frame[frame].surface);
        return 0;
    }
    d->h->frame[frame].duration = duration;
    switch(d->ni[frame].type){
    case NILE_TYPE_NONE:
        /* shouldn't happen */
        break;
    case NILE_TYPE_BITPLANES:
        nile_init(&tmp);
        nil_cairo_from_surface(&tmp, d->h->frame[frame].surface);
        nile_copy_planar_from_bitplanes(&tmp, &d->ni[frame]);
        nile_done(&tmp);
        break;
    case NILE_TYPE_INDEXED:
        nile_init(&tmp);
        nil_cairo_from_surface(&tmp, d->h->frame[frame].surface);
        nile_copy_planar_from_indexed(&tmp, &d->ni[frame]);
        nile_done(&tmp);
        break;
    case NILE_TYPE_PLANAR:
        /* The pixels are already copied to the surface. */
        break;
    }
    cairo_surface_mark_dirty(d->h->frame[frame].surface);
    if(frame != d->actual_frame)
        memcpy(&d->h->frame[d->actual_frame], &d->h->frame[frame], sizeof(frame_t));
    ++d->actual_frame;
    return 0;
}


static int
_ilbm_create_from_data(abydos_plugin_handle_t *h, const char *data, size_t len)
{
    load_data_t d;
    int ret;
    int i;
    d.h = h;
    d.ni = NULL;
    d.actual_frame = 0;
    ret = ni_ilbm_decode((uint8_t*)data, len, 4, NIL_CAIRO_FORMAT_RGB24, _on_request, _on_frame, &d);
    if(d.ni){
        for(i=0;i<h->info->frame_count;++i)
            nile_done(&d.ni[i]);
        free(d.ni);
    }
    return ret;
}


static cairo_surface_t *
_ilbm_get_image_surface(abydos_plugin_handle_t *h, int page, cairo_bool_t *layer_visibility, int variant, double time, int frame)
{
    return cairo_surface_reference(h->frame[frame].surface);
}


static void
_ilbm_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->frame[i].duration;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _ilbm_get_mime_types,
        .new = _ilbm_new,
        .free = _ilbm_free,
        .create_from_data = _ilbm_create_from_data,
        .get_image_surface = _ilbm_get_image_surface,
        .get_duration = _ilbm_get_duration
    };
    return &p;
}
