/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <stdlib.h>

#include "decode.h"

typedef struct{
    cairo_surface_t *surface;
}frame_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    frame_t *frame;
};


static const char **
_jpegxl_get_mime_types()
{
    static const char *mime_types[] = {
        "image/jxl",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_jpegxl_new(const char *mime_type, abydos_plugin_info_t *info){
    abydos_plugin_handle_t *h = malloc(sizeof(abydos_plugin_handle_t));
    info->threadsafe = 1;
    h->info = info;
    h->frame = NULL;

    return h;
}


static void
_jpegxl_free(abydos_plugin_handle_t *h){
    if(h->frame){
        int i;
        for(i=0;i<h->info->frame_count;++i)
            if(h->frame[i].surface)
                cairo_surface_destroy(h->frame[i].surface);
        free(h->frame);
    }
    free(h);
}


static int
_jpegxl_create_from_data(abydos_plugin_handle_t *h, const char *sdata, size_t len)
{
    JxlDecoder *dec = JxlDecoderCreate(NULL);
    JxlDecoderSubscribeEvents(dec, JXL_DEC_BASIC_INFO | JXL_DEC_FULL_IMAGE);
    JxlDecoderSetInput(dec, (uint8_t*)sdata, len);
    h->frame = NULL;
    h->info->frame_count = 0;
    for(;;){
        JxlBasicInfo info;
        switch(JxlDecoderProcessInput(dec)){
        case JXL_DEC_ERROR:
            DEBUG_p("error");
            goto done;
        case JXL_DEC_NEED_MORE_INPUT:
            DEBUG_p("need more data");
            goto done;
        case JXL_DEC_BASIC_INFO:
            if(JXL_DEC_SUCCESS != JxlDecoderGetBasicInfo(dec, &info))
                goto done;
            DEBUG_p("size: %dx%d", info.xsize, info.ysize);
            h->info->width = info.xsize;
            h->info->height = info.ysize;
            break;
        case JXL_DEC_FULL_IMAGE:
            DEBUG_p("full image");
            if(h->frame){
                h->frame = realloc(h->frame, sizeof(frame_t) * (h->info->frame_count + 1));
            }else{
                h->frame = malloc(sizeof(frame_t));
            }
            h->frame[h->info->frame_count++].surface = JxlDecoderImageSurface(dec);
            break;
        case JXL_DEC_SUCCESS:
            DEBUG_p("success");
            goto done;
        }
    }
done:
    JxlDecoderReleaseInput(dec);
    JxlDecoderDestroy(dec);
    return h->frame ? 0 : -1;
}


static cairo_surface_t *
_jpegxl_get_image_surface(abydos_plugin_handle_t *h,
                          int page,
                          cairo_bool_t *layer_visibility,
                          int variant,
                          double time,
                          int frame)
{
    return cairo_surface_reference(h->frame[frame].surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _jpegxl_get_mime_types,
        .new = _jpegxl_new,
        .free = _jpegxl_free,
        .create_from_data = _jpegxl_create_from_data,
        .get_image_surface = _jpegxl_get_image_surface
    };
    return &p;
}
