/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include "image_bundle.h"
#include "jxl-cairo.h"

namespace jxl{

#define MIN(a,b) ((a)<(b) ? (a) : (b))
#define MAX(a,b) ((a)>(b) ? (a) : (b))
#define CLAMP(x,lo,hi) MIN(hi, MAX(lo,x))


static cairo_surface_t *
_surface_from_image_rgb(const jxl::Image3F &img){
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24, img.xsize(), img.ysize());
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_skip = cairo_image_surface_get_stride(surface) / sizeof(uint32_t) - img.xsize();
    unsigned int y;
    for(y=0;y<img.ysize();++y){
        unsigned int x;
        const float *row[3];
        row[0] = img.PlaneRow(0, y);
        row[1] = img.PlaneRow(1, y);
        row[2] = img.PlaneRow(2, y);
        for(x=0;x<img.xsize();++x){
            int r = row[2][x];
            int g = row[1][x];
            int b = row[0][x];
            *dst++ = r << 0 | g << 8 | b << 16;
        }
        dst += dst_skip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static cairo_surface_t *
_surface_from_image_rgba(const jxl::Image3F &img, const jxl::ImageF &alpha){
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, img.xsize(), img.ysize());
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_skip = cairo_image_surface_get_stride(surface) / sizeof(uint32_t) - img.xsize();
    unsigned int y;
    for(y=0;y<img.ysize();++y){
        unsigned int x;
        const float *row[4];
        row[0] = img.PlaneRow(0, y);
        row[1] = img.PlaneRow(1, y);
        row[2] = img.PlaneRow(2, y);
        row[3] = alpha.Row(y);
        for(x=0;x<img.xsize();++x){
            int a = CLAMP(ceil(row[3][x] * 255), 0, 255);
            int r = a * CLAMP(ceil(row[2][x] * 255), 0, 255) / 255;
            int g = a * CLAMP(ceil(row[1][x] * 255), 0, 255) / 255;
            int b = a * CLAMP(ceil(row[0][x] * 255), 0, 255) / 255;
            *dst++ = r << 0 | g << 8 | b << 16 | a << 24;
        }
        dst += dst_skip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


cairo_surface_t *surface_from_image_bundle(const ImageBundle &ib){
    return ib.HasAlpha() ? _surface_from_image_rgba(ib.color(), ib.alpha())
                         : _surface_from_image_rgb(ib.color());
}

} /* namespace jxl */
