/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <nil-cairo.h>
#include <stdlib.h>
#include <string.h>

#include "lodepng.h"


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
_lpng_get_mime_types()
{
    static const char *mime_types[] = {
        "image/png",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_lpng_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    info->threadsafe = 1;
    h->info = info;
    h->surface = NULL;
    return h;
}


static void
_lpng_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


static int
_lpng_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    unsigned int error;
    uint8_t *pixels;
    unsigned int width;
    unsigned int height;

    error = lodepng_decode32(&pixels, &width, &height, (unsigned char*)data, len);
    if(error){
        DEBUG_p("error %u: %s", error, lodepng_error_text(error));
        return -1;
    }

    h->info->width = width;
    h->info->height = height;
    h->surface = nil_cairo_surface_create_from_planar(pixels,
                                                      h->info->width,
                                                      h->info->height,
                                                      -1,
                                                      4,
                                                      NILE_FORMAT_r8g8b8a8,
                                                      0);

    free(pixels);
    return 0;
}


static cairo_surface_t *
_lpng_get_image_surface(abydos_plugin_handle_t *h,
                        int page,
                        cairo_bool_t *layer_visibility,
                        int variant,
                        double time,
                        int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _lpng_get_mime_types,
        .new = _lpng_new,
        .free = _lpng_free,
        .create_from_data = _lpng_create_from_data,
        .get_image_surface = _lpng_get_image_surface
    };
    return &p;
}
