/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <rlottie_capi.h>
#include <stdlib.h>
#include <string.h>

#include "lottiecairo.h"


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    Lottie_Animation *anim;
};


static const char **
_lottie_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-lottie+json", "application/json",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_lottie_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->scalable_size = 1;
    h->info->threadsafe = 1;
    return h;
}


static void
_lottie_free(abydos_plugin_handle_t *h)
{
    if(h->anim)
        lottie_animation_destroy(h->anim);
    free(h);
}


static int
_lottie_create_from_file(abydos_plugin_handle_t *h, const char *filename)
{
    size_t width;
    size_t height;
    h->anim = lottie_animation_from_file(filename);
    if(!h->anim)
        return -1;
    lottie_animation_get_size(h->anim,&width,&height);
    h->info->width = width;
    h->info->height = height;
    h->info->frame_count = lottie_animation_get_totalframe(h->anim);
    return 0;
}


static void
_lottie_render(abydos_plugin_handle_t *h,
               cairo_t *cr,
               int page,
               cairo_bool_t *layer_visibility,
               int variant,
               double time,
               int frame)
{
    const LOTLayerNode* tree = lottie_animation_render_tree(h->anim,frame,h->info->width,h->info->height);
    lottie_cairo_render(tree,cr);
}


static void
_lottie_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int i;
    double d = (double)lottie_animation_get_framerate(h->anim)/1000;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = d;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _lottie_get_mime_types,
        .new = _lottie_new,
        .free = _lottie_free,
        .create_from_file = _lottie_create_from_file,
        .render = _lottie_render,
        .get_duration = _lottie_get_duration
    };
    return &p;
}
