/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <nile.h>
#include <nil-cairo.h>
#include <nil-cga.h>
#include <nil-interlace.h>
#include <nil-mcga.h>
#include <nil-pctext.h>
#include <stdlib.h>
#include <string.h>


typedef struct{
    cairo_surface_t *surface;
}page_t;

struct _abydos_plugin_handle_t{
    enum{
        T_SCRIPT,
        T_GRAPHIC,
        T_SCREEN,
        T_SPRITE
    }type;
    abydos_plugin_info_t *info;
    page_t *page;
};


static const char **
_datashow_get_mime_types()
{
    static const char *mime_types[] = {
        //"application/x-datashow-script",
        "image/x-datashow-graphic",
        "image/x-datashow-screen",
        "image/x-datashow-sprite",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_datashow_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    if(!strcasecmp(mime_type, "application/x-datashow-script")){
        h->type = T_SCRIPT;
    }else if(!strcasecmp(mime_type, "image/x-datashow-graphic")){
        h->type = T_GRAPHIC;
    }else if(!strcasecmp(mime_type, "image/x-datashow-screen")){
        h->type = T_SCREEN;
    }else if(!strcasecmp(mime_type, "image/x-datashow-sprite")){
        h->type = T_SPRITE;
    }else{
        /* not reached */
        free(h);
        return NULL;
    }
    info->threadsafe = 1;
    h->info = info;
    h->page = NULL;
    return h;
}


static void
_datashow_free(abydos_plugin_handle_t *h)
{
    if(h->page){
        int i;
        for(i=0;i<h->info->page_count;++i)
            if(h->page[i].surface)
                cairo_surface_destroy(h->page[i].surface);
        free(h->page);
    }
    free(h);
}


#define get8(a) (*(uint8_t*)(a))
#define get16(a) le16toh(*(uint16_t*)(a))


static void
_copy_pixels(nile_t *dst, uint8_t *src)
{
    int stride = nile_get_stride(dst);
    int x;
    for(x=0;x<dst->width;++x){
        int y;
        for(y=0;y<dst->height;++y)
            dst->pixels[x + y * stride] = *src++;
    }
}


#if 0
static int
_handle_script(abydos_plugin_handle_t *h, const uint8_t *data, size_t len)
{
    return -1;
}
#endif


static void
_swap(uint8_t **a, uint8_t **b){
    void *tmp = *a;
    *a = *b;
    *b = tmp;
}


static int
_handle_graphic(abydos_plugin_handle_t *h, const uint8_t *data, size_t len)
{
    const uint8_t *end = data + len;
    nile_t ni;

    if(len < 6)
        return -1;

    h->info->width = get16(data) + 1;
    h->info->height = get16(data+2) + 1;
    data += 4;

    if(data + h->info->width * h->info->height / 2 > end)
        return -1;

    nile_init(&ni);

    nil_interlace_set(&ni, data, h->info->width, h->info->height, h->info->width/8, 4);
    _swap(&ni.bitplanes[0], &ni.bitplanes[3]);
    _swap(&ni.bitplanes[1], &ni.bitplanes[2]);

    ni.palette = (uint8_t*)nil_cga_palette16;
    ni.colors = 16;
    ni.bpc = 3;
    ni.format = NILE_FORMAT_r8g8b8;

    h->page = malloc(sizeof(page_t));
    h->page[0].surface = nil_cairo_surface_create(&ni);
    h->info->pixel_ratio = 5.0/12;

    nile_done(&ni);
    return 0;
}


typedef struct{
    const uint8_t *data;
    const uint8_t *end;
    int copy;
    int run;
    int value;
}rle_t;


static void
_rle_init(rle_t *rle, const uint8_t *data, size_t len, int copy)
{
    rle->data = data;
    rle->end = data + len;
    rle->copy = copy;
    rle->run = 0;
}


static int
_rle_get(rle_t *rle)
{
    while(rle->copy <= 0 && rle->run <= 0){
        if(rle->data + 5 > rle->end)
            return -1;
        rle->copy = get16(rle->data);
        rle->data += 2;
        rle->value = 255 - get8(rle->data++);
        rle->run = get16(rle->data);
        rle->data += 2;
    }
    if(rle->run > 0){
        --rle->run;
        return rle->value;
    }else /* if rle->copy > 0 */ {
        if(rle->data + 1 > rle->end)
            return -1;
        --rle->copy;
        return 255 - get8(rle->data++);
    }
}


static int
_handle_screen(abydos_plugin_handle_t *h, const uint8_t *data, size_t len)
{
    const uint8_t *end = data + len;
    uint32_t palette[256];
    nile_t ni;
    int size;
    int rows;
    int cols;
    rle_t text_rle;
    rle_t attr_rle;
    int y;
    if(len < 6)
        return -1;
    size = get16(data);
    if(len < size)
        return -1;

    rows = get8(data+4);
    cols = get8(data+5);
    size = get16(data+6) -1;
    data += 8;

    {
        int text_start = get16(data);
        int attr_start = get16(data+2);
        data += 4;
        _rle_init(&text_rle, data, size, text_start);
        _rle_init(&attr_rle, data + size, end - (data + size), attr_start);
    }

    h->page = malloc(sizeof(page_t));
    h->page[0].surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24, cols*8, rows*16);
    nil_cga_copy_palette_colors((uint8_t*)palette, 4, NIL_CAIRO_FORMAT_RGB24, 16);
    nile_init(&ni);
    nil_cairo_from_surface(&ni, h->page[0].surface);
    for(y=0;y<rows;++y){
        int x;
        for(x=0;x<cols;++x){
            int ch = _rle_get(&text_rle);
            int color = _rle_get(&attr_rle);
            nil_pctext_draw_character(&ni, x*8, y*16, ch, palette[color >> 4], palette[color & 15], 8);
        }
    }
    nile_done(&ni);
    cairo_surface_mark_dirty(h->page[0].surface);
    h->info->pixel_ratio = 5.0/6;
    return 0;
}


static int
_handle_sprite(abydos_plugin_handle_t *h, const uint8_t *data, size_t len)
{
    const uint8_t *end = data + len;
    int garbage;
    nile_t ni;
    int i;

    if(data + 1 > end)
        return -1;
    /* skip pascal string */
    data += 1 + get8(data);

    if(data + 4 > end)
        return -1;
    h->info->page_count = get8(data);
    h->info->width = get8(data+2) + 1;
    h->info->height = get8(data+3) + 1;
    data += 4;

    garbage = (len-16) / h->info->page_count - h->info->width * h->info->height;
    DEBUG_p("%d bytes, %d pixels per frame, %d frames, %d garbage", (int)len, h->info->width * h->info->height, h->info->page_count, garbage);

    if(data + (h->info->width * h->info->height + garbage) * h->info->frame_count - garbage > end){
        DEBUG_p("data short");
        return -1;
    }

    nile_init(&ni);
    nile_ensure_indexed(&ni, h->info->width, h->info->height);

    nile_ensure_palette(&ni, 256, 4);
    ni.format = NILE_FORMAT_r8g8b8a8;
    nil_mcga_copy_palette_colors(ni.palette, ni.bpc, ni.format, ni.colors);
    /* Make the first color transparent */
    ni.palette[3] = 0;

    h->page = malloc(h->info->frame_count * sizeof(page_t));
    for(i=0;i<h->info->page_count;++i){
        _copy_pixels(&ni, (uint8_t*)data);
        h->page[i].surface = nil_cairo_surface_create(&ni);
        data += h->info->width * h->info->height + garbage;
    }

    nile_done(&ni);
    h->info->pixel_ratio = 5.0/6;
    return 0;
}


static int
_datashow_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    switch(h->type){
    case T_SCRIPT:
        return -1;
    case T_GRAPHIC:
        return _handle_graphic(h, (uint8_t*)data, len);
    case T_SCREEN:
        return _handle_screen(h, (uint8_t*)data, len);
    case T_SPRITE:
        return _handle_sprite(h, (uint8_t*)data, len);
    }
    /* not reached */
    return -1;
}


static cairo_surface_t *
_datashow_get_image_surface(abydos_plugin_handle_t *h,
                            int page,
                            cairo_bool_t *layer_visibility,
                            int variant,
                            double time,
                            int frame)
{
    return cairo_surface_reference(h->page[page].surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _datashow_get_mime_types,
        .new = _datashow_new,
        .free = _datashow_free,
        .create_from_data = _datashow_create_from_data,
        .get_image_surface = _datashow_get_image_surface
    };
    return &p;
}
