/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <nil-cairo.h>
#include <nil-pctext.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

/*

Header:

-------v------v--------------
offset | size | Description
-------+------+--------------
     0 |    3 | ID = "GX2"
     3 |    1 | version = 1
     4 |    2 | header size
     6 |    1 | bits per pixel
     7 |    2 | width
     9 |    2 | height
    11 |    2 | horizontal size
    13 |    2 | vertical size
    15 |    1 | unit

*/

typedef struct{
    const uint8_t *p;
    const uint8_t *end;
    int bytes;
    int runlength;
}rle_t;

static void
_rle_init(rle_t *rle, const uint8_t *data, size_t len)
{
    rle->p = data;
    rle->end = data+len;
    rle->bytes = 0;
}

static int
_rle_get(rle_t *rle)
{
    if(rle->bytes <= 0){
        if(rle->p >= rle->end){
            DEBUG_p("out of data");
            return -1;
        }
        rle->bytes = *rle->p++;
        if(0==rle->bytes)
            DEBUG_p("bytes should not be zero");
        if(rle->bytes >= 128){
            if(rle->p >= rle->end){
                DEBUG_p("unexcpected end of data");
                return -1;
            }
            rle->runlength = *rle->p++;
            rle->bytes &= 127;
            if(rle->bytes < 4)
                DEBUG_p("too few bytes");
        }else{
            rle->runlength = -1;
        }
    }
    --rle->bytes;
    if(rle->runlength >= 0){
        return rle->runlength;
    }else{
        if(rle->p >= rle->end){
            DEBUG_p("unexcpected end of data");
            return -1;
        }
        return *rle->p++;
    }
}


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    enum{
        T_GX1,
        T_GX2
    }type;
    cairo_surface_t *surface;
};


static const char **
_gxx_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-gx1",
        "image/x-gx2",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_gxx_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    if(!strcasecmp(mime_type,"image/x-gx1")){
        h->type = T_GX1;
    }else{
        h->type = T_GX2;
    }
    h->surface = NULL;
    h->info->threadsafe = 1;
    return h;
}


static void
_gxx_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


#define get16(a) le16toh(*(uint16_t*)(a))
#define get32(a) le32toh(*(uint32_t*)(a))


static ssize_t
_handle_palette_2bit(nile_t *ni, const uint8_t *data, size_t len)
{
    uint8_t *dst;
    int i;
    ni->colors = 1<<ni->depth;
    if(len < ni->colors)
        return -1;
    ni->bpc = 3;
    ni->format = NILE_FORMAT_r8g8b8;
    dst = ni->palette;
    for(i=0;i<ni->colors;++i){
        int c = *data++;
        int r = ((c>>1) & 2) | ((c>>5) & 1);
        int g = ((c>>0) & 2) | ((c>>4) & 1);
        int b = ((c<<1) & 2) | ((c>>1) & 1);
        r |= r<<2 | r<<4 | r<<6;
        g |= g<<2 | g<<4 | g<<6;
        b |= b<<2 | b<<4 | b<<6;
        *dst++ = r;
        *dst++ = g;
        *dst++ = b;
    }
    return ni->colors;
}


static ssize_t
_handle_palette_8bit(nile_t *ni, int type, const uint8_t *data, size_t len)
{
    switch(type){
    case 0:
        DEBUG_p("monochrome text");
        break;
    case 1:
        DEBUG_p("RGB text bitmap");
        if(len < 16*3)
            return -1;
        ni->colors = 16;
        ni->bpc = 3;
        ni->format = NILE_FORMAT_r8g8b8;
        memcpy(ni->palette, data, 16*3);
        return 16*3;
    case 2:
        DEBUG_p("RGB lookup table");
        ni->colors = 1<<ni->depth;
        if(len < ni->colors*3)
            return -1;
        ni->bpc = 3;
        ni->format = NILE_FORMAT_r8g8b8;
        memcpy(ni->palette, data, ni->colors*3);
        return ni->colors*3;
    case 3:
        DEBUG_p("RGB pixels");
        break;
    }
    DEBUG_p("block count: %d",data[1]);
    return -1;
}


static int
_unpack_planes(uint8_t **bitplane, int stride, int height, int depth, const uint8_t *data, size_t len)
{

    uint8_t changemap[4096];
    int maplen = (stride+7)/8;
    rle_t rle;
    int plane;

    _rle_init(&rle,data,len);

    for(plane=0;plane<depth;++plane){
        uint8_t *dst = bitplane[plane];
        int x;
        int y;
        int b;

        DEBUG_p("bytes @1: %d",rle.bytes);
        for(x=0;x<stride;++x){
            b = _rle_get(&rle);
            if(b<0){
                DEBUG_p("out of data #1");
                return -1;
            }
            *dst++ = b;
        }

        DEBUG_p("bytes @2: %d",rle.bytes);
        for(y=1;y<height;++y){
            for(x=0;x<maplen;++x){
                b = _rle_get(&rle);
                if(b<0){
                    DEBUG_p("out of data #2");
                    return -1;
                }
                changemap[x] = b;
                //DEBUG_p("changemap %d = %x",x,b);
            }
            for(x=0;x<stride;++x){
                int bits;
                if(!(x&7)){
                    bits = changemap[x/8];
                }
                bits<<=1;
                if(bits&0x100){
                    //DEBUG_p("there is a change at %dx%d",x,y);
                    b = _rle_get(&rle);
                    if(b<0){
                        DEBUG_p("out of data #3");
                        return -1;
                    }
                    *dst++ = b;
                }else{
                    *dst = dst[-stride];
                    ++dst;
                }
            }
        }
    }
    DEBUG_p("bytes left: %d",(int)(rle.end-rle.p));
    return 0;
}


static cairo_surface_t *
_surface_from_packed(uint8_t *src, int width, int height, int depth, nile_t *palette_ni)
{
    uint32_t palette[256];
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int skip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t) - width;
    int shifts = 8-depth;
    nile_copy_colors_rgb((uint8_t*)palette, 4, NIL_CAIRO_FORMAT_RGB24, palette_ni->palette, palette_ni->bpc, palette_ni->format, palette_ni->colors);
    while(height-->0){
        int x = width;
        for(;;){
            uint8_t pix = *src++;
            int i;
            for(i=8;i>0;i-=depth){
                *dst++ = palette[pix >> shifts];
                pix <<= depth;
                if(--x <= 0)
                    goto nextrow;
            }
        }
nextrow:
        dst += skip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static cairo_surface_t *
_surface_from_text(uint8_t *text, int width, int height)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24, width*8, height*16);
    nile_t ni;
    int y;
    nile_init(&ni);
    nil_cairo_from_surface(&ni, surface);
    for(y=0;y<height;++y){
        int x;
        for(x=0;x<width;++x){
            int ch = *text++;
            nil_pctext_draw_character(&ni, x*8, y*16, ch, 0x000000, 0xffffff, 8);
        }
    }
    nile_done(&ni);
    cairo_surface_mark_dirty(surface);
    return surface;
}


static cairo_surface_t *
_surface_from_text_and_attr(uint8_t *text, uint8_t *attr, int width, int height, nile_t *palette_ni)
{
    uint32_t palette[256];
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24, width*8, height*16);
    nile_t ni;
    int y;
    nile_copy_colors_rgb((uint8_t*)palette, 4, NIL_CAIRO_FORMAT_RGB24, palette_ni->palette, palette_ni->bpc, palette_ni->format, palette_ni->colors);
    nile_init(&ni);
    nil_cairo_from_surface(&ni, surface);
    for(y=0;y<height;++y){
        int x;
        for(x=0;x<width;++x){
            int ch = *text++;
            int color = *attr++;
            nil_pctext_draw_character(&ni, x*8, y*16, ch, palette[color >> 4], palette[color & 15], 8);
        }
    }
    nile_done(&ni);
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_handle_gx1(abydos_plugin_handle_t *h, const uint8_t *data, const uint8_t *end)
{
    nile_t ni;
    static const uint8_t bw_palette[] = {0,255};
    uint8_t palette[256*3];
    int size;
    int ret = -1;

    if(data+3 > end)
        return -1;

    if(data[0] != 0xfa){
        h->info->error = "Invalid data";
        return -1;
    }

    nile_init(&ni);
    ni.palette = palette;

    switch(data[1]){
    case 4:
        /* Header size = 3 */
        data += 3;

        h->info->width = ni.width = 640;
        h->info->height = ni.height = 200;
        h->info->pixel_ratio = 5.0/12;
        ni.depth = 1;

        ni.palette = (uint8_t*)bw_palette;
        ni.colors = 2;
        ni.bpc = 1;
        ni.format = NILE_FORMAT_g8;
        break;
    case 13:
        /* Header size = 4 */
        if(data+4 > end)
            goto out;
        data += 4;

        h->info->width = ni.width = 640;
        h->info->height = ni.height = 350;
        h->info->pixel_ratio = 35.0/48;
        ni.depth = 4;

        size = _handle_palette_2bit(&ni, data, end-data);
        if(size<0)
            goto out;
        data += size;
        break;
    case 14:
        /* Header size = 4 */
        if(data+4 > end)
            goto out;
        data += 4;

        h->info->width = ni.width = 640;
        h->info->height = ni.height = 480;
        ni.depth = 4;

        size = _handle_palette_2bit(&ni, data, end-data);
        if(size<0)
            goto out;
        data += size;
        break;
    default:
        h->info->error = "Unknown file type";
        goto out;
    }

    nile_ensure_bitplanes(&ni, ni.width, ni.height, ni.depth, 1);
    ret = _unpack_planes(ni.bitplanes, nile_get_stride(&ni), ni.height, ni.depth, data, end-data);
    if(ret>=0)
        h->surface = nil_cairo_surface_create(&ni);
out:
    nile_done(&ni);
    return ret;
}


static int
_handle_gx2(abydos_plugin_handle_t *h, const uint8_t *data, const uint8_t *end)
{
    nile_t ni;
    int palette_type;
    int bitmap_type;
    uint32_t palette[256];
    ssize_t size;
    int screen_width;
    int screen_height;
    int ret = -1;

    if(data+18 > end){
        h->info->error = "Data too short";
        return -1;
    }
    if(memcmp(data,"GX2",3)){
        h->info->error = "Invalid data";
        return -1;
    }
    if(data[3] != 1){
        h->info->error = "Unknown version";
        return -1;
    }

    nile_init(&ni);
    ni.palette = (uint8_t*)palette;
    ni.bpc = 4;
    ni.format = NIL_CAIRO_FORMAT_RGB24;

    size = get16(data+4);
    DEBUG_p("header size: %d",(int)size);
    ni.depth = data[6];
    DEBUG_p("bits per pixel: %d", ni.depth);
    h->info->width = ni.width = get16(data+7);
    h->info->height = ni.height = get16(data+9);
    DEBUG_p("dimensions: %dx%d",h->info->width,h->info->height);
    screen_width = get16(data+11);
    screen_height = get16(data+13);
    DEBUG_p("screen size %d:%d",screen_width, screen_height);
    h->info->pixel_ratio = (double)(h->info->height*screen_width)/(h->info->width*screen_height);
    data += size;

    palette_type = data[0] & 15;
    bitmap_type = data[0] >> 7;
    DEBUG_p("palette type: %d", palette_type);
    DEBUG_p("bitmap type: %d", bitmap_type);
    DEBUG_p("blocks: %d",data[1]);
    data += 2;

    size = _handle_palette_8bit(&ni, palette_type, (uint8_t*)data, end-data);
    if(size<0)
        goto out;
    data += size;

    switch(palette_type){
    case 0: /* Monochrome text */
        {
            int stride = h->info->width * h->info->height;
            uint8_t *text = (uint8_t*)malloc(stride);
            DEBUG_p("need %d bytes, got %d compressed bytes", stride, (int)(end-data));
            ret = _unpack_planes(&text, stride, 1, 1, data, end-data);
            if(ret >= 0)
                h->surface = _surface_from_text(text, h->info->width, h->info->height);
            h->info->width *= 8;
            h->info->height *= 16;
            free(text);
        }
        break;
    case 1: /* RGB Text Bitmap */
        {
            int bytes = h->info->width * h->info->height * 2;
            uint8_t *cells = (uint8_t*)malloc(bytes);
            DEBUG_p("need %d bytes, got %d compressed bytes", bytes, (int)(end-data));
            ret = _unpack_planes(&cells, bytes, 1, 1, data, end-data);
            if(ret >= 0)
                h->surface = _surface_from_text_and_attr(cells, cells + bytes/2, h->info->width, h->info->height, &ni);
            h->info->width *= 8;
            h->info->height *= 16;
            free(cells);
        }
        break;
    case 2: /* RGB Lookup table */
        switch(bitmap_type){
        case 0: /* Bitplanes */
            nile_ensure_indexed(&ni, ni.width, ni.height);
            ret = _unpack_planes(&ni.pixels, ni.width * ni.depth / 8, ni.height, 1, data, end-data);
            if(ret>=0)
                h->surface = _surface_from_packed(ni.pixels, ni.width, ni.height, ni.depth, &ni);
            break;
        case 1: /* Uniplanar */
            nile_ensure_bitplanes(&ni, ni.width, ni.height, ni.depth, 1);
            /* Decompress all bitplanes as one */
            ret = _unpack_planes(ni.bitplanes, nile_get_stride(&ni), ni.height * ni.depth, 1, data, end-data);
            if(ret>=0)
                h->surface = nil_cairo_surface_create(&ni);
            break;
        }
        break;
    default:
        DEBUG_p("unknown palette type");
        goto out;
    }
out:
    nile_done(&ni);
    return ret;
}


static int
_gxx_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    switch(h->type){
    case T_GX1:
        return _handle_gx1(h, (uint8_t*)data, (uint8_t*)data+len);
    case T_GX2:
        return _handle_gx2(h, (uint8_t*)data, (uint8_t*)data+len);
    }
    /* not reached */
    return -1;
}


static cairo_surface_t *
_gxx_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _gxx_get_mime_types,
        .new = _gxx_new,
        .free = _gxx_free,
        .create_from_data = _gxx_create_from_data,
        .get_image_surface = _gxx_get_image_surface,
    };
    return &p;
}
