/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

/*

offset | size | description
-------+------+--------------
     0 |    4 | ".NIC"
     4 |    2 | version (=3)
     6 |    1 | copyright notice length (=59)
     7 |      | copyright notice

word aligned after copyright notice:

     0 |    2 | Number of entries
     2 |    4 | creation date
     6 |    4 | modification date
    10 |   26 | author (null terminated)
    36 |   36 | comment 1
    72 |   36 | comment 2
   108 |   36 | comment 3
   144 |    2 | extraction code size
   146 |      | extraction code

*/


#define DECRYPT_KEY(a) ((55+(33*a))%256)
#define ICONSIZE_32x28 (4*28*2+20)


typedef uint32_t color_t;

typedef struct{
    /*
    int width;
    int height;
    */
    int colors;
    double pixel_ratio;
    cairo_surface_t *surface;
}variant_t;

typedef struct{
    int width;
    int height;
    variant_t *variant;
    int variant_count;
}page_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    page_t *page;
};

typedef int(*decode_func_t)(page_t *,const uint8_t *,size_t,int);


static const uint8_t _gdos2st[16] = { 0,15, 1, 2, 4, 6, 3, 5,
                                      7, 8, 9,10,12,14,11,13};

static const color_t _neodesk_palette4[4] = {
#if 0
    0xffff,
    0xf800,
    0x0333,
    0x0000
#else
    0xffffffff,
    0xffff0000,
    0xff006699,
    0xff000000
#endif
};

static const color_t _neodesk_palette16[16] = {
#if 0
    0xffff,
    0xf800,
    0x0333,
    0xffe0,
    0x001f,
    0xf81f,
    0x07ff,
    0xce79,
    0x8c51,
    0xa800,
    0x0540,
    0xad40,
    0x0015,
    0xa815,
    0x0555,
    0x0000
#else
    0xffffffff,
    0xffff0000,
    0xff006699,
    0xffffff00,
    0xff0000ff,
    0xffff00ff,
    0xff00ffff,
    0xffcccccc,
    0xff888888,
    0xffaa0000,
    0xff00aa00,
    0xffaaaa00,
    0xff0000aa,
    0xffaa00aa,
    0xff00aaaa,
    0xff000000
#endif
};


static const char **
_nic_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-neodesk-icon",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_nic_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->page = NULL;
    return h;
}


static void
_nic_free(abydos_plugin_handle_t *h)
{
    if(h->page){
        int i;
        for(i=0;i<h->info->page_count;++i)
            if(h->page[i].variant){
                int j;
                for(j=0;j<h->page[i].variant_count;++j)
                    if(h->page[i].variant[j].surface)
                        cairo_surface_destroy(h->page[i].variant[j].surface);
                free(h->page[i].variant);
            }
        free(h->page);
    }
    free(h);
}


#define get8(a) (*(uint8_t*)(a))
#define get16(a) be16toh(*(uint16_t*)(a))

#define get8decrypt(a,k) (get8(a)^(k))


static int
_get16decrypt(const uint8_t *a, uint8_t key)
{
    int ret = a[0] ^ key;
    key += 33;
    return ret << 8 | (a[1] ^ key);
}


static uint32_t
_get32decrypt(const uint8_t *a, uint8_t key)
{
    uint32_t ret = _get16decrypt(a, key);
    key += 33*2;
    return ret << 16 | _get16decrypt(a+2, key);
}


static void
_copy_decrypt(uint8_t *dst, const uint8_t *src, size_t len, uint8_t key)
{
    while(len-->0){
        *dst++ = *src++ ^ key;
        key += 33;
    }
}


static void
_copy_decrypt_uncompress(uint8_t *dst, const uint8_t *src, size_t len, uint8_t key)
{
    while(len-->0){
        int b = get8decrypt(src++,key);
        int count;
        key += 33;
        switch(b>>6){
        case 0:
            count = (b&63)+1;
            len -= count;
            while(count-->0){
                *dst++ = get8decrypt(src++,key);
                key += 33;
            }
            break;
        case 1:
            count = (b&63)+1;
            b = get8decrypt(src++,key);
            key += 33;
            --len;
            memset(dst, b, count);
            dst += count;
            break;
        case 2:
            count = ((b>>4)&3)+2;
            len -= count;
            b = (b&15)+1;
            DEBUG_p("copy %d blocks of size %d",b+1,count);
            {
                const uint8_t *p = dst;
                int count2 = count;
                while(count2-->0){
                    *dst++ = get8decrypt(src++,key);
                    key += 33;
                }
                while(b-->0){
                    memcpy(dst, p, count);
                    dst += count;
                }
            }
            break;
        default:
            return;
        }
    }
}


static uint32_t
_murmur32(const char *s, int len)
{
    uint32_t h = 3323198485ul;
    while(len-->0){
        h ^= *s++;
        h *= 0x5bd1e995;
        h ^= h >> 15;
    }
    return h;
}


static cairo_surface_t *
_surface_from_bits32x28(const uint8_t *src1, int apply_mask)
{
    const uint8_t *src2 = src1 + 112;
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, 32, 28);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    int y;
    for(y=0;y<28;++y){
        int x;
        for(x=0;x<32;++x){
            int bits;
            if(0==(x&7)){
                bits = *src1++;
                if(apply_mask)
                    bits ^= *src2++;
            }
            bits <<= 1;
            dst[x] =  bits&(1<<8) ? 0xff000000 : 0xffffffff;
        }
        dst += stride;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static void
_apply_mask(uint8_t **bitplane, const uint8_t *bitmask, int plane_size, int depth)
{
    int offs;
    for(offs=0;offs<plane_size;++offs){
        int plane;
        for(plane=0;plane<depth;++plane)
            bitplane[plane][offs] ^= bitmask[offs];
    }
}


static cairo_surface_t *
_surface_from_bitplanes(uint8_t **bitplane, int width, int height, int depth, const color_t *palette)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, width, height);
    color_t *dst = (color_t*)cairo_image_surface_get_data(surface);
    int stride = cairo_image_surface_get_stride(surface)/sizeof(color_t);
    off_t offs = 0;
    int plane;
    while(height-->0){
        int x;
        for(x=0;x<width;++x){
            int bits[8];
            int mask;
            int index;
            if(0==(x&7)){
                for(plane=0;plane<depth;++plane)
                    bits[plane] = bitplane[plane][offs];
                ++offs;
                mask = 1<<8;
            }
            index = 0;
            mask >>= 1;
            for(plane=0;plane<depth;++plane)
                if(bits[plane]&mask)
                    index |= 1<<plane;;
            dst[x] = palette[index];
        }
        dst += stride;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static cairo_surface_t *
_surface_from_bitplanes_with_mask(uint8_t **bitplane, const uint8_t *bitmask, int width, int height, int depth, const color_t *palette)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, width, height);
    color_t *dst = (color_t*)cairo_image_surface_get_data(surface);
    int stride = cairo_image_surface_get_stride(surface)/sizeof(color_t);
    off_t offs = 0;
    int plane;
    while(height-->0){
        int x;
        for(x=0;x<width;++x){
            int bits[8];
            int masks;
            int mask;
            if(0==(x&7)){
                for(plane=0;plane<depth;++plane)
                    bits[plane] = bitplane[plane][offs];
                masks = bitmask[offs];
                ++offs;
                mask = 1<<8;
            }
            mask >>= 1;
            if(masks&mask){
                int index = 0;
                for(plane=0;plane<depth;++plane)
                    if(bits[plane]&mask)
                        index |= 1<<plane;;
                dst[x] = palette[index];
            }else{
                dst[x] = 0;
            }
        }
        dst += stride;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_get_chunk(const uint8_t *src, uint8_t *dst, int len)
{
    uint8_t chunk[512];
    int chunk_size;
    int compression;

    _copy_decrypt(chunk, src, 512, DECRYPT_KEY(0));

    chunk_size = get16(src+0);
    DEBUG_p("chunk size: %d",chunk_size);
    compression = get8(src+2);
    if(compression&3){
        /* Compressed data */
        DEBUG_p("compressed data (%d/%d)",len,chunk_size);
        _copy_decrypt_uncompress(dst, src+3, len, DECRYPT_KEY(0));
    }else{
        /* Uncompressed data */
        DEBUG_p("uncompressed data (%d/%d)",len,chunk_size);
        _copy_decrypt(dst, src+3, len, DECRYPT_KEY(0));
    }
    return 0;
}


static int
_decode_3f2ed41f(page_t *page, const uint8_t *data, size_t len, int index)
{
    /*

    Entry:

    offset | size | description
    -------+------+------------
         0 |    1 | Width (in words)
         1 |    1 | Height (in pixels)
         3 |    1 | draw
         4 |    1 | color
        18 |    4 | unselected bitplanes low res offset
        22 |    4 | unselected mask low res offset
        26 |    4 | selected bitplanes low res offset
        30 |    4 | selected mask low res offset
        34 |    4 | unselected bitplanes med res offset
        38 |    4 | unselected mask med res offset
        42 |    4 | selected bitplanes med res offset
        46 |    4 | selected mask med res offset
        50 |    4 | unselected bitplanes high res offset
        54 |    4 | unselected mask high res offset
        58 |    4 | selected bitplanes high res offset
        62 |    4 | selected mask high res offset

    */
    //const uint8_t *end = data+len;
    const uint8_t *entry = data+index*66;
    uint8_t header[66];
    uint32_t offset[12];
    int page_count;
    int i;

    _copy_decrypt(header, entry, 66, DECRYPT_KEY(0));
    //depth = get8decrypt(entry+8,DECRYPT_KEY(8)) & 15;
#if 1
    _copy_decrypt((uint8_t*)offset, entry+18, 12*4, DECRYPT_KEY(18));
    for(i=0;i<12;++i)
        offset[i] = be32toh(offset[i]);
#else
    for(i=0;i<12;++i)
        offset[i] = _get32decrypt(entry+18+i*4,DECRYPT_KEY(18+i*4));
#endif
    page_count = 1;
    for(i=0;i<3;++i){
        if(offset[i*4+1] || offset[i*4+2]){
            page_count = 2;
            break;
        }
    }

    if(page){
        int stride = get8decrypt(entry+0,DECRYPT_KEY(0))*2;
        int height = get8decrypt(entry+1,DECRYPT_KEY(1));
        int plane_size = stride * height;
        uint8_t *mask = (uint8_t*)malloc(plane_size * 5);
        uint8_t *bits = mask + plane_size;
        color_t bicolor_palette[2];
        const color_t *palette[3];
        uint8_t *bitplane[4];

        bicolor_palette[0] = _neodesk_palette16[0];
        //bicolor_palette[1] = _neodesk_palette16[_gdos2st[get8decrypt(entry+4,DECRYPT_KEY(4))&15]];
        palette[0] = bicolor_palette;
        palette[1] = _neodesk_palette4;
        palette[2] = _neodesk_palette16;

        for(i=0;i<4;++i)
            bitplane[i] = bits + plane_size*i;

        for(i=0;i<page_count;++i){
            int j;
            page[i].width = stride*8;
            page[i].height = height;
            page[i].variant = (variant_t*)malloc(3*sizeof(variant_t));
            page[i].variant_count = 0;
            for(j=0;j<3;++j){
                if(offset[j*4+i*2] || offset[j*4+1]){
                    int depth = 1<<j;
                    variant_t *variant = &page[i].variant[page[i].variant_count++];
                    int bits_size = plane_size * depth;
                    DEBUG_p("bits size: %d",bits_size);
                     bicolor_palette[1] = _neodesk_palette16[15];
                    if(offset[j*4+i*2]){
                        /* Specific selected image exists */
                        _get_chunk(data+offset[j*4+i*2]-996, bits, bits_size);
                        if(offset[j*4+i*2+1]){
                            _get_chunk(data+offset[j*4+i*2+1]-996, mask, plane_size);
                            variant->surface = _surface_from_bitplanes_with_mask(bitplane, mask, page[i].width, page[i].height, depth, palette[j]);
                        }else{
                            bicolor_palette[1] = _neodesk_palette16[_gdos2st[get8decrypt(entry+4,DECRYPT_KEY(4))&15]];
                            variant->surface = _surface_from_bitplanes(bitplane, page[i].width, page[i].height, depth, palette[j]);
                        }
                    }else{
                        /* No specific selected image exists, but a mask */
                        _get_chunk(data+offset[j*4]-996, bits, bits_size);
                        _get_chunk(data+offset[j*4+1]-996, mask, plane_size);
                        _apply_mask(bitplane, mask, plane_size, depth);
                        variant->surface = _surface_from_bitplanes_with_mask(bitplane, mask, page[i].width, page[i].height, depth, palette[j]);
                    }
                }
            }
        }

        free(mask);
    }
    return page_count;
}


static int
_decode_db55f47c(page_t *page, const uint8_t *data, size_t len, int index)
{
    /*

    Entry:

    offset | size | description
    -------+------+-----------------
         0 |    1 | width in words
         1 |    1 | height in pixels
         6 |    1 | xchar
         7 |    1 | ychar
         8 |    1 | planes
         9 |    1 | type
        10 |   12 | filename mask

    */
    if(page){
        //const uint8_t *end = data+len;
        const uint8_t *entry = data+index*30;
        //uint8_t header[30];
        uint8_t plane_color[8];
        uint8_t *bitplane[8];
        color_t palette[2];
        int plane_size;
        uint8_t *mask;
        int bits_size;
        int stride = get8decrypt(entry+0,DECRYPT_KEY(0))*2;
        int depth = get8decrypt(entry+8,DECRYPT_KEY(8)) - 1;
        uint32_t offset;
        int i;
        //_copydecrypt(header, entry, 30, DECRYPT_KEY(0));
        page->width = stride*8;
        page->height = get8decrypt(entry+1,DECRYPT_KEY(1));
        plane_size = stride * page->height;
        bits_size = plane_size * (1+depth);
        offset = _get32decrypt(entry+22,DECRYPT_KEY(22)); /* = 0x00000e80 */
        DEBUG_p("colors offset %d",(int)offset);
        _copy_decrypt(plane_color, data+offset-592, 1<<depth, DECRYPT_KEY(0));

        offset = _get32decrypt(entry+26,DECRYPT_KEY(26));
        DEBUG_p("bits offset %d",(int)offset);
        mask = (uint8_t*)malloc(bits_size);
        _copy_decrypt(mask, data+offset-592, bits_size, DECRYPT_KEY(0));

        //page->variant[0].surface = _surface_from_bits(bits, page->width, page->height, depth, plane_color);
        for(i=0;i<1<<depth;++i)
            palette[i] = _neodesk_palette16[_gdos2st[plane_color[i]%15]];
        for(i=0;i<depth;++i)
            bitplane[i] = mask+plane_size*(i+1);
        for(i=0;i<2;++i){
            page[i].variant = (variant_t*)malloc(sizeof(variant_t));
            page[i].variant_count = 1;
            if(i)
                _apply_mask(bitplane, mask, plane_size, depth);
            page[i].variant[0].surface = _surface_from_bitplanes_with_mask(bitplane, mask, page->width, page->height, depth, palette);
        }
        free(mask);
    }
    return 2;
}


static int
_nic_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    if(!memcmp(data,".NIC",4)){
        if(3 == get8(data+4)){
            /* Version 3 */
            int entries;
            int size;
            uint32_t codehash;
            decode_func_t decode_func;

            size = 7 + get8(data+6);
            size+= size&1;
            DEBUG_p("skipping over %d bytes",size);
            data += size;
            len -= size;

            entries = _get16decrypt((uint8_t*)data+0,DECRYPT_KEY(0));
            DEBUG_p("icon entries: %d",entries);
#ifdef USE_DEBUG
            {
                char buffer[40];
                buffer[36] = '\0';
                _copy_decrypt((uint8_t*)buffer, (uint8_t*)data+10, 26, DECRYPT_KEY(8));
                DEBUG_p("author: %s",buffer);
                _copy_decrypt((uint8_t*)buffer, (uint8_t*)data+36, 36, DECRYPT_KEY(34));
                DEBUG_p("comment 1: %s",buffer);
                _copy_decrypt((uint8_t*)buffer, (uint8_t*)data+72, 36, DECRYPT_KEY(70));
                DEBUG_p("comment 2: %s",buffer);
                _copy_decrypt((uint8_t*)buffer, (uint8_t*)data+108, 36, DECRYPT_KEY(106));
                DEBUG_p("comment 3: %s",buffer);
            }
#endif

            size = _get16decrypt((uint8_t*)data+144, DECRYPT_KEY(0));
            DEBUG_p("extraction code size: %d", size);

            codehash = _murmur32(data+146,size);
            DEBUG_p("code hash: %08x",codehash);

            switch(codehash){
            case 0x3f2ed41f:
                decode_func = _decode_3f2ed41f;
                break;
            case 0xdb55f47c:
                decode_func = _decode_db55f47c;
                break;
            default:
                decode_func = NULL;
#ifdef USE_DEBUG
                {
                    char filename[256];
                    FILE *f;
                    sprintf(filename,"code-%08x.bin",codehash);
                    DEBUG_p("unknown decoder, writing to %s",filename);
                    f = fopen(filename,"w");
                    fwrite(data+146, 1, size, f);
                    fclose(f);
                }
#endif
                break;
            }
            if(decode_func){
                int n;
                int i;
                size += 146;
                DEBUG_p("skipping over %d bytes",size);
                data += size;
                len -= size;
                n = 0;
                for(i=0;i<entries;++i)
                    n += decode_func(NULL, (uint8_t*)data, len, i);
                h->info->page_count = n;
                h->page = (page_t*)malloc(n*sizeof(page_t));
                h->info->page_count = n;
                n = 0;
                for(i=0;i<entries;++i){
                    DEBUG_p("page %d of %d",n,h->info->page_count);
                    n += decode_func(&h->page[n], (uint8_t*)data, len, i);
                }
            }
            return 0;
        }else{
            h->info->error = "unknown version";
            return -1;
        }
    }
    if(2088 == len){
        /* Version 1 */
        int i;
        h->info->page_count = 9;
        h->page = (page_t*)malloc(h->info->page_count * sizeof(page_t));
        for(i=0;i<h->info->page_count;++i){
            DEBUG_p("[%s]",data+ICONSIZE_32x28*(i/2) + 4*28*2);
            h->page[i].width = 32;
            h->page[i].height = 28;
            h->page[i].variant = (variant_t*)malloc(sizeof(variant_t));
            h->page[i].variant_count = 1;
            h->page[i].variant[0].surface = _surface_from_bits32x28((uint8_t*)data+ICONSIZE_32x28*(i/2),i&1);
        }
        return 0;
    }
    if(1 /* 0 == len % ICONSIZE_32x28 */){
        /* Version 2 */
        int i;
        h->info->page_count = (len / ICONSIZE_32x28) * 2;
        h->page = (page_t*)malloc(h->info->page_count * sizeof(page_t));
        for(i=0;i<h->info->page_count;++i){
            DEBUG_p("[%s]",data+ICONSIZE_32x28*(i/2) + 4*28*2);
            h->page[i].width = 32;
            h->page[i].height = 28;
            h->page[i].variant = (variant_t*)malloc(sizeof(variant_t));
            h->page[i].variant_count = 1;
            h->page[i].variant[0].surface = _surface_from_bits32x28((uint8_t*)data+ICONSIZE_32x28*(i/2),i&1);
        }
        return 0;
    }

    DEBUG_p("unknown subtype of size %d",(int)len);
    return -1;
}


static cairo_surface_t *
_nic_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    /* FIXME!! remove this hack then using color count for selecting best variant. */
    variant = h->page[page].variant_count-1;
    //variant = h->page[page].variant_count>1 ? 1 : 0;

    return cairo_surface_reference(h->page[page].variant[variant].surface);
}


static void
_nic_get_page_info(abydos_plugin_handle_t *h, int page, abydos_plugin_page_info_t *info)
{
    info->width = h->page[page].width;
    info->height = h->page[page].height;
    info->variant_count = h->page[page].variant_count;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _nic_get_mime_types,
        .new = _nic_new,
        .free = _nic_free,
        .create_from_data = _nic_create_from_data,
        .get_image_surface = _nic_get_image_surface,
        .get_page_info = _nic_get_page_info
    };
    return &p;
}
