/*
 * Copyright (C) 1999         The Free Software Foundation
 * Copyright (C) 2019 - 2021  Magnus Bergman <magnus.bergman@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, see <http://www.gnu.org/licenses/>.
 */
/*
 * Based on io-ras.c from gdk-pixbuf 2.30
 * Authors: Arjan van de Ven <arjan@fenrus.demon.nl>
 *          Federico Mena-Quintero <federico@gimp.org>
 */
/*
 * Known bugs:
 * Compressed rasterfiles don't work yet
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <endian.h>
#include <nile.h>
#include <nil-cairo.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>

#include "buffer.h"

#define HEADER_SIZE 32

#if __BYTE_ORDER == __BIG_ENDIAN
# define RAS_MAGIC 0x59a66a95
#else
# define RAS_MAGIC 0x956aa659
#endif

/* Sun supported ras_types */
#define RT_OLD          0               /* Raw pixrect image in 68000 byte order */
#define RT_STANDARD     1               /* Raw pixrect image in 68000 byte order */
#define RT_BYTE_ENCODED 2               /* Run-length compression of bytes */
#define RT_FORMAT_RGB   3               /* XRGB or RGB instead of XBGR or BGR */
#define RT_FORMAT_TIFF  4               /* tiff <-> standard rasterfile */
#define RT_FORMAT_IFF   5               /* iff (TAAC format) <-> standard rasterfile */
#define RT_EXPERIMENTAL 0xffff          /* Reserved for testing */

/* Sun supported ras_maptypes */
#define RMT_NONE        0               /* ras_maplength is expected to be 0 */
#define RMT_EQUAL_RGB   1               /* red[ras_maplength/3],green[],blue[] */
/* Sun registered ras_maptypes */
#define RMT_RAW         2

/*
 * Some software, such as Imagemagick, uses the 8 bit padding in the
 * 32-format as an alpha channel. This is not originally part of the format
 * and therefore disabled by default.
 */
/* #define SUPPORT_ALPHA */

/*
   Header structure for sunras files.
   All values are in big-endian order on disk

   Note: Every scanline is padded to be a multiple of 16 bits
 */
typedef struct{
    uint32_t magic;
    uint32_t width;
    uint32_t height;
    uint32_t depth;
    uint32_t length;
    uint32_t type;
    uint32_t maptype;
    uint32_t maplength;
}header_t;

typedef struct _parser_t parser_t;

struct _parser_t{
    /* metadata */
    int header_type;                    /* Image encoding */
    int map_bytes;                      /* Size of the colormap in bytes */
    nile_t src;
    nile_blit_t blit;

    /* decoding state */
    abydos_buffer_t buf;
    enum{
        PS_HEADER,
        PS_CMAP,
        PS_DATA,
        PS_DONE
    }state;
    uint8_t *dst;
    int dst_stride;
    int current_line;
    /* decompression */
    uint8_t *line_buf;                  /* Buffer for 1 line */
    int LineDone;                       /* # of bytes in LineBuf */
    int DecoderState;

    /* user callbacks */
    abydos_plugin_info_func_t info_func;
    abydos_plugin_progress_func_t progress_func;
    void *userdata;
};

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
    parser_t parser;
};


static const char **
_sunras_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-cmu-raster",
        "image/x-sun-raster",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_sunras_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->surface = NULL;
    return h;
}


static void
_sunras_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


static parser_t *
_parser_init(parser_t *p)
{
    nile_init(&p->src);
    p->src.type = NILE_TYPE_PLANAR;
    p->src.stride = NILE_STRIDE_ALIGN16;
    p->src.bpc = 4;
    p->src.format = NIL_CAIRO_FORMAT_RGB24;
    abydos_buffer_clear(&p->buf);
    p->state = PS_HEADER;
    p->current_line = 0;
    p->line_buf = NULL;
    p->LineDone = 0;

    p->DecoderState = 0;
    return p;
}


static void
_parser_done(parser_t *p)
{
    nile_done(&p->src);
    if(p->line_buf)
        free(p->line_buf);
}


static void
_sunras_progressive_begin(abydos_plugin_handle_t *h,abydos_plugin_info_func_t info_func,abydos_plugin_progress_func_t progress_func,void *userdata)
{
    _parser_init(&h->parser);
    h->parser.info_func = info_func;
    h->parser.progress_func = progress_func;
    h->parser.userdata = userdata;
}


static const char *
_process_header(parser_t *p, header_t *header)
{
    int header_maptype;

    if(header->magic != RAS_MAGIC)
        return "invalid header";

    p->header_type = be32toh(header->type);
    header_maptype = be32toh(header->maptype);
    if(p->header_type > RT_FORMAT_RGB || header_maptype > RMT_EQUAL_RGB)
        return "unsupported RAS image variation";

    p->map_bytes = be32toh(header->maplength);
    p->src.colors = p->map_bytes / 3;
    if(p->src.colors > 256)
        return "bogus header data";

    p->src.width = be32toh(header->width);
    p->src.height = be32toh(header->height);
    p->src.depth = be32toh(header->depth);
    if(p->src.width <= 0 || p->src.height <= 0)
        return "zero size image";

    /* 2019-04-11 was the day this became less trivial */
    switch(p->src.depth){
    case 1:
        p->src.type = NILE_TYPE_BITPLANES;
        if(p->src.colors <= 1){
            static const uint32_t bw[2] = {0xffffff,0x000000};
            p->src.palette = (uint8_t*)bw;
            p->src.colors = 2;
        }
        break;
    case 8:
        p->src.type = NILE_TYPE_INDEXED;
        break;
    case 24:
        p->src.bpc = 3;
        p->src.format = RT_FORMAT_RGB == p->header_type ? NILE_FORMAT_r8g8b8 : NILE_FORMAT_b8g8r8;
        break;
    case 32:
#ifdef SUPPORT_ALPHA
        p->src.format = RT_FORMAT_RGB == p->header_type ? NILE_FORMAT_a8r8g8b8 : NILE_FORMAT_b8g8r8a8;
#else
        p->src.format = RT_FORMAT_RGB == p->header_type ? NILE_FORMAT_x8r8g8b8 : NILE_FORMAT_b8g8r8;
#endif
        break;
    default:
        return "unknown type";
    }
    p->line_buf = malloc(nile_get_stride(&p->src));
    return NULL;
}


static void
_process_line(parser_t *p, const uint8_t *src)
{
    p->LineDone = 0;
    if(p->current_line >= p->src.height)
        return;

    nile_blit_line(&p->blit, p->dst, src, p->src.width);
    p->dst += p->dst_stride;
    ++p->current_line;
}


static void
_decompress(parser_t *p,abydos_buffer_t *buf)
{
    int src_stride = nile_get_stride(&p->src);
    while(abydos_buffer_size(buf)>0){
        int b = abydos_buffer_pop(buf);
        switch(p->DecoderState){
        case 0:
            if(0x80==b){
                p->DecoderState = 1;
            }else{
                p->line_buf[p->LineDone++] = b;
            }
            break;
        case 1:
            if(0==b){
                p->line_buf[p->LineDone++] = 0x80;
                p->DecoderState = 0;
            }else{
                p->DecoderState = b+1;
            }
            break;
        default:
            for(; p->DecoderState; --p->DecoderState) {
                p->line_buf[p->LineDone++] = b;
                if(p->LineDone >= src_stride)
                    _process_line(p, p->line_buf);
            }
            break;
        }
        if(p->LineDone >= src_stride)
            _process_line(p, p->line_buf);
    }
}


static int
_sunras_progressive_feed(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    parser_t *p = &h->parser;
    abydos_buffer_add(&p->buf, (const unsigned char*)data, len);
    switch(p->state){
    case PS_HEADER:
        if(abydos_buffer_size(&p->buf) < HEADER_SIZE)
            break;
        h->info->error = _process_header(p, (header_t*)abydos_buffer_get(&h->parser.buf, HEADER_SIZE));
        if(h->info->error)
            return -1;
        h->info->width = p->src.width;
        h->info->height = p->src.height;
        h->surface = nil_cairo_surface_create_for_format(p->src.format, p->src.width, p->src.height);
        p->dst = cairo_image_surface_get_data(h->surface);
        p->dst_stride = cairo_image_surface_get_stride(h->surface);
        if(h->parser.info_func){
            /* Notify the user that we are ready to go */
            h->parser.info_func(h->parser.userdata);
        }
        p->state = PS_CMAP;
        /* fall through */
    case PS_CMAP:
        if(abydos_buffer_size(&p->buf) < p->map_bytes)
            break;
        if(p->map_bytes >= 6){
            uint32_t *cairo_palette;
            const uint8_t *src;
            int i;

            nile_ensure_palette(&p->src, p->src.colors, p->src.bpc);
            cairo_palette = (uint32_t*)p->src.palette;
            src = abydos_buffer_get(&p->buf, p->map_bytes);
            for(i=0;i<p->src.colors;++i){
                cairo_palette[i] = src[i+p->src.colors*0] << 16 |
                                   src[i+p->src.colors*1] <<  8 |
                                   src[i+p->src.colors*2] <<  0;
            }
        }else if(p->map_bytes > 0){
            abydos_buffer_skip(&p->buf, p->map_bytes);
        }
        switch(p->src.type){
        case NILE_TYPE_NONE:
            /* Should not happen */
            break;
        case NILE_TYPE_BITPLANES:
            nile_blit_init_packed_rgb(&p->blit, 4, NIL_CAIRO_FORMAT_RGB24, 1, p->src.palette, p->src.colors, p->src.bpc, p->src.format);
            break;
        case NILE_TYPE_INDEXED:
            nile_blit_init_indexed_rgb(&p->blit, 4, NIL_CAIRO_FORMAT_RGB24, p->src.palette, p->src.colors, p->src.bpc, p->src.format);
            break;
        case NILE_TYPE_PLANAR:
            nile_blit_init_planar_rgb(&p->blit, 4, NIL_CAIRO_FORMAT_RGB24, p->src.bpc, p->src.format);
            break;
        }
        p->state = PS_DATA;
        /* fall through */
    case PS_DATA:
        {
            cairo_rectangle_int_t r;
            r.y = h->parser.current_line;
            if(RT_BYTE_ENCODED == p->header_type){
                _decompress(p, &p->buf);
            }else{
                int stride = nile_get_stride(&p->src);
                while(abydos_buffer_size(&p->buf) >= stride){
                    _process_line(p, abydos_buffer_get(&p->buf, stride));
                }
            }
            cairo_surface_mark_dirty(h->surface);
            r.height = h->parser.current_line - r.y;
            if(p->progress_func && r.height > 0){
                r.x = 0;
                r.width = h->info->width;
                p->progress_func(p->userdata,&r);
            }
        }
        if(p->current_line < p->src.height)
            break;
        p->state = PS_DONE;
        /* fall through */
    case PS_DONE:
        DEBUG_p("ignoring %d trailing bytes",(int)abydos_buffer_size(&p->buf));
        abydos_buffer_clear(&p->buf);
        break;
    }
    abydos_buffer_privatize(&p->buf);
    return 0;
}


static int
_sunras_progressive_end(abydos_plugin_handle_t *h)
{
    /* FIXME this thing needs to report errors if
     * we have unused image data
     */
    _parser_done(&h->parser);
    return 0;
}


static cairo_surface_t *
_sunras_get_image_surface(abydos_plugin_handle_t *h,int page,cairo_bool_t *layer_visibility,int variant,double time,int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _sunras_get_mime_types,
        .new = _sunras_new,
        .free = _sunras_free,
        .progressive_begin = _sunras_progressive_begin,
        .progressive_feed = _sunras_progressive_feed,
        .progressive_end = _sunras_progressive_end,
        .get_image_surface = _sunras_get_image_surface
    };
    return &p;
}
