/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
#include <stddef.h>                     /* NULL */
#include <stdlib.h>                     /* malloc(), free() */

#include "loa-canvas.h"


loa_canvas_t *
loa_canvas_create(loa_canvas_type_t type)
{
    loa_canvas_t *c = malloc(sizeof(loa_canvas_t));
    c->refcount = 1;
    c->type = type;
    if(LOA_CANVAS_GROUP == type)
        c->group.first = NULL;
    return c;
}


loa_canvas_t *
loa_canvas_reference(loa_canvas_t *c)
{
    ++c->refcount;
    return c;
}


void
loa_canvas_destroy(loa_canvas_t *c, loa_abstract_free_func_t abstract_free, void *userdata)
{
    loa_canvas_part_t *part;
    if(!c)
        return;
    if(0 == --c->refcount){
        switch(c->type){
        case LOA_CANVAS_GROUP:
            for(part=c->group.first;part;){
                loa_canvas_part_t *tmp = part->next;
                abstract_free(userdata, part->abstract);
                free(part);
                part = tmp;
            }
            break;
        case LOA_CANVAS_ABSTRACT:
            abstract_free(userdata, c->abstract);
            break;
        }
        free(c);
    }
}


void loa_canvas_flatten(loa_canvas_t *c, const loa_abstract_interface_t *vt, void *userdata)
{
    loa_abstract_t *abst;
    loa_canvas_part_t *part;
    if(LOA_CANVAS_ABSTRACT == c->type)
        return;
    if(!c->group.first)
        return;
    abst = NULL;
    for(part=c->group.first;part;){
        loa_canvas_part_t *tmp = part->next;
        if(!abst){
            int width;
            int height;
            if(0 == part->x && 0 == part->y && 1 == part->xscale && 1 == part->yscale){
                abst = vt->copy(userdata, part->abstract, 1);
                goto done;
            }
            vt->get_size(userdata, part->abstract, &width, &height);
            abst = vt->create(userdata, width * part->xscale + part->x, height * part->yscale + part->y);
        }
        vt->paste(userdata, abst, part->x, part->y, part->xscale, part->yscale, part->flags, part->abstract, NULL);
done:
        vt->free(userdata, part->abstract);
        free(part);
        part = tmp;
    }
    c->type = LOA_CANVAS_ABSTRACT;
    c->abstract = abst;
}


loa_canvas_part_t *
loa_canvas_add_part(loa_canvas_t *o)
{
    loa_canvas_part_t *part = malloc(sizeof(loa_canvas_part_t));
    part->next = NULL;
    if(o->group.first){
        o->group.last->next = part;
        o->group.last = part;
    }else{
        o->group.first = o->group.last = part;
    }
    return part;
}


loa_canvas_t *
loa_canvas_copy(loa_canvas_t *src, int deep, const loa_abstract_interface_t *vt, void *userdata){
    if(deep){
        loa_canvas_t *dst;
        switch(src->type){
        case LOA_CANVAS_GROUP:
            dst = loa_canvas_create(LOA_CANVAS_GROUP);
            loa_canvas_paste(dst, 0, 0, 1, 1, 0, NULL, src, vt, userdata);
            break;
        case LOA_CANVAS_ABSTRACT:
            dst = loa_canvas_create(LOA_CANVAS_GROUP);
            dst->abstract = vt->copy(userdata, src->abstract, deep);
            break;
        }
        return dst;
    }else{
        return loa_canvas_reference(src);
    }
}


void
loa_canvas_scale(loa_canvas_t *c, double xscale, double yscale)
{
    loa_canvas_part_t *part;
    loa_abstract_t *tmp;
    switch(c->type){
    case LOA_CANVAS_GROUP:
        for(part=c->group.first;part;part=part->next){
            part->x *= xscale;
            part->y *= yscale;
            part->xscale *= xscale;
            part->yscale *= yscale;
        }
        break;
    case LOA_CANVAS_ABSTRACT:
        tmp = c->abstract;
        c->type = LOA_CANVAS_GROUP;
        c->group.first = NULL;
        part = loa_canvas_add_part(c);
        part->abstract = tmp;
        part->x = 0;
        part->y = 0;
        part->xscale = xscale;
        part->yscale = yscale;
        part->flags = 0;
        break;
    }
}


void
loa_canvas_paste(loa_canvas_t *dst,
                 double x,
                 double y,
                 double xscale,
                 double yscale,
                 int flags,
                 nil_clip_t *clip,
                 loa_canvas_t *src,
                 const loa_abstract_interface_t *vt,
                 void *userdata)
{
    switch(dst->type){
    case LOA_CANVAS_GROUP:
        switch(src->type){
        case LOA_CANVAS_GROUP:
            /* group <- group */
            {
                loa_canvas_part_t *src_part;
                for(src_part=src->group.first;src_part;src_part=src_part->next){
                    loa_canvas_part_t *dst_part = loa_canvas_add_part(dst);
                    dst_part->abstract = vt->copy(userdata, src_part->abstract, 0);
                    dst_part->x = x + src_part->x * xscale;
                    dst_part->y = y + src_part->y * yscale;
                    dst_part->xscale = src_part->xscale * xscale;
                    dst_part->yscale = src_part->yscale * yscale;
                    dst_part->flags = src_part->flags | flags;
                }
            }
            break;
        case LOA_CANVAS_ABSTRACT:
            /* group <- abstract */
            {
                loa_canvas_part_t *part = loa_canvas_add_part(dst);
                part->abstract = vt->copy(userdata, src->abstract, 0);
                part->x = x;
                part->y = y;
                part->xscale = xscale;
                part->yscale = yscale;
                part->flags = flags;
            }
            break;
        }
        break;
    case LOA_CANVAS_ABSTRACT:
        switch(src->type){
        case LOA_CANVAS_GROUP:
            /* abstract <- group */
            {
                loa_canvas_part_t *part;
                for(part=src->group.first;part;part=part->next)
                    vt->paste(userdata, dst->abstract,
                              x + part->x * xscale, y + part->y * yscale,
                              part->xscale * xscale, part->yscale * yscale,
                              part->flags | flags, part->abstract, clip);
            }
            break;
        case LOA_CANVAS_ABSTRACT:
            /* abstract <- abstract */
            vt->paste(userdata, dst->abstract, x, y, xscale, yscale, flags,
                      src->abstract, clip);
            break;
        }
        break;
    }
}
