/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <debug.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>

#include "nil-clip.h"
#include "ni-mng.h"
#include "loa-buffer.h"
#include "loa-canvas.h"

#define ENABLE_TWEENING

#ifndef ABS
# define ABS(a) ((a)<0?-(a):(a))
#endif

typedef struct{
    loa_canvas_t *canvas;
    int hide;
    int x;
    int y;
    nil_clip_t clip;
    int paste_x;
    int paste_y;
    int point;
}object_entry_t;

typedef struct{
    object_entry_t table[1<<16];
}object_table_t;

struct _loa_object_t{
    loa_abstract_t *abstract;
    int point;                          /* Animatable position and clipping. */
    int x;                              /* Offset to base position. */
    int y;
    double xscale;
    double yscale;
    int flags;
    nil_clip_t clip;
};

struct _loa_layer_t{
    int object;
    int object_count;
};

struct _loa_anim_t{
    enum{
        ANIM_MOVE_ABSOLUTE,
        ANIM_MOVE_RELATIVE,
        ANIM_CLIP_ABSOLUTE,
        ANIM_CLIP_RELATIVE
    }type;
    int point;
    union{
        struct{
            double x;
            double y;
        }move;
        struct{
            int x1;
            int y1;
            int x2;
            int y2;
        }clip;
    };
};

struct _loa_keyframe_t{
    double start;
    double duration;
    int layer;
    int layer_count;
    int anim;
    int anim_count;
};

typedef struct{
    int keyframe;
    int keyframe_count;
    char *name;
    int layer_count;
}page_t;

struct _ni_mng_t{
    loa_abstract_free_func_t free;
    void *userdata;
    int width;
    int height;
#ifdef ENABLE_TWEENING
    int scalable_time;
#endif
    loa_point_t *point;
    int point_count;
    loa_object_t *object;
    int object_count;
    loa_layer_t *layer;
    int layer_count;
    loa_anim_t *anim;
    loa_keyframe_t *keyframe;
    int keyframe_count;
    page_t *page;
    int page_count;
};

typedef struct{
    int id;
    int count;
    const char *start;
    int can_expand;
}loop_t;

typedef struct{
    enum{
        S_MNG,
        S_PNG,
        S_JNG,
        S_BASIS,
        S_LOOP
    }state;
    object_table_t *ot;
    ni_mng_mhdr_t mhdr;
    ni_mng_defi_t defi;
    ni_mng_basi_t basi;
    ni_mng_dhdr_t dhdr;
    ni_mng_back_t back;
#ifdef ENABLE_TWEENING
    int scalable_time;
#endif

    loa_buffer_t buf;
    loa_keyframe_t *current_keyframe;
    int is_first_layer;
    loa_array_t page_array;
    page_t *current_page;
    /* loops */
    loa_array_t loop_array;
    loop_t *current_loop;
    int loop_expand;

    nil_clip_t clip;
    int framing_mode;
    int ticks;
    loa_abstract_t *scratch;
    loa_abstract_t *background;
    //int has_background;
    int last_paste_x;
    int last_paste_y;
    double page_time;

    const ni_mng_callbacks_t *vt;
    void *userdata;
}read_state_t;

/* --- object entry --- */

static loa_layer_t *_begin_layer(read_state_t*);
static void _end_layer(read_state_t*);

static void
_object_entry_paste(object_entry_t *dst,
                    object_entry_t *src,
                    read_state_t *rs,
                    const ni_mng_past_t *past,
                    const ni_mng_past_entry_t *pe)
{
    int x;
    int y;
    int flags = 0;
    switch(pe->composition_mode){
    case 1:
        flags |= LOA_ABSTRACT_REPLACE;
        break;
    case 2:
        flags |= LOA_ABSTRACT_UNDER;
        break;
    }
    switch(pe->orientation){
    case 2:
        flags |= LOA_ABSTRACT_FLIP_HORIZONTAL | LOA_ABSTRACT_FLIP_VERTICAL;
        break;
    case 4:
        flags |= LOA_ABSTRACT_FLIP_HORIZONTAL;
        break;
    case 6:
        flags |= LOA_ABSTRACT_FLIP_VERTICAL;
        break;
    case 8:
        flags |= LOA_ABSTRACT_TILE;
        break;
    }
    switch(pe->offset_origin){
    case 0:
        x = 0;
        y = 0;
        break;
    case 1:
        x = past->target_x;
        y = past->target_y;
        switch(past->target_delta_type){
        case 0:
            break;
        case 1:
            x += dst->paste_x;
            y += dst->paste_y;
            break;
        case 2:
            x += rs->last_paste_x;
            y += rs->last_paste_y;
            break;
        }
        break;
    }
    x += pe->x_offset;
    y += pe->y_offset;
    rs->last_paste_x = dst->paste_x = x;
    rs->last_paste_y = dst->paste_y = y;
    loa_canvas_paste(dst->canvas, x, y, 1, 1, flags, &rs->clip, src->canvas, &rs->vt->abstract, rs->userdata);
}


static void _layer_add_canvas(read_state_t*, loa_canvas_t*, int flags, int point, nil_clip_t *clip);

static void
_object_entry_show(object_entry_t *e, read_state_t *rs)
{
    loa_canvas_t dst;
    _begin_layer(rs);
    _layer_add_canvas(rs, e->canvas, 0, e->point, &rs->clip);
    _end_layer(rs);

    dst.type = LOA_CANVAS_ABSTRACT;
    dst.abstract = rs->scratch;
    loa_canvas_paste(&dst, e->x, e->y, 1, 1, 0, &rs->clip, e->canvas, &rs->vt->abstract, rs->userdata);

    loa_point_t *p;
    e->point = loa_buffer_point_count(&rs->buf);
    p = loa_buffer_add_points(&rs->buf, 3);
    p[0].x = e->x;
    p[0].y = e->y;
    p[1].x = e->clip.x;
    p[1].y = e->clip.y;
    p[2].x = e->clip.x + e->clip.width;
    p[2].y = e->clip.y + e->clip.height;
}

/* --- object table --- */

static object_table_t *
_object_table_init(object_table_t *ot)
{
    memset(ot, 0, sizeof(object_table_t));
    return ot;
}


static void
_object_table_done(object_table_t *ot, read_state_t *rs)
{
    int i;
    for(i=0;i<(1<<16);++i)
        loa_canvas_destroy(ot->table[i].canvas, rs->vt->abstract.free, rs->userdata);
}


static loa_canvas_t *
_object_table_create(object_table_t *ot, int id, loa_canvas_type_t type, read_state_t *rs)
{
    object_entry_t *e = &ot->table[id];
    loa_canvas_destroy(e->canvas, rs->vt->abstract.free, rs->userdata);
    e->canvas = loa_canvas_create(type);
    return e->canvas;
}


#if 0
static void
_object_table_destroy(object_table_t *ot, int id)
{
    object_entry_t *e = &ot->table[id];
    _object_destroy(e->object);
    e->object = NULL;
}
#endif


#define _object_table_get_entry(ot, id) (&(ot)->table[id])
#define _object_table_get(ot, id) (_object_table_get_entry(ot, id)->canvas)


static void
_object_table_apply_defi(object_table_t *ot, ni_mng_defi_t *defi, read_state_t *rs)
{
    object_entry_t *e;
    e = &ot->table[defi->object_id];
    e->hide = defi->hide;
    e->x = defi->posx;
    e->y = defi->posy;
    DEBUG_p("id: %d, x: %d, y: %d, hide: %d", defi->object_id, e->x, e->y, e->hide);
    if(defi->clip_available){
        e->clip.x = defi->clipx1;
        e->clip.y = defi->clipy1;
        e->clip.width = defi->clipx2 - defi->clipx1;
        e->clip.height = defi->clipy2 - defi->clipy1;
    }else{
        e->clip.x = 0;
        e->clip.y = 0;
        e->clip.width = rs->mhdr.width;
        e->clip.height = rs->mhdr.height;
    }
}


static void
_object_table_apply_move(object_table_t *ot, ni_mng_move_t *move)
{
    int i;
    for(i=move->first;i<=move->last;++i){
        object_entry_t *e = &ot->table[i];
        switch(move->type){
        case NI_MNG_MOVE_ABSOLUTE:
            e->x = move->x;
            e->y = move->y;
            break;
        case NI_MNG_MOVE_RELATIVE:
            e->x += move->x;
            e->y += move->y;
            break;
        }
    }
}


static void
_object_table_apply_clip(object_table_t *ot, ni_mng_clip_t *clip)
{
    int i;
    for(i=clip->first;i<=clip->last;++i){
        object_entry_t *e = &ot->table[i];
        switch(clip->type){
        case NI_MNG_CLIP_ABSOLUTE:
            e->clip.x = clip->left_cb;
            e->clip.y = clip->top_cb;
            e->clip.width = clip->right_cb - clip->left_cb;
            e->clip.height = clip->bottom_cb - clip->top_cb;
            break;
        case NI_MNG_CLIP_RELATIVE:
            e->clip.x += clip->left_cb;
            e->clip.y += clip->top_cb;
            e->clip.width += clip->right_cb - clip->left_cb;
            e->clip.height += clip->bottom_cb - clip->top_cb;
            break;
        }
    }
}


static loa_canvas_t *
_object_table_copy(object_table_t *ot, int id, int deep, read_state_t *rs){
    loa_canvas_t *old_object = _object_table_get(ot, id);
    loa_canvas_t *new_object;
    if(old_object){
        loa_canvas_reference(old_object);
        new_object = _object_table_create(ot, id, old_object->type, rs);
        switch(old_object->type){
        case LOA_CANVAS_GROUP:
            {
                loa_canvas_part_t *src;
                for(src=old_object->group.first;src;src=src->next){
                    loa_canvas_part_t *dst = loa_canvas_add_part(new_object);
                    *dst = *src;
                    dst->abstract = rs->vt->abstract.copy(rs->userdata, src->abstract, 0);
                }
            }
            break;
        case LOA_CANVAS_ABSTRACT:
            new_object->abstract = rs->vt->abstract.copy(rs->userdata, old_object->abstract, deep);
            break;
        }
        loa_canvas_destroy(old_object, rs->vt->abstract.free, rs->userdata);
    }else{
        new_object = _object_table_create(ot, id, LOA_CANVAS_GROUP, rs);
    }
    return new_object;
}


#if 0
static loa_canvas_t *
_object_table_soft_copy(object_table_t *ot, int id, loa_canvas_type_t type, read_state_t *rs){
    loa_canvas_t *old_object = _object_table_get(ot, id);
    loa_canvas_t *new_object;
    if(old_object){
        loa_canvas_reference(old_object);
        new_object = _object_table_create(ot, id, type, rs);
        loa_canvas_destroy(old_object, rs->vt->abstract.free, rs->userdata);
    }else{
        new_object = _object_table_create(ot, id, type, rs);
    }
    return new_object;
}
#endif


static void
_object_table_maybe_show(object_table_t *ot, int id, read_state_t *rs)
{
    object_entry_t *e = &ot->table[id];
    if(e->hide){
        DEBUG_p("object %d is hidden", id);
        return;
    }
    _object_entry_show(e, rs);
}

/* --- utility functions --- */

static void
_layer_add_abstract(read_state_t *rs,
                    loa_abstract_t *abst,
                    int point,
                    int x, int y,
                    double xscale, double yscale,
                    int flags, nil_clip_t *clip)
{
    loa_object_t *o = loa_buffer_add_object(&rs->buf);
    o->abstract = rs->vt->abstract.copy(rs->userdata, abst, 0);
    o->point = point;
    o->x = x;
    o->y = y;
    o->xscale = xscale;
    o->yscale = yscale;
    o->flags = flags;
    if(clip){
        o->clip = *clip;
    }else{
        o->clip.x = 0;
        o->clip.y = 0;
        o->clip.width = rs->mhdr.width;
        o->clip.height = rs->mhdr.height;
    }
}


static void
_layer_add_canvas(read_state_t *rs, loa_canvas_t *c, int flags, int point, nil_clip_t *clip)
{
    loa_abstract_t *abst;
#if 1
    loa_canvas_flatten(c, &rs->vt->abstract, rs->userdata);
#endif
    switch(c->type){
    case LOA_CANVAS_GROUP:
        {
            loa_canvas_part_t *part;
            for(part=c->group.first;part;part=part->next){
                abst = rs->vt->abstract.copy(rs->userdata, part->abstract, 0);
                _layer_add_abstract(rs, abst, point,
                                    part->x, part->y, part->xscale, part->yscale,
                                    part->flags | flags, clip);
            }
        }
        break;
    case LOA_CANVAS_ABSTRACT:
        abst = rs->vt->abstract.copy(rs->userdata, c->abstract, 1);
        _layer_add_abstract(rs, abst, point, 0, 0, 1, 1, flags, clip);
        break;
    }
}


static int
_has_frames(read_state_t *rs)
{
    return rs->current_page && rs->current_page->keyframe < loa_buffer_keyframe_count(&rs->buf);
}


static int
_has_duration(read_state_t *rs)
{
    return (double)rs->ticks / rs->mhdr.ticks_per_second > 0.001;
}

static int
_has_layers(read_state_t *rs)
{
    return rs->current_keyframe && rs->current_keyframe->layer < loa_buffer_layer_count(&rs->buf);
}


static void
_ensure_page(read_state_t *rs)
{
    if(rs->current_page)
        return;
    DEBUG_p("page %d", loa_buffer_keyframe_count(&rs->buf));
    rs->current_page = (page_t*)loa_array_append(&rs->page_array, 1);
    rs->current_page->keyframe = loa_buffer_keyframe_count(&rs->buf);
    rs->current_page->name = NULL;
    rs->current_page->layer_count = 0;
    rs->page_time = 0;
}


static void
_ensure_keyframe(read_state_t *rs)
{
    _ensure_page(rs);
    if(rs->current_keyframe)
        return;
    DEBUG_p("keyframe %d", loa_buffer_keyframe_count(&rs->buf));
    rs->current_keyframe = loa_buffer_add_keyframe(&rs->buf);
    rs->current_keyframe->start = rs->page_time;
#if 1
    rs->current_keyframe->duration = (double)rs->ticks / rs->mhdr.ticks_per_second;
#else
    rs->current_keyframe->duration = 0.1;
#endif
    rs->current_keyframe->layer = loa_buffer_layer_count(&rs->buf);
    rs->current_keyframe->anim = loa_buffer_anim_count(&rs->buf);
}


static void
_flush_keyframe(read_state_t *rs)
{
    if(!rs->current_keyframe)
        return;
    if(!_has_layers(rs))
        return;
    rs->current_keyframe->layer_count = loa_buffer_layer_count(&rs->buf) - rs->current_keyframe->layer;
    if(rs->current_keyframe->layer_count > rs->current_page->layer_count)
        rs->current_page->layer_count = rs->current_keyframe->layer_count;
    rs->current_keyframe->anim_count = loa_buffer_anim_count(&rs->buf) - rs->current_keyframe->anim;
    rs->current_keyframe->duration *= rs->loop_expand;
    rs->page_time += rs->current_keyframe->duration;
    rs->current_keyframe = NULL;
}


static void
_flush_page(read_state_t *rs)
{
    if(!rs->current_page)
        return;
    if(!_has_frames(rs))
        return;
    _flush_keyframe(rs);
    rs->current_page->keyframe_count = loa_buffer_keyframe_count(&rs->buf) - rs->current_page->keyframe;
    rs->current_page = NULL;
}


static void
_show_background(read_state_t *rs)
{
    loa_layer_t *layer;
    DEBUG_p("framing mode: %d, is first layer: %d", rs->framing_mode, rs->is_first_layer);
    _ensure_keyframe(rs);
    layer = loa_buffer_add_layer(&rs->buf);
    layer->object = loa_buffer_object_count(&rs->buf);
    if(rs->is_first_layer || rs->framing_mode >= 3){
        if(rs->back.mandatory & 1){
            loa_abstract_t *abst;
            if(!rs->background){
                rs->background = rs->vt->abstract.create(rs->userdata, rs->mhdr.width, rs->mhdr.height);
                rs->vt->abstract.fill(rs->userdata, rs->background, rs->back.red, rs->back.green, rs->back.blue, 1, NULL);
            }
            abst = rs->vt->abstract.copy(rs->userdata, rs->background, 0);
            _layer_add_abstract(rs, abst, 0, 0, 0, 1, 1, LOA_ABSTRACT_TILE, &rs->clip);
            rs->vt->abstract.paste(rs->userdata, rs->scratch, 0, 0, 1, 1, 0, abst, &rs->clip);
        }
        if(rs->back.mandatory >= 2 && rs->back.object_id >= 0){
            object_entry_t *e = _object_table_get_entry(rs->ot, rs->back.object_id);
            if(e->canvas){
                int flags = rs->back.tiling ? LOA_ABSTRACT_TILE : 0;
                loa_canvas_t dst;
                _layer_add_canvas(rs, e->canvas, flags, e->point, &rs->clip);
                dst.type = LOA_CANVAS_ABSTRACT;
                dst.abstract = rs->scratch;
                loa_canvas_paste(&dst, 0, 0, 1, 1, flags, &e->clip, e->canvas, &rs->vt->abstract, rs->userdata);
            }
        }
        rs->is_first_layer = 0;
    }else{
        loa_abstract_t *abst = rs->vt->abstract.copy(rs->userdata, rs->scratch, 1);
        _layer_add_abstract(rs, abst, 0, 0, 0, 1, 1, 0, &rs->clip);
    }
    layer->object_count = loa_buffer_object_count(&rs->buf) - layer->object;
}


static void
_ensure_background(read_state_t *rs)
{
    if(rs->is_first_layer || !_has_layers(rs))
        _show_background(rs);
}


static loa_layer_t *
_begin_layer(read_state_t *rs)
{
    loa_layer_t *layer;
    _ensure_keyframe(rs);
    _ensure_background(rs);
    DEBUG_p("adding layer %d", loa_buffer_layer_count(&rs->buf));
    layer = loa_buffer_add_layer(&rs->buf);
    layer->object = loa_buffer_object_count(&rs->buf);
    return layer;
}


static void
_end_layer(read_state_t *rs)
{
    loa_layer_t *layer = loa_buffer_get_layer(&rs->buf, loa_buffer_layer_count(&rs->buf) - 1);
    layer->object_count = loa_buffer_object_count(&rs->buf) - layer->object;
    if(rs->framing_mode & 1){
        if(rs->current_keyframe && _has_duration(rs))
            _flush_keyframe(rs);
    }
}



static int
_on_chunk(void *userdata, uint32_t chunk_id, const char *data, size_t len)
{
    read_state_t *rs = userdata;
    for(;;){
        switch(rs->state){
        case S_MNG:
            switch(chunk_id){
            case MNG_MHDR:
                if(ni_mng_parse_mhdr(&rs->mhdr, data, len))
                    return -1;
                DEBUG_p("width: %d", rs->mhdr.width);
                DEBUG_p("height: %d", rs->mhdr.height);
                DEBUG_p("ticks per second: %d", rs->mhdr.ticks_per_second);
                DEBUG_p("layers count: %d", rs->mhdr.layer_count);
                DEBUG_p("frame count: %d", rs->mhdr.frame_count);
                DEBUG_p("play time: %d", rs->mhdr.play_time);
                DEBUG_p("simplicity profile: 0x%04x", rs->mhdr.simplicity_profile);
                rs->clip.x = 0;
                rs->clip.y = 0;
                rs->clip.width = rs->mhdr.width;
                rs->clip.height = rs->mhdr.height;
                {
                    loa_point_t *p = loa_buffer_get_point(&rs->buf, 2);
                    p->x = rs->mhdr.width;
                    p->y = rs->mhdr.height;
                }
                rs->scratch = rs->vt->abstract.create(rs->userdata, rs->mhdr.width, rs->mhdr.height);
                break;
            case MNG_MEND:
                break;
            case MNG_LOOP:
                {
                    ni_mng_loop_t loop;
                    if(ni_mng_parse_loop(&loop, data, len))
                        return -1;
                    rs->current_loop = (loop_t*)loa_array_append(&rs->loop_array, 1);
                    rs->current_loop->id = loop.nest_level;
                    rs->current_loop->count = loop.iteration_count;
                    rs->current_loop->start = data + len +4;
#ifdef ENABLE_TWEENING
                    rs->current_loop->can_expand = rs->framing_mode >= 3;
#else
                    rs->current_loop->can_expand = 0;
#endif
                    rs->state = S_LOOP;
                }
                break;
            case MNG_DEFI:
                if(ni_mng_parse_defi(&rs->defi, data, len))
                    return -1;
                {
                    loa_point_t *p;
#if 1
                    object_entry_t *e = _object_table_get_entry(rs->ot, rs->defi.object_id);
                    loa_canvas_destroy(e->canvas, rs->vt->abstract.free, rs->userdata);
                    e->canvas = loa_canvas_create(LOA_CANVAS_GROUP);
                    e->point = loa_buffer_point_count(&rs->buf);
#else
                    loa_canvas_t *c = _object_table_create(rs->ot, rs->defi.object_id, LOA_CANVAS_GROUP, rs);
                    c->point = loa_buffer_point_count(&rs->buf);
#endif
                    p = loa_buffer_add_points(&rs->buf, 3);
                    p[0].x = rs->defi.posx;
                    p[0].y = rs->defi.posy;
                    if(rs->defi.clip_available){
                        p[1].x = rs->defi.clipx1;
                        p[1].y = rs->defi.clipy1;
                        p[2].x = rs->defi.clipx2;
                        p[2].y = rs->defi.clipy2;
                    }else{
                        p[1].x = 0;
                        p[1].y = 0;
                        p[2].x = rs->mhdr.width;
                        p[2].y = rs->mhdr.height;
                    }
                }
                _object_table_apply_defi(rs->ot, &rs->defi, rs);
                break;
            case MNG_IHDR:
                rs->vt->png_begin(rs->userdata);
                rs->vt->png_feed(rs->userdata, PNG_SIGNATURE, 8);
                rs->state = S_PNG;
                continue;
            case MNG_JHDR:
                rs->vt->jng_begin(rs->userdata);
                rs->vt->jng_feed(rs->userdata, JNG_SIGNATURE, 8);
                rs->state = S_JNG;
                continue;
            case MNG_BASI:
                if(ni_mng_parse_basi(&rs->basi, data, len))
                    return -1;
                rs->state = S_BASIS;
                break;
            case MNG_CLON:
                {
                    ni_mng_clon_t clon;
                    object_entry_t *dst;
                    object_entry_t *src;
                    if(ni_mng_parse_clon(&clon, data, len))
                        return -1;
                    dst = _object_table_get_entry(rs->ot, clon.dst_id);
                    loa_canvas_destroy(dst->canvas, rs->vt->abstract.free, rs->userdata);
                    src = _object_table_get_entry(rs->ot, clon.src_id);
                    *dst = *src;
                    switch(clon.clone_type){
                    case 0:
                        dst->canvas = loa_canvas_copy(src->canvas, 1, &rs->vt->abstract, rs->userdata);
                        break;
                    case 1:
                        dst->canvas = loa_canvas_copy(src->canvas, 0, &rs->vt->abstract, rs->userdata);
                        break;
                    case 2:
                        dst->canvas = src->canvas;
                        src->canvas = NULL;
                        break;
                    }
                    {
                        loa_point_t *p;
                        dst->point = loa_buffer_point_count(&rs->buf);
                        p = loa_buffer_add_points(&rs->buf, 3);
                        p[0].x = dst->x;
                        p[0].y = dst->y;
                        p[1].x = dst->clip.x;
                        p[1].y = dst->clip.y;
                        p[2].x = dst->clip.x + dst->clip.width;
                        p[2].y = dst->clip.y + dst->clip.height;
                    }
                    if(clon.hide >= 0)
                        dst->hide = clon.hide;
#if 0
                    if(clon.concrete >= 0)
                        dst->concrete = clon.concrete;
#endif
                    if(clon.loca_delta_type >= 0){
                        loa_anim_t *anim = loa_buffer_add_anim(&rs->buf);
                        anim->move.x = clon.x_location;
                        anim->move.y = clon.y_location;
                        switch(clon.loca_delta_type){
                        case 0:
                            anim->type = ANIM_MOVE_ABSOLUTE;
                            break;
                        case 1:
                            anim->type = ANIM_MOVE_RELATIVE;
                            anim->move.x *= rs->loop_expand;
                            anim->move.y *= rs->loop_expand;
                            break;
                        }
                        anim->point = dst->point;
                    }
                }
            case MNG_DHDR:
                ni_mng_parse_dhdr(&rs->dhdr, data, len);
                switch(rs->dhdr.image_type){
                case 1:
                    rs->vt->png_begin(rs->userdata);
                    rs->vt->png_feed(rs->userdata, PNG_SIGNATURE, 8);
                    rs->state = S_PNG;
                    break;
                case 2:
                    rs->vt->jng_begin(rs->userdata);
                    rs->vt->jng_feed(rs->userdata, JNG_SIGNATURE, 8);
                    rs->state = S_JNG;
                    break;
                }
                break;
            case MNG_PAST:
                {
                    ni_mng_past_t past;
                    object_entry_t *dst;
                    const char *data2;
                    int count;
                    int i;
                    if(ni_mng_parse_past(&past, data, len))
                        return -1;
                    dst = _object_table_get_entry(rs->ot, past.destination_id);
                    count = (len - 11) / 30;
                    for(i=0,data2=data+11;i<count;++i,data2+=30){
                        ni_mng_past_entry_t pe;
                        object_entry_t *src;
                        if(ni_mng_parse_past_entry(&pe, data2, 30))
                            return -1;
                        src = _object_table_get_entry(rs->ot, pe.source_id);
                        _object_entry_paste(dst, src, rs, &past, &pe);
                    }
                }
                break;
            case MNG_MAGN:
                {
                    ni_mng_magn_t magn;
                    int i;
                    ni_mng_parse_magn(&magn, data, len);
                    for(i=magn.first_id;i<=magn.last_id;++i){
                        loa_canvas_t *c = _object_table_get(rs->ot, i);
                        loa_canvas_scale(c, magn.mx, magn.my);
                    }
                }
                break;
            case MNG_BACK:
                {
                    ni_mng_parse_back(&rs->back, data, len);
                    //back.green = 0;
                    //back.blue = 0;
                    if(rs->background)
                        rs->vt->abstract.free(rs->userdata, rs->background);
                    rs->background = NULL;
                }
                break;
            case MNG_FRAM:
                /*
                    Scenario 1:
                      old duration = 0
                      new duration > 0
                      draw no background

                    Scenario 2:
                      old duration = 0
                      new duration = 0
                      draw background

                    Scenario 3:
                      old duration > 0
                      flush keyframe

                */
                {
                    ni_mng_fram_t fram;
                    int old_has_duration = _has_duration(rs);
                    int new_has_duration;
                    ni_mng_parse_fram(&fram, data, len);
                    if(fram.framing_mode != 0)
                        rs->framing_mode = fram.framing_mode;
                    if(fram.interframe_delay.change)
                        rs->ticks = fram.interframe_delay.ticks;
                    if(fram.clipping_boundaries.change){
                        rs->clip.x = fram.clipping_boundaries.left;
                        rs->clip.y = fram.clipping_boundaries.top;
                        rs->clip.width = fram.clipping_boundaries.right - fram.clipping_boundaries.left;
                        rs->clip.height = fram.clipping_boundaries.bottom - fram.clipping_boundaries.top;
                    }
                    if(fram.sync_id_list.change){
                        /* FIXME!! */
                    }
                    new_has_duration = _has_duration(rs);
                    /* If no other layer_exists and the frame duration is non
                     * zero, there should at least be a background.
                     */
                    //if(_has_duration(rs))
                    if(!old_has_duration && !new_has_duration)
                        _ensure_background(rs);
                    /* Sometimes a FRAM chunk apears in a keyframe with zero
                     * duration to set the actual duration, without starting
                     * a new keyframe.
                     */
                    if(rs->current_keyframe)
                        rs->current_keyframe->duration = (double)rs->ticks / rs->mhdr.ticks_per_second;
                    if(old_has_duration)
                        _flush_keyframe(rs);
                    rs->is_first_layer = 1;
                }
                break;
            case MNG_MOVE:
                {
                    ni_mng_move_t move;
                    int i;
                    ni_mng_parse_move(&move, data, len);
                    _ensure_keyframe(rs);
                    for(i=move.first;i<=move.last;++i){
                        object_entry_t *e = _object_table_get_entry(rs->ot, i);
                        loa_anim_t *anim;
                        anim = loa_buffer_add_anim(&rs->buf);
                        anim->move.x = move.x;
                        anim->move.y = move.y;
                        switch(move.type){
                        case NI_MNG_MOVE_ABSOLUTE:
                            anim->type = ANIM_MOVE_ABSOLUTE;
                            break;
                        case NI_MNG_MOVE_RELATIVE:
                            anim->type = ANIM_MOVE_RELATIVE;
                            anim->move.x *= rs->loop_expand;
                            anim->move.y *= rs->loop_expand;
                            break;
                        }
                        anim->point = e->point;
                    }
                    _object_table_apply_move(rs->ot, &move);
                }
                break;
            case MNG_CLIP:
                {
                    ni_mng_clip_t clip;
                    int i;
                    ni_mng_parse_clip(&clip, data, len);
                    _ensure_keyframe(rs);
                    for(i=clip.first;i<=clip.last;++i){
                        object_entry_t *e = _object_table_get_entry(rs->ot, i);
                        loa_anim_t *anim;
                        anim = loa_buffer_add_anim(&rs->buf);
                        anim->clip.x1 = clip.left_cb;
                        anim->clip.y1 = clip.top_cb;
                        anim->clip.x2 = clip.right_cb;
                        anim->clip.y2 = clip.bottom_cb;
                        switch(clip.type){
                        case 0:
                            anim->type = ANIM_CLIP_ABSOLUTE;
                            break;
                        case 1:
                            anim->type = ANIM_CLIP_RELATIVE;
                            anim->clip.x1 *= rs->loop_expand;
                            anim->clip.y1 *= rs->loop_expand;
                            anim->clip.x2 *= rs->loop_expand;
                            anim->clip.y2 *= rs->loop_expand;
                            break;
                        }
                        anim->point = e->point + 1;
                    }
                    _object_table_apply_clip(rs->ot, &clip);
                }
                break;
            case MNG_SHOW:
                {
                    ni_mng_show_t show;
                    int i;
                    ni_mng_parse_show(&show, data, len);
                    //DEBUG_p("show objects %d - %d", show.first, show.last);
                    for(i=show.first;i<=show.last;++i){
                        object_entry_t *e = _object_table_get_entry(rs->ot, i);
                        e->hide = 0; /* Is this correct and needed? */
                        _object_entry_show(e, rs);
                    }
                }
                //_end_layer(&rs);
                break;
            case MNG_SEEK:
                _flush_page(rs);
                _ensure_page(rs);
                rs->current_page->name = strndup(data, len);
                break;
            }
            break;
        case S_PNG:
        case S_JNG:
        case S_BASIS:
            switch(rs->state){
            case S_MNG:
            case S_LOOP:
                /* not reached */
                break;
            case S_PNG:
                rs->vt->png_feed(rs->userdata, data - 8, len + 12);
                break;
            case S_JNG:
                rs->vt->jng_feed(rs->userdata, data - 8, len + 12);
                break;
            case S_BASIS:
                break;
            }
            switch(chunk_id){
            case MNG_IEND:
                {
                    loa_abstract_t *abst;
                    switch(rs->state){
                    case S_MNG:
                    case S_LOOP:
                        /* not reached */
                        break;
                    case S_PNG:
                        abst = rs->vt->png_end(rs->userdata);
                        break;
                    case S_JNG:
                        abst = rs->vt->jng_end(rs->userdata);
                        DEBUG_p("loaded JNG image %p", abst);
                        break;
                    case S_BASIS:
                        abst = rs->vt->abstract.create(rs->userdata, rs->basi.width, rs->basi.height);
                        rs->vt->abstract.fill(rs->userdata, abst, rs->basi.red, rs->basi.green, rs->basi.blue, rs->basi.alpha, &rs->clip);
                        break;
                    }
                    if(!abst){
                        DEBUG_p("failed to load sub image");
                    }
                    if(rs->dhdr.object_id >= 0){
                        DEBUG_p("delta type: %d", rs->dhdr.delta_type);
                        if(rs->dhdr.delta_type < 7){
                            loa_canvas_t *dst;
                            int flags;
                            loa_canvas_t src;
                            dst = _object_table_copy(rs->ot, rs->dhdr.object_id, 1, rs);
                            switch(rs->dhdr.delta_type){
                            case 0:
                                flags = 0;
                                break;
                            case 1:
                                flags = LOA_ABSTRACT_ADD;
                                break;
                            case 2:
                                flags = LOA_ABSTRACT_MASK_ALPHA | LOA_ABSTRACT_ADD;
                                break;
                            case 3:
                                flags = LOA_ABSTRACT_MASK_COLOR | LOA_ABSTRACT_ADD;
                                break;
                            case 4:
                                flags = 0;
                                break;
                            case 5:
                                flags = LOA_ABSTRACT_MASK_ALPHA;
                                break;
                            case 6:
                                flags = LOA_ABSTRACT_MASK_COLOR;
                                break;
                            }
                            src.type = LOA_CANVAS_ABSTRACT;
                            src.abstract = abst;
                            loa_canvas_paste(dst, rs->dhdr.x_location, rs->dhdr.y_location, 1, 1, flags, &rs->clip, &src, &rs->vt->abstract, rs->userdata);
                        }
                        rs->vt->abstract.free(rs->userdata, abst);
                        _object_table_maybe_show(rs->ot, rs->dhdr.object_id, rs);
                        rs->dhdr.object_id = -1;
                    }else if(rs->defi.object_id >= 0){
                        loa_canvas_t *c = _object_table_create(rs->ot, rs->defi.object_id, LOA_CANVAS_ABSTRACT, rs);
                        c->abstract = abst;
                        _object_table_maybe_show(rs->ot, rs->defi.object_id, rs);
                    }else{
                        _begin_layer(rs);
                        _layer_add_abstract(rs, abst, 0, 0, 0, 1, 1, 0, NULL);
                        _end_layer(rs);
                        rs->vt->abstract.paste(rs->userdata, rs->scratch, 0, 0, 1, 1, 0,
                                               abst, &rs->clip);
                    }
                    rs->state = S_MNG;
                }
                break;
            }
            break;
        case S_LOOP:
            switch(chunk_id){
            case MNG_ENDL:
                {
                    ni_mng_endl_t endl;
                    if(ni_mng_parse_endl(&endl, data, len))
                        return -1;
                    if(rs->current_loop->id == endl.nest_level){
                        size_t loop_len = data - rs->current_loop->start - 8;
                        int i;
                        rs->state = S_MNG;
                        if(rs->current_loop->count > 1 && rs->current_loop->can_expand){
                            DEBUG_p("expanding loop %d times", rs->current_loop->count);
                            rs->loop_expand = rs->current_loop->count;
                            if(ni_xng_parse_chunks(rs->current_loop->start, loop_len, 0, _on_chunk, userdata))
                                return -1;
                            rs->loop_expand = 1;
#ifdef ENABLE_TWEENING
                            rs->scalable_time = 1;
#endif
                        }else{
                            DEBUG_p("running loop %d times", rs->current_loop->count);
                            for(i=0;i<rs->current_loop->count;++i)
                                if(ni_xng_parse_chunks(rs->current_loop->start, loop_len, 0, _on_chunk, userdata))
                                    return -1;
                            DEBUG_p("after loop");
                        }
                        loa_array_pop(&rs->loop_array);
                    }
                }
                break;
#ifdef ENABLE_TWEENING
            case MNG_MOVE:
                if(rs->current_loop->can_expand){
                    ni_mng_move_t move;
                    if(ni_mng_parse_move(&move, data, len))
                        return -1;
                    if(NI_MNG_MOVE_RELATIVE == move.type && (ABS(move.x) == rs->clip.width || ABS(move.y) == rs->clip.height))
                        rs->current_loop->can_expand = 0;
                }
                break;
#endif
            }
            break;
        }
        break;
    }
    return 0;
}


ni_mng_t *
ni_mng_decode(const char *data, size_t len, const ni_mng_callbacks_t *vt, void *userdata)
{
    read_state_t rs;
    ni_mng_t *mng;
    int res;

    rs.state = S_MNG;
#ifdef ENABLE_TWEENING
    rs.scalable_time = 0;
#endif

    rs.current_page = NULL;
    rs.current_keyframe = NULL;
    rs.loop_expand = 1;
    rs.framing_mode = 1;
    rs.ticks = 1;
    rs.is_first_layer = 1;
    //rs.has_background = 0;
    rs.scratch =  NULL;
    rs.background = NULL;

    rs.ot = malloc(sizeof(object_table_t));
    _object_table_init(rs.ot);
    loa_buffer_init(&rs.buf, sizeof(loa_object_t), sizeof(loa_layer_t), sizeof(loa_anim_t), sizeof(loa_keyframe_t));
    loa_array_init(&rs.page_array, sizeof(page_t));
    loa_array_init(&rs.loop_array, sizeof(loop_t));
    rs.defi.object_id = -1;
    rs.dhdr.object_id = -1;
    rs.back.mandatory = 0;

    /* The first point is always (0, 0), clipping region is the whole screen. */
    {
        loa_point_t *p  = loa_buffer_add_points(&rs.buf, 3);
        p[0].x = 0;
        p[0].y = 0;
        p[1].x = 0;
        p[1].y = 0;
    }

    rs.vt = vt;
    rs.userdata = userdata;

    res = ni_mng_parse(data, len, 0, _on_chunk, &rs);
    _flush_page(&rs);

    if(rs.scratch)
        vt->abstract.free(userdata, rs.scratch);
    if(rs.background)
        vt->abstract.free(userdata, rs.background);

    DEBUG_p("points: %d", loa_buffer_point_count(&rs.buf));
    DEBUG_p("objects: %d", loa_buffer_object_count(&rs.buf));
    DEBUG_p("layers: %d", loa_buffer_layer_count(&rs.buf));
    DEBUG_p("frames: %d", loa_buffer_keyframe_count(&rs.buf));
    DEBUG_p("pages: %d", loa_array_length(&rs.page_array));
    DEBUG_p("loops: %d", loa_array_length(&rs.loop_array));

    if(res){
        int i;
        mng = NULL;
        for(i=0;i<loa_buffer_object_count(&rs.buf);++i)
            vt->abstract.free(mng->userdata, loa_buffer_get_object(&rs.buf, i)->abstract);
        for(i=0;i<loa_array_length(&rs.page_array);++i){
            page_t *page = (page_t*)loa_array_get(&rs.page_array, i);
            if(page->name)
                free(page->name);
        }
    }else{
        mng = malloc(sizeof(ni_mng_t));
        mng->free = vt->abstract.free;
        mng->userdata = userdata;
        mng->width = rs.mhdr.width;
        mng->height = rs.mhdr.height;
#ifdef ENABLE_TWEENING
        mng->scalable_time = rs.scalable_time;
#endif
        //mng->ticks_per_second = get32(data+8);
        //mng->frame_count = get32(data+12);
        mng->point = loa_buffer_copy_points(&rs.buf);
        mng->point_count = loa_buffer_point_count(&rs.buf);
        mng->object = loa_buffer_copy_objects(&rs.buf);
        mng->object_count = loa_buffer_object_count(&rs.buf);
        mng->layer = loa_buffer_copy_layers(&rs.buf);
        mng->layer_count = loa_buffer_layer_count(&rs.buf);
        mng->anim = loa_buffer_copy_anims(&rs.buf);
        mng->keyframe = loa_buffer_copy_keyframes(&rs.buf);
        mng->keyframe_count = loa_buffer_keyframe_count(&rs.buf);
        mng->page = (page_t*)loa_array_copy(&rs.page_array);
        mng->page_count = loa_array_length(&rs.page_array);
    }
    loa_buffer_done(&rs.buf);
    loa_array_done(&rs.page_array);
    loa_array_done(&rs.loop_array);
    _object_table_done(rs.ot, &rs);
    free(rs.ot);
    return mng;
}


void ni_mng_free(ni_mng_t *mng)
{
    int i;
    free(mng->point);
    for(i=0;i<mng->object_count;++i)
        mng->free(mng->userdata, mng->object[i].abstract);
    free(mng->layer);
    free(mng->anim);
    free(mng->keyframe);
    for(i=0;i<mng->page_count;++i){
        page_t *page = &mng->page[i];
        if(page->name)
            free(page->name);
    }
    free(mng->page);
    free(mng);
}


int ni_mng_get_width(ni_mng_t *mng)
{
    return mng->width;
}


int ni_mng_get_height(ni_mng_t *mng)
{
    return mng->height;
}


int ni_mng_get_scalable_time(ni_mng_t *mng)
{
#ifdef ENABLE_TWEENING
    return mng->scalable_time;
#else
    return 0;
#endif
}


int ni_mng_get_page_count(ni_mng_t *mng)
{
    return mng->page_count;
}


const char *
ni_mng_get_page_name(ni_mng_t *mng, int pagenr)
{
    return mng->page[pagenr].name;
}


int ni_mng_get_layer_count(ni_mng_t *mng, int pagenr)
{
    return mng->page[pagenr].layer_count;
}


int ni_mng_get_frame_count(ni_mng_t *mng, int pagenr)
{
    return mng->page[pagenr].keyframe_count;
}


double ni_mng_get_frame_duration(ni_mng_t *mng, int pagenr, int frame)
{
    page_t *page = &mng->page[pagenr];
    return mng->keyframe[page->keyframe + frame].duration;
}


static int
_frame_from_time(const ni_mng_t *mng, page_t *page, double time)
{
    int frame;
    if(page->keyframe_count <= 1)
        return 0;
    for(frame=0;;){
        double duration = mng->keyframe[page->keyframe + frame].duration;
        if(time < duration)
            return frame;
        time -= duration;
        if(++frame >= page->keyframe_count)
            frame = 0;
    }
}


#ifdef _USE_DEBUG
static void
_dump_page(const ni_mng_t *mng, page_t *page)
{
    int i;
    for(i=0;i<page->keyframe_count;++i){
        loa_keyframe_t *kf = &mng->keyframe[page->keyframe + i];
        DEBUG_p("keyframe %d - %f, %f seconds", i, kf->start, kf->duration);
    }
}
#endif


static void
_apply_anim(loa_point_t *point, loa_anim_t *anim, double progress)
{
    loa_point_t *p = &point[anim->point];
    switch(anim->type){
    case ANIM_MOVE_ABSOLUTE:
        p[0].x = anim->move.x;
        p[0].y = anim->move.y;
        break;
    case ANIM_MOVE_RELATIVE:
        p[0].x += anim->move.x * progress;
        p[0].y += anim->move.y * progress;
        break;
    case ANIM_CLIP_ABSOLUTE:
        p[0].x = anim->clip.x1;
        p[0].y = anim->clip.y1;
        p[1].x = anim->clip.x2;
        p[1].y = anim->clip.y2;
        break;
    case ANIM_CLIP_RELATIVE:
        p[0].x += (double)anim->clip.x1 * progress;
        p[0].y += (double)anim->clip.y1 * progress;
        p[1].x += (double)anim->clip.x2 * progress;
        p[1].y += (double)anim->clip.y2 * progress;
        break;
    }
}


int
ni_mng_draw(const ni_mng_t *mng, int pagenr, int *layer_visibility, double time, ni_mng_draw_func_t draw, void *userdata)
{
    page_t *page = &mng->page[pagenr];
    int frame = _frame_from_time(mng, page, time);
    //frame = 1;
    loa_keyframe_t *kf = &mng->keyframe[page->keyframe + frame];
    loa_point_t *point;
#ifdef ENABLE_TWEENING
    double progress = mng->scalable_time ? (time - kf->start) / kf->duration : 1;
#else
    static const double progress = 1;
#endif
    int i;

#ifdef _USE_DEBUG
    _dump_page(mng, page);
#endif

    DEBUG_p("page: %d, frame: %d", pagenr, frame);

    point = (loa_point_t*)malloc(mng->point_count * sizeof(loa_point_t));
    memcpy(point, mng->point, mng->point_count * sizeof(loa_point_t));

    /* Apply complete animations */
    for(i=0;i<kf->anim;++i)
        _apply_anim(point, &mng->anim[i], 1);

    /* Apply partial animations */
    for(i=0;i<kf->anim_count;++i)
        _apply_anim(point, &mng->anim[kf->anim + i], progress);

    for(i=0;i<kf->layer_count;++i){
        loa_layer_t *layer = &mng->layer[kf->layer + i];
        int j;
        if(!layer_visibility[i])
            continue;
        DEBUG_p("drawing layer %d", i);
        for(j=0;j<layer->object_count;++j){
            loa_object_t *o = &mng->object[layer->object + j];
            loa_point_t *p = &point[o->point];
            nil_clip_t clip;
            clip.x = p[1].x;
            clip.y = p[1].y;
            clip.width = p[2].x - clip.x;
            clip.height = p[2].y - clip.y;
            nil_clip_intersect(&clip, &o->clip);
            draw(userdata, p->x + o->x, p->y + o->y, o->xscale, o->yscale, o->flags, o->abstract, &clip);
        }
    }
    return 0;
}
