/*
 * Copyright (C) 2018 - 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <stdlib.h>
#include <string.h>

#include "converter.h"


#define ENABLE_BMP
//#define ENABLE_FITS /* Upside down */
//#define ENABLE_GEM /* Wrong palette. */
#define ENABLE_GIF
#define ENABLE_ILBM
#define ENABLE_JPEG
#define ENABLE_JPEG2000
#define ENABLE_MACPAINT
#define ENABLE_MTV
//#define ENABLE_NEOCHROME /* Incorrectly decoded. */
#define ENABLE_PCX
//#define ENABLE_PHOTOCD /* Application not included with netpbm by default. */
#define ENABLE_SGI
#define ENABLE_SPECTRUM512
//#define ENABLE_WORKBENCHICON /* Incorrectly decoded. */

/*

Formats supported by netpbm:

  411     | Sony Mavic camera
  ascii   | ASCII
  atk     | Andrew Toolkit raster object
  avs     | AVS X image
  biorad  | Biorad confocal
* bmp     | Windows BMP or OS/2 BMP
  brush   | Doodle brush
  camera  | Camera raw image
  cis     | Compuserve RLE image
  cmuwm   | CMU window manager bitmap
  ddbug   | Diddle or DiddleBug sketch
  escp2   | ESC/P2 printer file
  eyuv    | Berkley YUV
  fits    | FITS
  fs      | Usenix FaceSaver
  g3      | Group 3 fax file
* gem     | GEM .img
* gif     | Compuserve GIF
  gould   | Gould scanner file
  hdiff   | Horizontal difference image
  hips    | Hunan Informational Processing Library HIPS
* hpcd    | Photo-CD image
* ilbm    | Interlaced Bitmap
  img     | Img-whatnot
* info    | Amiga .info icon
* jpek2k
* jpeg
  leaf    | Interleaf image
  lispm   | Lisp Machine bitmap
* macp    | MacPaint
  mda     | Microdesign .mda or .mdp
  mgr     | MGR window manager bitmap
* mtv     | MTV or PRT ray tracer output
* neo     | Neochrome
  palm    | Palm bitmap
  pc1
  pcdov   | Photo CD overview
* pcx
  pdbimg  | Palm Pilot PDB image
  pfm     | Portable Float Map
  pi1
  pi3
  pict    | Macintosh PICT
  pj      | HP PaintJet
  pk      | packed format font
  png
  psid    | PostScript image data
  ps
  qrt     | QRT raytracer output
  rast    | Sun resterfile
  rla     | Alias/Wavefront RLA and RPF image
  sbig    | Santa Barbera Instrument Group astronomical CCD camera file (CCDOPS)
  sgi     | SGI image
  sir     | Solitair Image Recorder file
  sld     | AutoCAD slide
* spc     | compressed spectrum file
  spot    | SPOT satellite image
* spu     | uncompressed spectrum file
  srf     | SRF image
  st4     | Santa Barbera Instrument Group astronimical ST-4 camera file
  sunicon | Sun icon
  svg     |
  tga     |
  thinkjet| HP ThinkJet printer file
  tiff    |
  wbmp    | wireless bitmap
  winicon | Windows icon
  xbm     | X10 or X11 bitmap
  xpm     | X11 pixmap
  xvmini  | XV thumbnail
  xwd     | X10 or X11 window dump
  ybm     | Bennet Yee face file
  zeiss   | Zeiss confocal file
*/


#define ATARI_PIXEL_RATIO (5.0/6)


static const char **
_netpbm_conv_get_mime_types()
{
    static const char *mime_types[] = {
#ifdef ENABLE_BMP
        "image/bmp",
#endif
#ifdef ENABLE_FITS
        "application/fits", "image/fits",
#endif
#ifdef ENABLE_GEM
        "image/x-gem",
#endif
#ifdef ENABLE_GIF
        "image/gif",
#endif
#ifdef ENABLE_ILBM
        "image/x-ilbm",
#endif
#ifdef ENABLE_JPEG2000
        "image/jp2", "image/jpeg2000", "image/jpx",
#endif
#ifdef ENABLE_MACPAINT
        "application/mac-paint", "image/x-macpaint",
#endif
#ifdef ENABLE_MTV
        "image/x-mtv",
#endif
#ifdef ENABLE_NEOCHROME
        "image/x-neo", "image/x-neochrome",
#endif
#ifdef ENABLE_PC1
        "image/x-pc1",
#endif
#ifdef ENABLE_PC2
        "image/x-pc2",
#endif
#ifdef ENABLE_PC3
        "image/x-pc3",
#endif
#ifdef ENABLE_PCX
        "image/vnd.zbrush.pcx",
#endif
#ifdef ENABLE_PHOTOCD
        "image/x-photo-cd",
#endif
#ifdef ENABLE_SGI
        "image/x-rgb", "image/x-sgi",
#endif
#ifdef ENABLE_SPECTRUM512
        "image/x-spectrum512-compressed",
        "image/x-spectrum512-uncompressed",
#endif
#ifdef ENABLE_WORKBENCHICON
        "image/x-amiga-icon",
#endif
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_netpbm_conv_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    const char *converter;
    const char *out_mime_type = "image/x-portable-arbitrarymap";
    h->info = info;
    h->info->pixel_ratio = 1;
#ifdef ENABLE_BMP
    if(!strcasecmp(mime_type,"image/bmp")){
        converter = "bmptopnm";
    }else
#endif
#ifdef ENABLE_FITS
    if(!strcasecmp(mime_type,"application/fits") ||
       !strcasecmp(mime_type,"image/fits"))
    {
        converter = "fitstopnm";
    }else
#endif
#ifdef ENABLE_GEM
    if(!strcasecmp(mime_type,"image/x-gem")){
        converter = "gemtopnm";
    }else
#endif
#ifdef ENABLE_GIF
    if(!strcasecmp(mime_type,"image/gif")){
        converter = "giftopnm";
    }else
#endif
#ifdef ENABLE_ILBM
    if(!strcasecmp(mime_type,"image/x-ilbm")){
        converter = "ilbmtoppm";
    }else
#endif
#ifdef ENABLE_JPEG2000
    if(!strcasecmp(mime_type,"image/jp2") ||
       !strcasecmp(mime_type,"image/jpeg2000") ||
       !strcasecmp(mime_type,"image/jpx"))
    {
        converter = "jpeg2ktopam";
    }else
#endif
#ifdef ENABLE_MACPAINT
    if(!strcasecmp(mime_type,"application/mac-paint") ||
       !strcasecmp(mime_type,"image/x-macpaint"))
    {
        converter = "macptopbm";
    }else
#endif
#ifdef ENABLE_MTV
    if(!strcasecmp(mime_type,"image/x-mtv")){
        converter = "mtvtoppm";
    }else
#endif
#ifdef ENABLE_NEOCHROME
    if(!strcasecmp(mime_type,"image/x-neo") ||
       !strcasecmp(mime_type,"image/x-neochrome"))
    {
        converter = "neotoppm";
        h->info->pixel_ratio = ATARI_PIXEL_RATIO;
    }else
#endif
#ifdef ENABLE_PC1
    if(!strcasecmp(mime_type,"image/x-pc1")){
        converter = "pc1toppm";
        h->info->pixel_ratio = ATARI_PIXEL_RATIO;
    }else
#endif
#ifdef ENABLE_PI1
    if(!strcasecmp(mime_type,"image/x-pi1")){
        converter = "pi1toppm";
        h->info->pixel_ratio = ATARI_PIXEL_RATIO;
    }else
#endif
#ifdef ENABLE_PCX
    if(!strcasecmp(mime_type,"image/vnd.zbrush.pcx")){
        converter = "pcxtoppm";
    }else
#endif
#ifdef ENABLE_PHOTOCD
    if(!strcasecmp(mime_type,"image/x-photo-cd")){
        converter = "hpcdtoppm";
    }else
#endif
#ifdef ENABLE_SGI
    if(!strcasecmp(mime_type,"image/x-rgb") ||
       !strcasecmp(mime_type,"image/x-sgi"))
    {
        converter = "sgitopnm";
    }else
#endif
#ifdef ENABLE_SPECTRUM512
    if(!strcasecmp(mime_type,"image/x-spectrum512-compressed")){
        converter = "spctoppm";
        h->info->pixel_ratio = ATARI_PIXEL_RATIO;
    }else if(!strcasecmp(mime_type,"image/x-spectrum512-uncompressed")){
        converter = "sputoppm";
        h->info->pixel_ratio = ATARI_PIXEL_RATIO;
    }else
#endif
#ifdef ENABLE_WORKBENCHICON
    if(!strcasecmp(mime_type,"image/x-amiga-icon")){
        converter = "infotopam";
    }else
#endif
    {
        free(h);
        return NULL;
    }
    if(!abydos_converter_init(h, converter, out_mime_type, 0)){
        free(h);
        return NULL;
    }
    info->threadsafe = 1;
    return h;
}


static void
_netpbm_conv_free(abydos_plugin_handle_t *h)
{
    abydos_converter_done(h);
    free(h);
}


const abydos_plugin_interface_t *
Abydos0(){
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _netpbm_conv_get_mime_types,
        .new = _netpbm_conv_new,
        .free = _netpbm_conv_free,
        .progressive_begin = abydos_converter_begin,
        .progressive_feed = abydos_converter_feed,
        .progressive_end = abydos_converter_end,
        .get_image_surface = abydos_converter_get_image_surface
    };
    return &p;
}
