/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
/*

    A file format that could represent a static image or multiple static images or an animation.
    JPEG or H.264 raw video can be used for data compression.

    File Structure Overview
    ------------------------
    'T''A''P''0'
    ------------------------
    width
    ------------------------
    height
    ------------------------
    codec
    ------------------------
    nimage
    ------------------------
    delayms[0]
    delayms[1]
      :
      :
    delayms[nimage-1]
    ------------------------
    nchunk
    ------------------------
    chunkid[0]
    chunksize[0]
    chunkid[1]
    chunksize[1]
      :
      :
    chunkid[nchunk-1]
    chunksize[nchunk-1]
    ------------------------
    chunkdata[0] (padding zero to 4 bytes)
    chunkdata[1] (padding zero to 4 bytes)
      :
      :
    chunkdata[nchunk-1] (padding zero to 4 bytes)
    ------------------------

    'T''A''P''0'
    MAGIC flag, TAP document

    width
    image width, signed int little endian

    height
    image height, signed int little endian

    codec
    compression codec used, signed int little endian
    can be one of the following
    0 TAP_CODEC_NONE
    undefined codec
    1 TAP_CODEC_JPEG
    JPEG image
    2 TAP_CODEC_H264
    H264 raw video

    nimage
    the count of images or the count of frames in animation, signed int little endian

    delayms array
    frame delay duration in ms, signed int little endian
    can be one of the following
    0 static image or skipped/hidden frame
    >0 delay duration in ms

    nchunk
    the count of followed data chunks, signed int little endian

    chunkid chunksize pair array
    chunk id and corresponding chunk size, signed int little endian
    chunk id starts from 0

    chunkdata array
    chunk data which length is chunksize, binary
    padding zero at the end of each chunk data to meet the alignment of 4 bytes

*/
#include <abydos.h>
#include <abydos-plugin.h>
#include <debug.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>


typedef struct{
    cairo_surface_t *surface;
    double duration;
}multi_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    enum{
        MT_PAGE,
        MT_FRAME
    }multi_type;
    multi_t *multi;
};

typedef struct{
    int frame;
    abydos_t *ar;
}stream_t;


static const char **
_tap_get_mime_types()
{
    static const char *mime_types[] = {
        "image/vnd.tencent.tap",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_tap_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    info->threadsafe = 1;
    h->multi = NULL;

    return h;
}


static void
_tap_free(abydos_plugin_handle_t *h)
{
    if(h->multi){
        int count = 0; /* Just to shut up the compiler. */
        int i;
        switch(h->multi_type){
        case MT_PAGE:
            count = h->info->page_count;
            break;
        case MT_FRAME:
            count = h->info->frame_count;
            break;
        }
        for(i=0;i<count;++i)
            if(h->multi[i].surface)
                cairo_surface_destroy(h->multi[i].surface);
        free(h->multi);
    }
    free(h);
}


#define get32(a) le32toh(*(uint32_t*)(a))


static int
_tap_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    int result = -1;
    int codec;
    int multi_count;
    size_t len2;
    int frame_count;
    stream_t *stream = NULL;
    int chunk_count;
    const char *chunkp;
    int i;
    if(len < 20)
        return -1;
    if(strncmp(data,"TAP0",4))
        return -1;
    codec = get32(data+12);
    if(codec != 1){
        DEBUG_p("unhandled format");
        return -1;
    }
    h->info->width = get32(data+4);
    h->info->height = get32(data+8);
    multi_count = get32(data+16);
    data += 20;
    len -= 20;

    len2 = multi_count*4+4;
    if(len < len2)
        return -1;
    stream = (stream_t*)malloc(multi_count*sizeof(stream_t));
    frame_count = 0;
    for(i=0;i<multi_count;++i)
        if(get32(data+i*4)>0)
            ++frame_count;
    if(frame_count > 0){
        /* this is an animation */
        h->info->frame_count = frame_count;
        h->multi_type = MT_FRAME;
        h->multi = (multi_t*)malloc(frame_count*sizeof(multi_t));
        frame_count = 0;
        for(i=0;i<multi_count;++i){
            int delay = get32(data+i*4);
            if(delay > 0){
                h->multi[frame_count].surface = NULL;
                h->multi[frame_count].duration = (double)delay/1000;
                stream[i].frame = frame_count++;
                stream[i].ar = abydos_create("image/jpeg");
                abydos_load_begin(stream[i].ar,NULL,NULL,NULL);
            }else{
                stream[i].frame = -1;
                stream[i].ar = NULL;
            }
        }
    }else{
        /* this is a set of pages */
        h->info->page_count = multi_count;
        h->multi_type = MT_PAGE;
        h->multi = (multi_t*)calloc(multi_count,sizeof(multi_t));
    }
    data += len2;
    len -= len2;

    chunk_count = get32(data-4);
    len2 = chunk_count*8;
    if(len2 > len)
        goto error;

    chunkp = data+len2;
    for(i=0;i<chunk_count;++i)
        chunkp += (get32(data+i*8+4)+3) & ~3;

    if(chunkp > data+len)
        goto error;

    chunkp = data+len2;
    for(i=0;i<chunk_count;++i){
        int id = get32(data+i*8+0);
        size_t size = get32(data+i*8+4);
        if(stream[id].frame >= 0){
            if(abydos_load_feed(stream[id].ar,chunkp,size)<0)
                goto error;
        }
        chunkp += (size+3) & ~3;
    }

    for(i=0;i<multi_count;++i){
        if(stream[i].frame < 0)
            continue;
        if(abydos_load_end(stream[i].ar)<0)
            goto error;
        h->multi[stream[i].frame].surface = abydos_get_image_surface(stream[i].ar);
    }
    result = 0;
error:
    if(stream){
        for(i=0;i<multi_count;++i)
            if(stream[i].ar)
                abydos_destroy(stream[i].ar);
        free(stream);
    }
    return result;
}


static cairo_surface_t *
_tap_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    switch(h->multi_type){
    case MT_PAGE:
        return cairo_surface_reference(h->multi[page].surface);
    case MT_FRAME:
        return cairo_surface_reference(h->multi[frame].surface);
    }
    /* not reached */
    return NULL;
}


static void
_tap_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->multi[i].duration;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _tap_get_mime_types,
        .new = _tap_new,
        .free = _tap_free,
        .create_from_data = _tap_create_from_data,
        .get_image_surface = _tap_get_image_surface,
        .get_duration = _tap_get_duration
    };
    return &p;
}
