/*
 * Copyright (C) 2019 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <ctype.h>
#include <debug.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>

#include "cebra.h"
#include "cons_buf.h"
#include "t42.h"
#include "teletext_cons.h"
#include "teletext_dec.h"
#include "tti7.h"

/* #define TEST_RENDER */

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    Console *cons;
    enum{
        T_CEBRA,
        T_BBCRAM,
        T_SOFTEL,
        T_T42,
        T_TTI7,
        T_RAW
    }type;
#ifndef TEST_RENDER
    cairo_surface_t *surface[2];
#endif
};


static int
_atoin(const char* s,size_t len)
{
    const char *end = s+len;
    int neg;
    int res;
    while(s<end && isspace(*s))
        ++s;
    if(s>=end)
        return 0;
    if('-'==*s){
        neg = 1;
        ++s;
    }else{
        neg = 0;
    }
    res = 0;
    while(s<end && *s>='0' && *s<='9'){
        res *= 10;
        res += *s-'0';
        ++s;
    }
    return neg ? -res : res;
}


static const char **
_tt_get_mime_types()
{
    static const char *mime_types[] = {
        "application/x.teletext.ttx",
        "image/x-bbc-micro-screendump",
        "text/x-softel-teletext",
        "text/x-t42-packets",
        "text/x.teletext.tti",
        "text/x-raw-teletext",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_tt_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    if(!strcasecmp(mime_type, "application/x.teletext.ttx")){
        h->type = T_CEBRA;
    }else if(!strcasecmp(mime_type, "image/x-bbc-micro-screendump")){
        h->type = T_BBCRAM;
    }else if(!strcasecmp(mime_type, "text/x-softel-teletext")){
        h->type = T_SOFTEL;
    }else if(!strcasecmp(mime_type, "text/x-t42-packets")){
        h->type = T_T42;
    }else if(!strcasecmp(mime_type, "text/x.teletext.tti")){
        h->type = T_TTI7;
    }else{
        h->type = T_RAW;
    }
    h->cons = cons_buf_new();
    cons_buf_set_size(h->cons,TT_SCREEN_WIDTH,TT_SCREEN_HEIGHT,0);
    h->surface[0] = NULL;
    h->surface[1] = NULL;
    return h;
}


static void
_tt_free(abydos_plugin_handle_t *h)
{
    int i;
    cons_buf_free(h->cons);
    for(i=0;i<h->info->frame_count;++i)
        if(h->surface[i])
            cairo_surface_destroy(h->surface[i]);
    free(h);
}


static int
_on_cebra_data(void *userdata, int level, int type, const uint8_t *data, size_t len)
{
    TeletextDecoder *dec = (TeletextDecoder*)userdata;
    switch(level){
#if 0
    case 6: /* header */
        switch(type){
        case 2:
            DEBUG_p("line number size: %d, line: %d", (int)len, data[0]);
            teletext_decoder_goto(dec, 0, 0);
            break;
        case 4:
            DEBUG_p("--- headers ---");
            break;
        }
        break;
#endif
    case 8: /* body */
        DEBUG_p("type %d",type);
        switch(type){
        case 2:
            DEBUG_p("line number size: %d, line: %d", (int)len, data[0]);
            teletext_decoder_goto(dec, 0, data[0]);
            break;
        case 4:
            DEBUG_p("--- body row ---");
            while(len-->0)
                teletext_decoder_write(dec, *data++);
            break;
        }
        break;
    }
    return 0;
}


static int
_on_t42_packet(void *userdata, int page, int row, uint8_t *data)
{
    TeletextDecoder *dec = (TeletextDecoder*)userdata;
    int i;
    teletext_decoder_goto(dec, 0, row);
    for(i=0;i<40;++i)
        teletext_decoder_write(dec, data[i]);
    return 0;
}


static int
_on_tti7_data(void *userdata, tti7_line_t line, int num, str_t * str)
{
    TeletextDecoder *dec = (TeletextDecoder*)userdata;
    const char *p;
    const char *end;
    switch(line){
    case TTI7_OL:
        if(num < 2)
            break;
        teletext_decoder_goto(dec, 0, _atoin(str[0].str, str[0].len)-1);
        p = str[1].str;
        end = p + str[1].len;
        for(;;){
            int c;
            if(p >= end)
                break;
            c = *p++;
            if('\e'==c){
                if(p >= end)
                    break;
                c = *p++ - 64;
            }
            teletext_decoder_write(dec, c);
        }
        break;
    default:
        break;
    }
    return 0;
}


static int
_tt_create_from_data(abydos_plugin_handle_t *h, const char *data,size_t len)
{
    TeletextDecoder *dec;
    int ret = -1;
    h->info->width = TT_SCREEN_WIDTH * TT_CHAR_WIDTH;
    h->info->height = TT_SCREEN_HEIGHT * TT_CHAR_HEIGHT;
    h->info->pixel_ratio = 25.0/36;

    dec = teletext_decoder_new(&cons_buf_if,h->cons);

    switch(h->type){
    case T_CEBRA:
        if(len < 91)
            goto error;
        if(memcmp(data, "CEBRA Teletext", 14))
            goto error;
        data += 19;
        len -= 19;
        cebra_parse(0, (uint8_t*)data, len, _on_cebra_data, dec);
        break;
    case T_SOFTEL:
        if(*(uint8_t*)data!=0xfe || data[1]!=1)
            goto error;
        DEBUG_p("language: %d",*(uint8_t*)&data[2]);
        data += 3;
        len -= 3;
        /* Skip header */
        {
            int headerlen = 3 + le16toh(*(uint8_t*)(data+1));
            data += headerlen;
            len -= headerlen;
        }
        while(len--)
            teletext_decoder_write(dec,*data++);
        break;
    case T_T42:
        t42_parse((uint8_t*)data, len, _on_t42_packet, dec);
        break;
    case T_TTI7:
        tti7_parse(data, len, _on_tti7_data, dec);
        break;
    case T_BBCRAM:
    case T_RAW:
        while(len--)
            teletext_decoder_write(dec,*data++);
        break;
    }

#ifndef TEST_RENDER
    h->surface[0] = abydos_image_surface_create_from_teletext_console(h->cons, S_ON);
    if(teletext_console_has_flash(h->cons)){
        h->info->frame_count = 2;
        h->surface[1] = abydos_image_surface_create_from_teletext_console(h->cons, S_OFF);
    }
#endif

    ret = 0;
error:
    teletext_decoder_free(dec);
    return ret;
}


#ifdef TEST_RENDER
static void
_set_color(cairo_t *cr, int idx)
{
    static const struct{
        double r;
        double g;
        double b;
    }color[8] = {
        {0,0,0}, /* black */
        {1,0,0}, /* red */
        {0,1,0}, /* green */
        {1,1,0}, /* yellow */
        {0,0,1}, /* blue */
        {1,0,1}, /* magenta */
        {0,1,1}, /* cyan */
        {1,1,1}  /* white */
    };
    cairo_set_source_rgb(cr,color[idx].r,color[idx].g,color[idx].b);
}
#endif


#ifdef TEST_RENDER
static void
_draw_character(cairo_t *cr,int cx,int cy,int ch)
{
    uint16_t *fontchar = &teletext_12x10[(ch-' ')*CHAR_HEIGHT];
    int x;
    int y;
    for(y=0;y<CHAR_HEIGHT;++y){
        int pixels = *fontchar++;
        for(x=0;x<CHAR_WIDTH;++x){
            pixels <<= 1;
            if(pixels & (1<<CHAR_WIDTH))
                cairo_rectangle(cr,cx+x,cy+y,1,1);
        }
    }
}
#endif


#ifdef TEST_RENDER
static void
_draw_screen(cairo_t *cr,Console *cons)
{
    const ConsoleCell *src = cons_buf_buffer(cons);
    int x;
    int y;
    cairo_save(cr);
    for(y=0;y<SCREEN_HEIGHT;++y){
        for(x=0;x<SCREEN_WIDTH;++x){
            union{
                TeletextState state;
                int attr;
            }s;
            ConsoleCell cell = *src++;
            s.attr = cell.attr;
            _set_color(cr,s.state.bg);
            cairo_rectangle(cr,x*CHAR_WIDTH,y*CHAR_HEIGHT,CHAR_WIDTH,CHAR_HEIGHT);
            cairo_fill(cr);
            if(cell.ch>' '){
                _set_color(cr,s.state.fg);
                if(s.state.graphic && (cell.ch&0x60) != 0x40){
                    _draw_graphic(cr,x*CHAR_WIDTH,y*CHAR_HEIGHT,cell.ch);
                }else{
                    _draw_character(cr,x*CHAR_WIDTH,y*CHAR_HEIGHT,cell.ch);
                }
                cairo_fill(cr);
            }
        }
    }
    cairo_restore(cr);
}
#endif


#ifdef TEST_RENDER
static void
_tt_render(abydos_plugin_handle_t *h,
           cairo_t *cr,
           int page,
           cairo_bool_t *layer_visibility,
           int varient,
           double time,
           int frame)
{
    _draw_screen(cr,h->cons);
}
#else
static cairo_surface_t *
_tt_get_image_surface(abydos_plugin_handle_t *h,int page,cairo_bool_t *layer_visibility,int variant,double time,int frame)
{
    return cairo_surface_reference(h->surface[frame]);
}
#endif


static void
_tt_get_duration(abydos_plugin_handle_t *h, int page, double *frame_duration)
{
    DEBUG_p("%d frames",h->info->frame_count);
    frame_duration[0] = 2.25;
    if(h->info->frame_count>1)
        frame_duration[1] = 0.75;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _tt_get_mime_types,
        .new = _tt_new,
        .free = _tt_free,
        .create_from_data = _tt_create_from_data,
#ifdef TEST_RENDER
        .render = _tt_render,
#else
        .get_image_surface = _tt_get_image_surface,
#endif
        .get_duration = _tt_get_duration
    };
    return &p;
}
