/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <atomic.h>
#include <cairo.h>
#include <debug.h>
#include <stddef.h>                     /* needed by gdevdsp.h */
#include <ghostscript/gdevdsp.h>
#include <ghostscript/gserrors.h>
#include <ghostscript/iapi.h>
#include <pthread.h>

#include "gs-types.h"

#include "gxdevice.h"
#include "gxdevsop.h"
#include "gzpath.h"
#include "tribble.h"


#define DPI 72


struct _tribble_page_t{
    double width;
    double height;
    cairo_surface_t *surface;
};

struct _tribble_document_t{
    int refcount;
    tribble_page_t *page;
    int page_count;
};

typedef struct{
    gx_device_common;
    tribble_document_t *doc;
    tribble_page_t *page; /* current page */
    cairo_t *cr; /* cairo context for the current page */
}tribble_device_t;


void gs_lib_register_device(const gx_device*);


static bool
_cairo_path(cairo_t *cr, gx_path *path)
{
#if 1
    const segment *seg = (segment*)path->segments->contents.subpath_first;
    const curve_segment *cseg;
    DEBUG_p("subpath count: %d", path->subpath_count);
    if(path->subpath_count <= 0)
        return false;
    while(seg){
        switch(seg->type){
        case s_start:
        case s_gap:
            DEBUG_p("move to: %fx%f", fixed2float(seg->pt.x), fixed2float(seg->pt.y));
            cairo_move_to(cr, fixed2float(seg->pt.x), fixed2float(seg->pt.y));
            break;
        case s_line:
            DEBUG_p("line to: %fx%f", fixed2float(seg->pt.x), fixed2float(seg->pt.y));
            cairo_line_to(cr, fixed2float(seg->pt.x), fixed2float(seg->pt.y));
            break;
        case s_line_close:
            cairo_close_path(cr);
            break;
        case s_curve:
            cseg = (curve_segment*)seg;
            cairo_curve_to(cr, fixed2float(cseg->p1.x), fixed2float(cseg->p1.y),
                               fixed2float(cseg->p2.x), fixed2float(cseg->p2.y),
                               fixed2float(seg->pt.x), fixed2float(seg->pt.y));
            break;
        }
        seg = seg->next;
    }
#else
    gs_path_enum penum;
    gx_path_enum_init(&penum, path);
    DEBUG_p("subpath count: %d", path->subpath_count);
    for(;;){
        gs_fixed_point pts[3];
        switch(gx_path_enum_next(&penum, pts)){
        case gs_pe_moveto:
        case gs_pe_gapto:
            DEBUG_p("move to: %fx%f", fixed2float(pts[0].x), fixed2float(pts[0].y));
            cairo_move_to(cr, fixed2float(pts[0].x), fixed2float(pts[0].y));
            continue;
        case gs_pe_lineto:
            cairo_line_to(cr, fixed2float(pts[0].x), fixed2float(pts[0].y));
            continue;
        case gs_pe_closepath:
            cairo_close_path(cr);
            continue;
        case gs_pe_curveto:
            cairo_curve_to(cr, fixed2float(pts[0].x), fixed2float(pts[0].y),
                               fixed2float(pts[1].x), fixed2float(pts[1].y),
                               fixed2float(pts[2].x), fixed2float(pts[2].y));
            continue;
        default:
            break;
        }
        break;
    }
#endif
    return true;
}


static void
_cairo_set_index(cairo_t *cr, gx_color_index c)
{
    DEBUG_p("%06X", (int)c);
    cairo_set_source_rgb(cr,
                         (double)((c>>0)&255) / 255,
                         (double)((c>>8)&255) / 255,
                         (double)((c>>16)&255) / 255);
}


static void
_cairo_set_color(cairo_t *cr, const gx_device_color *c)
{
    if(gx_dc_is_pure(c)){
        gx_color_index pc = gx_dc_pure_color(c);
        DEBUG_p("color is pure");
        _cairo_set_index(cr, pc);
    }else if(c->ccolor_valid){
        cairo_set_source_rgb(cr, c->ccolor.paint.values[0],
                                 c->ccolor.paint.values[1],
                                 c->ccolor.paint.values[2]);
    }
}


#if 0
static void
_clear_marks()
{
    DEBUG_p("begin");
}
#endif


#if 0
static const gs_ptr_procs_t *
_enum_ptrs(const gs_memory_t *mem,
           const void *x,
           unsigned int a,
           int b,
           enum_ptr_t *ptr,
           const gs_memory_struct_type_t *type,
           gc_state_t *state)
{
    DEBUG_p("begin");
    return NULL;
}


static void
_reloc_ptrs()
{
    DEBUG_p("begin");
}


static void
_finalize()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_dev_proc()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_finalize()
{
    DEBUG_p("begin");
}
#endif


static int
_open_device(gx_device *gdev)
{
    tribble_device_t *dev = (tribble_device_t*)gdev;
    DEBUG_p("begin");

    dev->doc = (tribble_document_t*)malloc(sizeof(tribble_document_t));
    dev->doc->refcount = 1;
    dev->doc->page = (tribble_page_t*)malloc(1 * sizeof(tribble_page_t));
    dev->doc->page_count = 1;

    dev->page = &dev->doc->page[0];
    dev->page->surface = cairo_recording_surface_create(CAIRO_CONTENT_COLOR_ALPHA, NULL);

    gx_callout(gdev, 0, 0, dev->doc);

    dev->cr = cairo_create(dev->doc->page[0].surface);
    return 0;
}

#if 0
static void
_get_initial_matrix(gx_device *dev, gs_matrix *pmat)
{
    DEBUG_p("begin");
    gx_default_get_initial_matrix(dev, pmat);
}
#endif


static int
_sync_output(gx_device *dev)
{
    DEBUG_p("begin");
    return 0;
}


static int
_output_page(gx_device *gdev, int a, int b)
{
    tribble_device_t *dev = (tribble_device_t*)gdev;
    DEBUG_p("page size: %dx%d", (int)dev->MediaSize[0], (int)dev->MediaSize[1]);
    dev->page->width = dev->MediaSize[0];
    dev->page->height = dev->MediaSize[1];
    return 0;
}


static int
_close_device(gx_device *gdev)
{
    tribble_device_t *dev = (tribble_device_t*)gdev;
    int i;
    DEBUG_p("begin");

    for(i=0;i<dev->doc->page_count;++i){
        tribble_page_t *p = &dev->doc->page[i];
        if(p->width <= 0 || p->height <= 0){
            double x, y, w, h;
            DEBUG_p("page %d has no size", i);
            cairo_recording_surface_ink_extents(p->surface, &x, &y, &w, &h);
            DEBUG_p("ink extents: %fx%f %fx%f", x, y, w, h);
            if(p->width <= 0)
                p->width = w;
            if(p->height <=0)
                p->height = h;
        }
    }

    //cairo_surface_flush(dev->doc->page[0].surface);
    cairo_destroy(dev->cr);
    tribble_document_destroy(dev->doc);
    return 0;
}


#if 0
static void
_map_rgb_color()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_map_color_rgb()
{
    DEBUG_p("begin");
}
#endif


static int
_fill_rectangle(gx_device *gdev,
                int x,
                int y,
                int width,
                int height,
                gx_color_index color)
{
    cairo_t *cr = ((tribble_device_t*)gdev)->cr;
    DEBUG_p("%dx%d %dx%d", x, y, width, height);
#if 1
    cairo_rectangle(cr, x, y, width, height);
    _cairo_set_index(cr, color);
    cairo_fill(cr);
#endif
    return 0;
}


#if 0
static int
_tile_rectangle(gx_device *dev,
                const gx_tile_bitmap *bm,
                int a,
                int b,
                int c,
                int d,
                gx_color_index color1,
                gx_color_index color2,
                int e,
                int f)
{
    DEBUG_p("begin");
    return 0;
}
#endif


#if 0
static int
_copy_mono(gx_device *dev,
           const byte *a,
           int b,
           int c,
           gx_bitmap_id bm,
           int d,
           int e,
           int f,
           int g,
           gx_color_index color1,
           gx_color_index color2)
{
    DEBUG_p("begin");
    return 0;
}
#endif


#if 0
static int
_copy_color(gx_device *dev,
            const byte *a,
            int b,
            int c,
            gx_bitmap_id bm,
            int d,
            int e,
            int f,
            int g)
{
    DEBUG_p("begin");
    return 0;
}
#endif


#if 0
static int
_get_bits(gx_device *dev, int a, byte *b, byte **c)
{
    DEBUG_p("begin");
    return 0;
}
#endif


static int
_get_params(gx_device *dev, gs_param_list *plist)
{
    bool bool_true = true;
    int code;
    DEBUG_p("begin");
    if((code = gx_default_get_params(dev, plist)) < 0)
        return code;
    if((code = param_write_bool(plist, "HighLevelDevice", &bool_true)) < 0)
        return code;
    if((code = param_write_bool(plist, "NoInterpolateImagemasks", &bool_true)) < 0)
        return code;
    return 0;
}


#if 0
static int
_put_params(gx_device *dev, gs_param_list *plist)
{
    DEBUG_p("begin");
    return gx_default_put_params(dev, plist);
}
#endif


static gx_color_index
_map_cmyk_color(gx_device *dev, const gx_color_value *color)
{
    DEBUG_p("begin");
    return 0xff00ff;
}


#if 0
static gx_xfont_procs *
_get_xfont_procs(gx_device *dev)
{
    static const gx_xfont_procs proct = {
    };
    DEBUG_p("begin");
    return procs;
}
#endif


#if 0
static gx_color_index
_map_rgb_alpha_color(gx_device *dev,
                     gx_color_value *a,
                     gx_color_value *b,
                     gx_color_value *c)
{
    DEBUG_p("begin");
    return 0xff00ff;
}
#endif


#if 0
static int
_get_alpha_bits(gx_device *dev, graphics_object_type type)
{
    DEBUG_p("begin");
    return 0;
}
#endif


#if 0
static int
_copy_alpha(gx_device *dev,
            const byte *a,
            int b,
            int c,
            gx_bitmap_id bm,
            int d,
            int e,
            int f,
            int g,
            gx_color_index color,
            int h)
{
    DEBUG_p("begin");
    return 0;
}
#endif


#if 0
static int
_get_band(gx_device *dev, int a, int b)
{
    DEBUG_p("begin");
    return 0;
}
#endif


#if 0
static int
_copy_rop(gx_device *dev,
          const byte *a,
          int b,
          unsigned int c,
          gx_bitmap_id bm,
          const gx_color_index *color,
          const gx_tile_bitmap *bm2,
          const gx_color_index *color2,
          int d,
          int e,
          int f,
          int g,
          int h,
          int i,
          gs_logical_operation_t op)
{
    DEBUG_p("begin");
    return 0;
}
#endif


static int
_fill_path(gx_device *gdev,
           const gs_gstate *state,
           gx_path *path,
           const gx_fill_params *params,
           const gx_drawing_color *color,
           const gx_clip_path *clip)
{
    cairo_t *cr = ((tribble_device_t*)gdev)->cr;
    DEBUG_p("begin");
    if(_cairo_path(cr, path)){
        _cairo_set_color(cr, color);
        cairo_fill(cr);
    }
    return 0;
}


static int
_stroke_path(gx_device *gdev,
             const gs_gstate *state,
             gx_path *path,
             const gx_stroke_params *params,
             const gx_drawing_color *color,
             const gx_clip_path *clip)
{
    cairo_t *cr = ((tribble_device_t*)gdev)->cr;
    DEBUG_p("begin");
    _cairo_path(cr, path);
    _cairo_set_color(cr, color);
    cairo_stroke(cr);
    return 0;
}


#if 0
static int
_fill_mask(gx_device *dev,
           const byte *a,
           int b,
           int c,
           gx_bitmap_id bm,
           int d,
           int e,
           int f,
           int g,
           const gx_drawing_color *color,
           int h,
           gs_logical_operation_t op,
           const gx_clip_path *clip)
{
    DEBUG_p("begin");
    return 0;
}
#endif


#if 0
static void
_fill_trapezoid()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_fill_parallelogram()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_fill_triangle()
{
    DEBUG_p("begin");
}
#endif


static int
_draw_thin_line(gx_device *dev,
                fixed a,
                fixed b,
                fixed c,
                fixed d,
                const gx_drawing_color *color,
                gs_logical_operation_t op,
                fixed e,
                fixed f)
{
    DEBUG_p("begin");
    return 0;
}


static int
_begin_image(gx_device *dev,
             const gs_gstate *state,
             const gs_image_t *img,
             gs_image_format_t format,
             const gs_int_rect *rect,
             const gx_drawing_color *color,
             const gx_clip_path *clip,
             gs_memory_t *mem,
             gx_image_enum_common_t **ienum)
{
    DEBUG_p("begin");
    return 0;
}


static int
_image_data(gx_device *dev,
            gx_image_enum_common_t *ienum,
            const byte **a,
            int b,
            unsigned int c,
            int d)
{
    DEBUG_p("begin");
    return 0;
}


static int
_end_image(gx_device *dev, gx_image_enum_common_t *ienum, bool b)
{
    DEBUG_p("begin");
    return 0;
}


#if 0
static void
_strip_tile_rectangle()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_strip_copy_rop()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_get_clipping_box()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_begin_typed_image()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_get_bits_rectangle()
{
    DEBUG_p("begin");
}
#endif


#if 0
static int
_map_color_rgb_alpha(gx_device *dev,
                     gx_color_index color,
                     gx_color_value *val)
{
    DEBUG_p("begin");
    return 0;
}
#endif


#if 0
static void
_create_compositor()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_get_hardware_params()
{
    DEBUG_p("begin");
}
#endif


static int
_text_begin(gx_device *dev,
            gs_gstate *state,
            const gs_text_params_t *params,
            gs_font *font,
            gx_path *path,
            const gx_device_color *color,
            const gx_clip_path *clip,
            gs_memory_t *mem,
            gs_text_enum_t **tenum)
{
    DEBUG_p("begin");
    return 0;
}


#if 0
static int
_finish_copydevice(gx_device *gdev, const gx_device *from_dev)
{
    DEBUG_p("begin");
    return 0;
}
#endif


#if 0
static void
_begin_transparency_group()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_end_transparency_group()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_begin_transparency_mask()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_end_transparency_mask()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_discard_transparency_layer()
{
    DEBUG_p("begin");
}
#endif


static int
_get_color_comp_index(gx_device *dev, const char *pname, int name_size, int component_type)
{
    DEBUG_p("name: %.*s (%d)", name_size, pname, component_type);
    return 0;
}


static gx_color_index
_encode_color(gx_device *gdev, const gx_color_value *cv)
{
    //DEBUG_p("%d %d %d", (int)cv[0], (int)cv[1], (int)cv[2]);
    gx_color_index c = gx_color_value_to_byte(cv[0]) << 0 |
                       gx_color_value_to_byte(cv[1]) << 8 |
                       gx_color_value_to_byte(cv[2]) << 16;
    //DEBUG_p("%06X", (int)c);
    return c;
}


#if 0
static void
_decode_color()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_pattern_manage()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_fill_rectangle_hl_color()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_include_color_space()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_fill_linear_color_scanline()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_fill_linear_color_trapezoid()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_fill_linear_color_triangle()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_update_spot_equivalent_colors()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_ret_devn_params()
{
    DEBUG_p("begin");
}
#endif


static int
_push_transparency_state(gx_device *dev, gs_gstate *state)
{
    DEBUG_p("begin");
    return 0;
}


static int
_pop_transparency_state(gx_device *dev, gs_gstate *state)
{
    DEBUG_p("begin");
    return 0;
}


static int
_put_image(gx_device *dev,
           gx_device *dev2,
           const byte **a,
           int b,
           int c,
           int d,
           int e,
           int f,
           int g,
           int h,
           int i)
{
    DEBUG_p("begin");
    return 0;
}


static int
_dev_spec_op(gx_device *dev, int op, void *data, int datasize)
{
    switch(op){
    case gxdso_supports_devn:
    case gxdso_skip_icc_component_validation:
        //return (dev_proc(dev, fill_rectangle_hl_color) == _cairo_fill_rectangle_hl_color);
        return 1;
    case gxdso_reopen_after_init:
        return 1;
    case gxdso_adjust_bandheight:
        return 0;
    }
    DEBUG_p("op: %d", op);
    return gx_default_dev_spec_op(dev, op, data, datasize);
}


#if 0
static void
_copy_planes()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_set_graphics_type_tag(gx_device *dev, gs_graphics_type_tag_t tag)
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_strip_copy_rop2()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_strip_tile_rect_devn()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_copy_alpha_hl_color()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_process_page()
{
    DEBUG_p("begin");
}
#endif


#if 0
static void
_transform_pixel_region()
{
    DEBUG_p("begin");
}
#endif


static int
_fill_stroke_path(gx_device *gdev,
                  const gs_gstate *state,
                  gx_path *path,
                  const gx_fill_params *fill_params,
                  const gx_drawing_color *fill_color,
                  const gx_stroke_params *stroke_params,
                  const gx_drawing_color *stroke_color,
                  const gx_clip_path *clip)
{
    cairo_t *cr = ((tribble_device_t*)gdev)->cr;
    DEBUG_p("begin");
    _cairo_path(cr, path);
    _cairo_set_color(cr, fill_color);
    cairo_fill_preserve(cr);
    _cairo_set_color(cr, stroke_color);
    cairo_stroke(cr);
    return 0;
}


static int
_tribble_init(void)
{
    static const tribble_device_t dev = {
        .params_size = sizeof(tribble_device_t),
        //.static_procs = _dev_proc,
        .dname = "tribble",
        //.memory = 0,
        //.stype = &stype,
        //.stype_is_dynamic = 0,
        //.finalize = _finalize,
        //.rc = {0},
        //.retained = false,
        //.parent = NULL,
        //.child = NULL,
        //.subclass_data = NULL,
        //.PageList = NULL,
        //.is_open = false,
        /* color_info goes here */
        //.max_fill_band = 0,
        .color_info = {
            .max_components = 3,
            .num_components = 3,
            .polarity = GX_CINFO_POLARITY_ADDITIVE,
            .depth = 24,
            .gray_index = GX_CINFO_COMP_NO_INDEX,
            .max_gray = 255,
            .max_color = 255,
            .dither_grays = 256,
            .dither_colors = 256,
            .anti_alias = {
                .text_bits = 1,
                .graphics_bits = 1
            },
            .separable_and_linear = 1,
            //{ 0 } /* component shift */
            //{ 0 } /* component bits */
            //{ 0 } /* component mask */
            // cn /* process color name */
            .opmode = GX_CINFO_OPMODE_UNKNOWN,
            .process_comps = 0
        },
        .cached_colors = {
            gx_no_color_index,
            gx_no_color_index
        },
        .width = DEFAULT_WIDTH_10THS * DPI,
        .height = DEFAULT_HEIGHT_10THS * DPI,
        .is_planar = 0,
        .LeadingEdge = 0,
        /*
        .MediaSize = {
            ((WIDTH_10THS * 72.0 + 0.5) - 0.5) / DPI,
            ((HEIGHT_10THS * 72.0 + 0.5) - 0.5) / DPI
        },
        */


        .BLS_force_memory = true,


        //.icc_struct = NULL,
        .HWResolution = {DPI, DPI},
        .page_procs = {
            gx_default_install,
            gx_default_begin_page,
            gx_default_end_page
        },
        .procs = {
            .open_device = _open_device,
            .get_initial_matrix = gx_default_get_initial_matrix,
            .sync_output = _sync_output,
            .output_page = _output_page,
            .close_device = _close_device,
            .map_rgb_color = gx_default_w_b_map_rgb_color,
            .map_color_rgb = gx_default_w_b_map_color_rgb,
            .fill_rectangle = _fill_rectangle,
            //.tile_rectangle = _tile_rectangle,
            //.copy_mono = _copy_mono,
            //.copy_color = _copy_color,
            //.obsolete_draw_line = _obsolete_draw_line,
            //.get_bits = _get_bits,
            .get_params = _get_params,
            //.put_params = _put_params,
            .map_cmyk_color = _map_cmyk_color,
            //.get_xfont_procs = _get_xfont_procs,
            //.get_xfont_device = _get_xfont_device,
            //.map_rgb_alpha_color = _map_rgb_alpha_color,
            //.get_page_device = gx_page_device_get_page_device,
            /* extra entries */
            //.get_alpha_bits = _get_alpha_bits,
            //.copy_alpha = _copy_alpha,
            //.get_band = _get_band,
            //.copy_rop = _copy_rop,
            .fill_path = _fill_path,
            .stroke_path = _stroke_path,
            //.fill_mask = _fill_mask,
            //.fill_trapezoid = _fill_trapezoid,
            //.fill_parallelogram = _fill_parallelogram,
            //.fill_triangle = _fill_triangle,
            .draw_thin_line = _draw_thin_line,
            .begin_image = _begin_image,
            .image_data = _image_data,
            .end_image = _end_image,
            //.strip_tile_rectangle = _strip_tile_rectangle,
            //.strip_copy_rop = _strip_copy_rop,
            //.get_clipping_box = _get_clipping_box,
            //.begin_typed_image = _begin_typed_image,
            //.get_bits_rectangle = _get_bits_rectangle,
            //.map_color_rgb_alpha = _map_color_rgb_alpha,
            //.create_compositor = _create_compositor,
            //.get_hardware_params = _get_hardware_params,
            .text_begin = _text_begin,
            //.finish_copydevice = _finish_copydevice,
            //.begin_transparency_group = _begin_transparency_group,
            //.end_transparency_group = _end_transparency_group,
            //.begin_transparency_mask = _begin_transparency_mask,
            //.end_transparency_mask = _end_transparency_mask,
            //.discard_transparency_layer = _discard_transparency_layer,
            //.get_color_mapping_procs = _get_color_mapping_procs,
            .get_color_comp_index = _get_color_comp_index,
            .encode_color = _encode_color,
            //.decode_color = _decode_color,
            //.pattern_manage = _pattern_manage,
            //.fill_rectangle_hl_color = _fill_rectangle_hl_color,
            //.include_color_space = _include_color_space,
            //.fill_linear_color_scanline = _fill_linear_color_scanline,
            //.fill_linear_color_trapezoid = _fill_linear_color_trapezoid,
            //.fill_linear_color_triangle = _fill_linear_color_triangle,
            //.update_spot_equivalent_colors = _update_spot_equivalent_colors,
            //.ret_devn_params = _ret_devn_params,
            //.fillpage = _fillpage,
            .push_transparency_state = _push_transparency_state,
            .pop_transparency_state = _pop_transparency_state,
            .put_image = _put_image,
            .dev_spec_op = _dev_spec_op,
            //.copy_planes = _copy_planes,
            //.get_profile = _get_profile,
            //.set_graphics_type_tag = _set_graphics_type_tag,
            //.strip_copy_rop2 = _strip_copy_rop2,
            //.strip_tile_rect_devn = _strip_tile_rect_devn,
            //.copy_alpha_hl_color = _copy_alpha_hl_color,
            //.process_page = _process_page,
            //.transform_pixel_region = _transform_pixel_region,
            .fill_stroke_path = _fill_stroke_path
        }
    };
    static pthread_mutex_t init_lock = PTHREAD_MUTEX_INITIALIZER;
    static int initialized;
    pthread_mutex_lock(&init_lock);
    if(!initialized){
        gs_lib_register_device((gx_device*)&dev);
        initialized = 1;
    }
    pthread_mutex_unlock(&init_lock);
    return 0;
}



static int
_on_callout(void *instance,
            void *userdata,
            const char *device_name,
            int id,
            int size,
            void *data)
{
    *(tribble_document_t**)userdata = tribble_document_reference((tribble_document_t*)data);
    return 0;
}


typedef struct{
    const char *argv[16];
    int argc;
}arguments_t;


static void
_arguments_add(arguments_t *args, const char *arg)
{
    args->argv[args->argc++] = arg;
}


static tribble_document_t *
_tribble_document_create(const char *filename, const char *data, size_t len)
{
    tribble_document_t *doc = NULL;
    void *minst = NULL;
    arguments_t args;
    char format[64];
    int code, code1;

    DEBUG_p("begin");

    if(_tribble_init())
        return NULL;

    code = gsapi_new_instance(&minst, NULL);
    if (code < 0)
        return NULL;
    code = gsapi_set_arg_encoding(minst, GS_ARG_ENCODING_UTF8);
    code = gsapi_register_callout(minst, _on_callout, &doc);

    args.argc = 0;
    _arguments_add(&args, "tribble");
    _arguments_add(&args, "-sDEVICE=tribble");
    _arguments_add(&args, "-dNOPAUSE");
    _arguments_add(&args, "-dBATCH");
    _arguments_add(&args, "-dSAFER");
#ifdef _USE_DEBUG
    _arguments_add(&args, "-dDEBUG");
#endif
    sprintf(format, "-dDisplayFormat=%d", DISPLAY_COLORS_RGB |
            DISPLAY_ALPHA_NONE | DISPLAY_DEPTH_8 |
            DISPLAY_LITTLEENDIAN | DISPLAY_BOTTOMFIRST);
    _arguments_add(&args, format);
    if(filename)
        _arguments_add(&args, filename);

    if (code == 0)
        code = gsapi_init_with_args(minst, args.argc, (char**)args.argv);

    DEBUG_p("after gsapi_init_with_args()");

    if(data){
        gsapi_run_string_begin(minst, 0, &code1);
        gsapi_run_string_continue(minst, data, len, 0, &code1);
        gsapi_run_string_end(minst, 0, &code1);
    }

    code1 = gsapi_exit(minst);
    if ((code == 0) || (code == gs_error_Quit))
        code = code1;

    gsapi_delete_instance(minst);
    return doc;
}


tribble_document_t *
tribble_document_create_from_data(const char *data, size_t len)
{
    return _tribble_document_create(NULL, data, len);
}


tribble_document_t *
tribble_document_create_from_file(const char *filename)
{
    return _tribble_document_create(filename, NULL, 0);
}


tribble_document_t *
tribble_document_reference(tribble_document_t *doc)
{
    atomic_int_inc(&doc->refcount);
    return doc;
}


void
tribble_document_destroy(tribble_document_t *doc)
{
    if(atomic_int_dec_and_test(&doc->refcount)){
        if(doc->page){
            int i;
            for(i=0;i<doc->page_count;++i)
                cairo_surface_destroy(doc->page[i].surface);
            free(doc->page);
        }
        free(doc);
    }
}


int
tribble_document_get_n_pages(const tribble_document_t *doc)
{
    return doc->page_count;
}


const tribble_page_t *
tribble_document_get_page(const tribble_document_t *doc, int pagenr)
{
    return &doc->page[pagenr];
}



void
tribble_page_get_size(const tribble_page_t *page, double *width, double *height)
{
    if(width)
        *width = page->width;
    if(height)
        *height = page->height;
}


void
tribble_page_render(const tribble_page_t *page, cairo_t *cr)
{
    DEBUG_p("painting replay surface");
    cairo_set_source_surface(cr, page->surface, 0.0, 0.0);
    cairo_paint(cr);
}

/*

{params_size = 1800, static_procs = 0x0, dname = 0x7ffff7fb316c "tribble", memory = 0x0,
  stype = 0x0, stype_is_dynamic = 0, finalize = 0x0, rc = {ref_count = 0, memory = 0x0, free =
    0x0}, retained = 0, parent = 0x0, child = 0x0, subclass_data = 0x0, PageList = 0x0,
  is_open = 0, max_fill_band = 0, color_info = {max_components = 3 '\003',
    num_components = 3 '\003', polarity = GX_CINFO_POLARITY_ADDITIVE, depth = 24,
    gray_index = 255 '\377', max_gray = 255, max_color = 255, dither_grays = 256,
    dither_colors = 256, anti_alias = {text_bits = 1, graphics_bits = 1},
    separable_and_linear = GX_CINFO_SEP_LIN, comp_shift = '\000' <repeats 63 times>,
    comp_bits = '\000' <repeats 63 times>, comp_mask = {0 <repeats 64 times>}, cm_name = 0x0,
    opmode = GX_CINFO_OPMODE_UNKNOWN, process_comps = 0, black_component = 0,
    use_antidropout_downscaler = 0}, cached_colors = {black = 18446744073709551615,
    white = 18446744073709551615}, width = 6120, height = 7920, pad = 0, log2_align_mod = 0,
  is_planar = 0, LeadingEdge = 0, MediaSize = {0, 0}, ImagingBBox = {0, 0, 0, 0},

  ImagingBBox_set = 0, HWResolution = {72, 72}, Margins = {0, 0}, HWMargins = {0, 0, 0, 0},
  FirstPage = 0, LastPage = 0,
  PageHandlerPushed = 0, @948
  DisablePageHandler = 0, @952
  ObjectFilter = 0, ObjectHandlerPushed = 0, PageCount = 0, ShowpageCount = 0, NumCopies = 0,
  NumCopies_set = 0, IgnoreNumCopies = 0, UseCIEColor = 0, LockSafetyParams = 0,
  band_offset_x = 0, band_offset_y = 0, BLS_force_memory = 0, sgr = {stroke_stored = 0,
    orig = {{x = 0, y = 0}, {x = 0, y = 0}, {x = 0, y = 0}, {x = 0, y = 0}}, adjusted = {{x = 0,
        y = 0}, {x = 0, y = 0}, {x = 0, y = 0}, {x = 0, y = 0}}}, MaxPatternBitmap = 0,
  page_uses_transparency = 0, space_params = {MaxBitmap = 0, BufferSpace = 0, band = {
      BandWidth = 0, BandHeight = 0, BandBufferSpace = 0, tile_cache_size = 0},
    params_are_read_only = -161575618,

    banding_type = (BandingAlways | BandingNever | unknown: 0x7ffc)},
  icc_struct = 0x7ffff65e8d44 <gx_default_begin_page>,
  graphics_type_tag = (GS_IMAGE_TAG | GS_UNKNOWN_TAG | unknown: 0xf65e8d08),
  interpolate_control = 32767, page_procs = {install = 0x7ffff7fb1a37 <_open_device>,
    begin_page = 0x7ffff65e6127 <gx_default_get_initial_matrix>, end_page =
    0x7ffff7fb1b89 <_sync_output>}

{params_size = 1800, static_procs = 0x0, dname = 0x7ffff7fb316c "tribble", memory = 0x0,
  stype = 0x0, stype_is_dynamic = false, finalize = 0x0, rc = {ref_count = 0, memory = 0x0,
    free = 0x0}, retained = false, parent = 0x0, child = 0x0, subclass_data = 0x0,
  PageList = 0x0, is_open = false, max_fill_band = 0, color_info = {max_components = 3 '\003',
    num_components = 3 '\003', polarity = GX_CINFO_POLARITY_ADDITIVE, depth = 24,
    gray_index = 255 '\377', max_gray = 255, max_color = 255, dither_grays = 256,
    dither_colors = 256, anti_alias = {text_bits = 1, graphics_bits = 1},
    separable_and_linear = GX_CINFO_SEP_LIN, comp_shift = '\000' <repeats 63 times>,
    comp_bits = '\000' <repeats 63 times>, comp_mask = {0 <repeats 64 times>}, cm_name = 0x0,
    opmode = GX_CINFO_OPMODE_UNKNOWN, process_comps = 0, black_component = 0,
    use_antidropout_downscaler = false}, cached_colors = {black = 18446744073709551615,
    white = 18446744073709551615}, width = 6120, height = 7920, pad = 0, log2_align_mod = 0,
  is_planar = 0, LeadingEdge = 0, MediaSize = {0, 0}, ImagingBBox = {0, 0, 0, 0},

  ImagingBBox_set = false, HWResolution = {72, 72}, Margins = {0, 0}, HWMargins = {0, 0, 0, 0},
  FirstPage = 0,
  LastPage = 0,
  PageHandlerPushed = false, @948
  DisablePageHandler = false, @949
  ObjectFilter = 0,
  ObjectHandlerPushed = false,
  PageCount = 0, ShowpageCount = 0, NumCopies = 0,
  NumCopies_set = false, IgnoreNumCopies = false, UseCIEColor = false, LockSafetyParams = false,
  band_offset_x = 0, band_offset_y = 0, BLS_force_memory = false, sgr = {stroke_stored = false,
    orig = {{x = 0, y = 0}, {x = 0, y = 0}, {x = 0, y = 0}, {x = 0, y = 0}}, adjusted = {{x = 0,
        y = 0}, {x = 0, y = 0}, {x = 0, y = 0}, {x = 0, y = 0}}}, MaxPatternBitmap = 0,
  page_uses_transparency = false, space_params = {MaxBitmap = 0, BufferSpace = 0, band = {
      BandWidth = 0, BandHeight = 0, BandBufferSpace = 0, tile_cache_size = 0},
    params_are_read_only = false, banding_type = BandingAuto}, icc_struct = 0x0,
  graphics_type_tag = GS_UNTOUCHED_TAG, interpolate_control = 0, page_procs = {install =
    0x7ffff65e8d3e <gx_default_install>, begin_page = 0x7ffff65e8d44 <gx_default_begin_page>,
    end_page = 0x7ffff65e8d4a <gx_default_end_page>}
*/
