/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <debug.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* memcpy() */

#include "sixeldec.h"

#define MAX_COLORS 256

struct _sixel_decoder_t{
    sixel_func_t userfunc;
    void *userdata;

    int command;
    int param[8];
    int paramnr;

    sixel_color_t color[MAX_COLORS];
    int colornr;
    int x;
    int y;
};


sixel_decoder_t *
sixel_decoder_new(sixel_func_t userfunc, void *userdata)
{
    static const sixel_color_t default_palette[16] = {
        0xff000000, /* black */
        0xff3333cc, /* blue */
        0xffcc2222, /* red */
        0xff33cc33, /* green */
        0xffcc33cc, /* magenta */
        0xff33cccc, /* cyan */
        0xffcccc33, /* yellow */
        0xff888888, /* gray 50% */
        0xff444444, /* gray 25% */
        0xff555599, /* blue */
        0xff994444, /* red */
        0xff559955, /* green */
        0xff995599, /* magenta */
        0xff559999, /* cyan */
        0xff999955, /* yellow */
        0xffcccccc  /* gray 75% */
    };
    sixel_decoder_t *dec = (sixel_decoder_t*)malloc(sizeof(sixel_decoder_t));
    dec->userfunc = userfunc;
    dec->userdata = userdata;
    dec->command = 0;
    dec->param[0] = 0;
    dec->paramnr = 0;
    memcpy(dec->color, default_palette, 16*sizeof(sixel_color_t));
    /* default color */
    dec->colornr = 15;
    dec->x = 0;
    dec->y = 0;
    return dec;
}


void
sixel_decoder_free(sixel_decoder_t *dec)
{
    free(dec);
}


void
sixel_decoder_data(sixel_decoder_t *dec, int ch)
{
    if(ch>='0' && ch<='9'){
        dec->param[dec->paramnr] *= 10;
        dec->param[dec->paramnr] += (ch-'0');
        return;
    }
    ++dec->paramnr;
    if(';'==ch){
        dec->param[dec->paramnr] = 0;
        return;
    }
    switch(dec->command){
    case '#':
        dec->colornr = dec->param[0] % MAX_COLORS;
        if(dec->paramnr >= 3){
            DEBUG_p("color %d, (%d params)",dec->param[0],dec->paramnr);
            switch(dec->param[1]){
            case 2: /* RGB color */
                if(dec->paramnr >= 5){
                    int r = 255*dec->param[2]/100;
                    int g = 255*dec->param[3]/100;
                    int b = 255*dec->param[4]/100;
                    dec->color[dec->colornr] = r<<16|g<<8|b<<0|255<<24;
                }
                break;
            }
        }
        break;
    }
    switch(ch){
    case '!':
    case '#':
        break;
    case '$':
        /* Carriage return */
        dec->x = 0;
        break;
    case '-':
        /* Line feed */;
        dec->x = 0;
        ++dec->y;
        break;
    default:
        if(ch>=63 && ch<=(63+63)){
            int repeat = '!'==dec->command ? dec->param[0] : 1;
            DEBUG_p("color: %d",dec->colornr);
            dec->userfunc(dec->userdata, dec->x, dec->y, dec->color[dec->colornr], ch-63, repeat);
            dec->x += repeat;
        }else{
            DEBUG_p("unknown command: '%c'",ch);
        }
        break;
    }
    dec->command = ch;
    dec->param[0] = 0;
    dec->paramnr = 0;
}
