/*
 * VTF image loader
 *
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 * Copyright (C) 2010  Forrest Voight <voights@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
/*
 * Based on gdk-pixbuf-vtf by Forrest Voight <voights@gmail.com>
 * and Jan Dudek <jd@jandudek.com>.
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <inttypes.h>
#include <math.h>                       /* lrintf() */
#include <nil-cairo.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "vtf.h"


typedef struct{
    cairo_surface_t *surface;
    double duration;
} frame_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    frame_t *frame;
};


static const char **
_vtf_get_mime_types()
{
    static const char *mime_types[] = {
        "image/vnd.valve.source.texture",
        "image/x-vtf",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_vtf_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    info->threadsafe = 1;
    h->frame = NULL;

    return h;
}


static void
_vtf_free(abydos_plugin_handle_t *h)
{
    if(h->frame){
        int i;
        for(i=0;i<h->info->frame_count;++i)
            if(h->frame[i].surface)
                cairo_surface_destroy(h->frame[i].surface);
        free(h->frame);

    }
    free(h);
}


static size_t
_frame_size(int image_format, int width, int height) {
    switch (image_format) {
        case IMAGE_FORMAT_NONE:              return 0;
        case IMAGE_FORMAT_RGBA8888:          return width * height * 4;
        case IMAGE_FORMAT_ABGR8888:          return width * height * 4;
        case IMAGE_FORMAT_RGB888:            return width * height * 3;
        case IMAGE_FORMAT_BGR888:            return width * height * 3;
        case IMAGE_FORMAT_RGB565:            return width * height * 2;
        case IMAGE_FORMAT_I8:                return width * height * 1;
        case IMAGE_FORMAT_IA88:              return width * height * 2;
//      case IMAGE_FORMAT_P8:                return TODO;
        case IMAGE_FORMAT_A8:                return width * height * 1;
//      case IMAGE_FORMAT_RGB888_BLUESCREEN: return TODO;
//      case IMAGE_FORMAT_BGR888_BLUESCREEN: return TODO;
        case IMAGE_FORMAT_ARGB8888:          return width * height * 4;
        case IMAGE_FORMAT_BGRA8888:          return width * height * 4;
        case IMAGE_FORMAT_DXT1:              return ((width+3)/4) * ((height+3)/4) * 8;
//      case IMAGE_FORMAT_DXT3:              return TODO;
        case IMAGE_FORMAT_DXT5:              return ((width+3)/4) * ((height+3)/4) * 16;
        case IMAGE_FORMAT_BGRX8888:          return width * height * 4;
        case IMAGE_FORMAT_BGR565:            return width * height * 2;
        case IMAGE_FORMAT_BGRX5551:          return width * height * 2;
        case IMAGE_FORMAT_BGRA4444:          return width * height * 2;
//      case IMAGE_FORMAT_DXT1_ONEBITALPHA:  return TODO;
        case IMAGE_FORMAT_BGRA5551:          return width * height * 2;
//      case IMAGE_FORMAT_UV88:              return TODO;
//      case IMAGE_FORMAT_UVWQ8888:          return TODO;
        case IMAGE_FORMAT_RGBA16161616F:     return width * height * 8;
        case IMAGE_FORMAT_RGBA16161616:      return width * height * 8;
//      case IMAGE_FORMAT_UVLX8888:          return TODO;
        /* not yet supported or illegal value */
        default:                             return SIZE_MAX;
    }
}


#define get16(a) le16toh(*(uint16_t*)(a))
#define get32(a) le32toh(*(uint32_t*)(a))


/*
#if __BYTE_ORDER == __BIG_ENDIAN
static inline float
_leftoh(float f)
{
    uint32_t v = le32toh(*(uint32_t*)&f);
    return *(float*)&v;
}
#else
# define _leftoh(f) (f)
#endif
*/


static inline float
_to_float(int v)
{
    union{
        uint16_t b[2];
        float f;
    }val;
    val.b[0] = 0;
    val.b[1] = v;
    return val.f;
}


/* cube maps have 6 or 7 faces, other maps have only 1 */
/*
static size_t
face_count(const VTFHEADER *header) {
    return header->flags & TEXTUREFLAGS_ENVMAP ?
           (header->firstFrame != 0xffff && (header->version[0] < 7 || (header->version[0] == 7 && header->version[1] < 5)) ? 7 : 6) :
           1;
}
*/


static int
_vtf_mip_size(VTFHEADER *header, int mip_level, int depth){
    int mip_width  = header->width  >> mip_level;
    int mip_height = header->height >> mip_level;
    int mip_depth  = depth          >> mip_level;

    if(mip_width < 1)
        mip_width = 1;
    if(mip_height < 1)
        mip_height = 1;
    if(mip_depth < 1)
        mip_depth = 1;

    return _frame_size(header->highResImageFormat, mip_width, mip_height) * mip_depth;
}


static int
_vtf_offset(VTFHEADER *header, int frame, int face, int slice, int mip_level)
{
    int offset = 0;
    int facecount = 1;
    int volume_bytes;
    int slice_bytes;
    int i;

    for(i = header->mipmapCount - 1; i > mip_level; --i)
        offset += _vtf_mip_size (header, i, header->depth);

    offset *= header->frames * facecount;

    volume_bytes = _vtf_mip_size (header, mip_level, header->depth);
    slice_bytes  = _vtf_mip_size (header, mip_level, 1);

    offset += volume_bytes * (frame * facecount + face);
    offset += slice_bytes  * slice;

    return offset;
}


static cairo_surface_t *
_surface_from_dxt1(const uint8_t *src, int width, int height)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24, width, height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    int y;
    for(y=0;y<height;y+=4){
        int x;
        for(x=0;x<width;x+=4) {
            int r[4];
            int g[4];
            int b[4];
            int a[4];
            int c0 = get16(src);
            int c1 = get16(src+2);
            uint32_t sel;
            int x2;
            int y2;
            src += 4;

            r[0] = (c0 >> 11) & 31;
            r[0] = (r[0] << 3) | (r[0] >> 5);
            g[0] = (c0 >> 5) & 63;
            g[0] = (g[0] << 2) | (g[0] >> 6);
            b[0] = (c0 >> 0) & 31;
            b[0] = (b[0] << 3) | (b[0] >> 5);

            r[1] = (c1 >> 11) & 31;
            r[1] = (r[1] << 3) | (r[1] >> 5);
            g[1] = (c1 >> 5) & 63;
            g[1] = (g[1] << 2) | (g[1] >> 6);
            b[1] = (c1 >> 0) & 31;
            b[1] = (b[1] << 3) | (b[1] >> 5);

            a[0] = 255;
            a[1] = 255;

            if(c0 > c1) {
                r[2] = (4*r[0] + 2*r[1] + 3)/6;
                g[2] = (4*g[0] + 2*g[1] + 3)/6;
                b[2] = (4*b[0] + 2*b[1] + 3)/6;
                a[2] = 255;

                r[3] = (2*r[0] + 4*r[1] + 3)/6;
                g[3] = (2*g[0] + 4*g[1] + 3)/6;
                b[3] = (2*b[0] + 4*b[1] + 3)/6;
                a[3] = 255;
            }else{
                r[2] = (r[0] + r[1] + 1)/2;
                g[2] = (g[0] + g[1] + 1)/2;
                b[2] = (b[0] + b[1] + 1)/2;
                a[2] = 255;

                r[3] = 0;
                g[3] = 0;
                b[3] = 0;
                a[3] = 0;
            }
            sel = get32(src);
            src += 4;

            for(y2=0;y2<4;++y2)
                for(x2=0;x2<4;++x2){
                    dst[x+x2 + (y+y2)*stride] = r[sel&3]<<16 |
                                                g[sel&3]<< 8 |
                                                b[sel&3]<< 0 |
                                                a[sel&3]<< 24;
                    sel >>= 2;
                }
        }
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static cairo_surface_t *
_surface_from_dxt5(const uint8_t *src, int width, int height)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, width, height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    int y;
    for(y=0;y<height;y+=4){
        int x;
        for(x=0;x<width;x+=4){
            {
                uint16_t a[8];
                uint64_t sel;
                int x2;
                int y2;

                a[0] = *src++;
                a[1] = *src++;

                if(a[0] > a[1]) {
                    a[2] = (12*a[0] + 2*a[1] + 7)/14;
                    a[3] = (10*a[0] + 4*a[1] + 7)/14;
                    a[4] = (8*a[0] + 6*a[1] + 7)/14;
                    a[5] = (6*a[0] + 8*a[1] + 7)/14;
                    a[6] = (4*a[0] + 10*a[1] + 7)/14;
                    a[7] = (2*a[0] + 12*a[1] + 7)/14;
                } else {
                    a[2] = (8*a[0] + 2*a[1] + 5)/10;
                    a[3] = (6*a[0] + 4*a[1] + 5)/10;
                    a[4] = (4*a[0] + 6*a[1] + 5)/10;
                    a[5] = (2*a[0] + 8*a[1] + 5)/10;
                    a[6] = 0;
                    a[7] = 255;
                }

                sel  = *src++;
                sel |= (uint64_t)*src++ << 8;
                sel |= (uint64_t)*src++ << 16;
                sel |= (uint64_t)*src++ << 24;
                sel |= (uint64_t)*src++ << 32;
                sel |= (uint64_t)*src++ << 40;

                for(y2=0;y2<4;++y2)
                    for(x2=0;x2<4;++x2){
                        dst[x+x2 + (y+y2)*stride] = a[sel&7] << 24;
                        sel >>= 3;
                    }
            }
            {
                int c0;
                int c1;
                int r[4];
                int g[4];
                int b[4];
                uint32_t sel;
                int x2;
                int y2;
                c0 = get16(src);
                c1 = get16(src+2);
                src += 4;

                r[0] = (c0 >> 11) & 31;
                r[0] = (r[0] << 3) | (r[0] >> 5);
                g[0] = (c0 >> 5) & 63;
                g[0] = (g[0] << 2) | (g[0] >> 6);
                b[0] = (c0 >> 0) & 31;
                b[0] = (b[0] << 3) | (b[0] >> 5);

                r[1] = (c1 >> 11) & 31;
                r[1] = (r[1] << 3) | (r[1] >> 5);
                g[1] = (c1 >> 5) & 63;
                g[1] = (g[1] << 2) | (g[1] >> 6);
                b[1] = (c1 >> 0) & 31;
                b[1] = (b[1] << 3) | (b[1] >> 5);

                r[2] = (4*r[0] + 2*r[1] + 3)/6;
                g[2] = (4*g[0] + 2*g[1] + 3)/6;
                b[2] = (4*b[0] + 2*b[1] + 3)/6;

                r[3] = (2*r[0] + 4*r[1] + 3)/6;
                g[3] = (2*g[0] + 4*g[1] + 3)/6;
                b[3] = (2*b[0] + 4*b[1] + 3)/6;

                sel = get32(src);
                src += 4;

                for(y2=0;y2<4;++y2)
                    for(x2=0;x2<4;++x2){
                        uint32_t *pix = &dst[x+x2 + (y+y2)*stride];
                        int a = *pix >> 24;
                        *pix |= (a*r[sel&3]/255) << 16 |
                                (a*g[sel&3]/255) <<  8 |
                                (a*b[sel&3]/255) <<  0;
                        sel >>= 2;
                    }
            }
        }
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static cairo_surface_t *
_surface_from_rgb565le(const uint16_t *src, int width, int height)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB16_565, width, height);
    uint16_t *dst = (uint16_t*)cairo_image_surface_get_data(surface);
    int skip = cairo_image_surface_get_stride(surface)/sizeof(uint16_t) - width;
#if __BYTE_ORDER == __BIG_ENDIAN
    while(height-->0){
        int x;
        for(x=0;x<width;++x)
            *dst++ = le32toh(*src++);
        dst += skip;
    }
#else
    if(skip){
        while(height-->0){
            memcpy(dst,src,width*sizeof(uint16_t));
            src += width;
            dst += width+skip;
        }
    }else{
        memcpy(dst,src,width*height*sizeof(uint16_t));
    }
#endif
    cairo_surface_mark_dirty(surface);
    return surface;
}


static cairo_surface_t *
_surface_from_a8(const uint8_t *src, int width, int height)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, width, height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int skip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t) - width;
    while(height-->0){
        int x;
        for(x=0;x<width;++x){
            int a = *src++;
            *dst++ = a<<16 | a<<8 | a<<0 | a<<24;
        }
        dst += skip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static cairo_surface_t *
_surface_from_rgba64le(const uint16_t *src, int width, int height)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, width, height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int skip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t) - width;
    while(height-->0){
        int x;
        for(x=0;x<width;++x){
            int r = le16toh(src[0]) / 257;
            int g = le16toh(src[1]) / 257;
            int b = le16toh(src[2]) / 257;
            int a = le16toh(src[3]) / 257;
            src += 4;
            *dst++ = r<<16 | g<<8 | b | a<<24;
        }
        dst += skip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static cairo_surface_t *
_surface_from_rgba64fle(const uint16_t *src, int width, int height)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, width, height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int skip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t) - width;
    while(height-->0){
        int x;
        for(x=0;x<width;++x){
            int r = lrintf(_to_float(le16toh(src[0]))*255);
            int g = lrintf(_to_float(le16toh(src[1]))*255);
            int b = lrintf(_to_float(le16toh(src[2]))*255);
            int a = lrintf(_to_float(le16toh(src[3]))*255);
            src += 4;
            *dst++ = r<<16 | g<<8 | b | a<<24;
        }
        dst += skip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static cairo_surface_t *
_surface_from_format(int format, const uint8_t *src, int width, int height)
{
    DEBUG_p("creating surface from format %d",format);
    switch(format){
    case IMAGE_FORMAT_RGBA8888:
        return nil_cairo_surface_create_from_planar(src, width, height, width*4,
                                                    4, NILE_FORMAT_r8g8b8a8|NILE_PREMULTIPLIED,
                                                    0);
    case IMAGE_FORMAT_ABGR8888:
        return nil_cairo_surface_create_from_planar(src, width, height, width*4,
                                                    4, NILE_FORMAT_a8r8g8b8|NILE_PREMULTIPLIED,
                                                    0);
    case IMAGE_FORMAT_RGB888:
        return nil_cairo_surface_create_from_planar(src, width, height, width*3,
                                                    3, NILE_FORMAT_r8g8b8|NILE_PREMULTIPLIED,
                                                    0);
    case IMAGE_FORMAT_BGR888:
        return nil_cairo_surface_create_from_planar(src, width, height, width*3,
                                                    3, NILE_FORMAT_b8g8r8|NILE_PREMULTIPLIED,
                                                    0);
    case IMAGE_FORMAT_RGB565:
        return _surface_from_rgb565le((const uint16_t*)src, width, height);
    case IMAGE_FORMAT_I8:
        return nil_cairo_surface_create_from_planar(src, width, height, width,
                                                    1, NILE_FORMAT_g8|NILE_PREMULTIPLIED,
                                                    0);
    case IMAGE_FORMAT_IA88:
        return nil_cairo_surface_create_from_planar(src, width, height, width*2,
                                                    2, NILE_FORMAT_g8a8|NILE_PREMULTIPLIED,
                                                    0);
    case IMAGE_FORMAT_A8:
        return _surface_from_a8(src, width, height);
    case IMAGE_FORMAT_DXT1:
        return _surface_from_dxt1(src, width, height);
    case IMAGE_FORMAT_DXT5:
        return _surface_from_dxt5(src, width, height);
    case IMAGE_FORMAT_ARGB8888:
        return nil_cairo_surface_create_from_planar(src, width, height, width*4,
                                                    4, NILE_FORMAT_a8r8g8b8|NILE_PREMULTIPLIED,
                                                    0);
    case IMAGE_FORMAT_BGRA8888:
        return nil_cairo_surface_create_from_planar(src, width, height, width*4,
                                                    4, NILE_FORMAT_b8g8r8a8|NILE_PREMULTIPLIED,
                                                    0);
    case IMAGE_FORMAT_RGBA16161616F:
        return _surface_from_rgba64fle((const uint16_t*)src, width, height);
    case IMAGE_FORMAT_RGBA16161616:
        return _surface_from_rgba64le((const uint16_t*)src, width, height);
    default:
        DEBUG_p("unhandled format: %d",format);
        return NULL;
    }
}


static int
_vtf_create_from_data(abydos_plugin_handle_t *h, const char *data, size_t len)
{
    VTFHEADER header;
    int fulldata;
    int base;

    if(len < sizeof(VTFHEADER))
        return -1;

    memcpy(&header,data,sizeof(VTFHEADER));

    if(memcmp(header.signature,"VTF",4))
        return -1;

    /*
    header.version[0] = le32toh(header.version[0]);
    header.version[1] = le32toh(header.version[1]);
    */
    header.headerSize = le32toh(header.headerSize);
    header.width = le16toh(header.width);
    header.height = le16toh(header.height);
    //header.flags = le32toh(header.flags);
    header.frames = le16toh(header.frames);
    /*
    header.firstFrame = le16toh(header.firstFrame);
    header.reflectivity[0] = _leftoh(header.reflectivity[0]);
    header.reflectivity[1] = _leftoh(header.reflectivity[1]);
    header.reflectivity[2] = _leftoh(header.reflectivity[2]);
    header.reflectivity[3] = _leftoh(header.reflectivity[3]);
    header.bumpmapScale = _leftoh(header.bumpmapScale);
    */
    header.highResImageFormat = le32toh(header.highResImageFormat);
    //header.lowResImageFormat = le32toh(header.lowResImageFormat);
#if 1
    if(offsetof(VTFHEADER,depth)>=header.headerSize){
        header.depth = 1;
    }else{
        header.depth = le16toh(header.depth);
    }
#else
    if (header.version[0] < 7 || (header.version[0] == 7 && header.version[1] < 2)){
        header.depth = 1;
    }else{
        header.depth = le16toh(header.depth);
    }
#endif
    /*
#if 1
    if(offsetof(VTFHEADER,resources)>=header.headerSize){
        header.resources = 0;
    }else{
        header.resources = le32toh(header.resources);
    }
#else
    if (header.version[0] < 7 || (header.version[0] == 7 && header.version[1] < 3)){
        header.resources = 0;
    }else{
        header.resources = le32toh(header.resources);
    }
#endif
    */

    fulldata = _vtf_offset(&header, 0, 0, 0, -1);
    if(len < fulldata)
        return -1;

    base = len - fulldata;
    h->info->width = header.width;
    h->info->height = header.height;
    h->info->frame_count = header.frames;
    h->frame = (frame_t*)calloc(h->info->frame_count,sizeof(frame_t));

    for(int i=0; i < h->info->frame_count;++i){
        int offs = base + _vtf_offset(&header, i, 0, 0, 0);
        if(offs + _frame_size(header.highResImageFormat,h->info->width,h->info->height) > len)
            return -1;
        h->frame[i].surface = _surface_from_format(header.highResImageFormat, (uint8_t*)data+offs, h->info->width, h->info->height);
        if(!h->frame[i].surface)
            return -1;
    }
    return 0;
}


static cairo_surface_t *
_vtf_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->frame[frame].surface);
}


static void
_vtf_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = 0.2;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _vtf_get_mime_types,
        .new = _vtf_new,
        .free = _vtf_free,
        .create_from_data = _vtf_create_from_data,
        .get_image_surface = _vtf_get_image_surface,
        .get_duration = _vtf_get_duration
    };
    return &p;
}
