/*
 * Copyright (C) 2019 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos.h>
#include <abydos-plugin.h>
#include <assert.h>
#include <debug.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>


/* #define ENABLE_WINICON */ /* Incomplete. */


typedef struct{
    abydos_t *ar;
    double duration;
}frame_t;

typedef struct{
    cairo_surface_t *surface;
    int width;
    int height;
}variant_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    enum{
        T_ANI,
        T_WINICON
    }type;
    frame_t *frame;
    variant_t *variant;
};


static const char **
_ani_get_mime_types()
{
    static const char *mime_types[] = {
        "application/x-navi-animation",
#ifdef ENABLE_WINICON
        "image/vnd.microsoft.icon",
#endif
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_ani_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
#ifdef ENABLE_WINICON
    /* This block isn't needed if there can be only only one MIME type. */
    if(!strcasecmp(mime_type,"application/x-navi-animation"){
        h->type = T_ANI;
    }else if(!strcasecmp(mime_type,"image/vnd.microsoft.icon")){
        h->type = T_WINICON;
    }else{
        free(h);
        return NULL;
    }
#else
    h->type = T_ANI;
#endif
    h->info = info;
    h->info->threadsafe = 1;
    return h;
}


static void
_ani_free(abydos_plugin_handle_t *h)
{
    free(h);
}


#define _cmdid(a,b) (*(int32_t*)(a) - *(int32_t*)(b))
#define _get16(p) le16toh(*(uint16_t*)(p))
#define _get32(p) le32toh(*(uint32_t*)(p))


#ifdef USE_DEBUG
static int
_read_INFO(abydos_plugin_handle_t *h,const char *p, size_t len)
{
    while(len >= 8){
        int size = _get32(p+4);
        if(!_cmdid(p,"INAM")){
           DEBUG_p("INAM (%.*s)",(int)size,p+8);
        }else if(!_cmdid(p,"IART")){
            int length = _get32(p+4);
            DEBUG_p("IART (%.*s)",(int)length,p+8);
        }else{
            DEBUG_p("[unknown]");
        }
        size+=8+(size&1);
        p+=size;
        len-=size;
    }
    return 0;
}
#endif


static int
_read_fram(abydos_plugin_handle_t *h, frame_t *frame, int max, const char *p, size_t len)
{
    int nr = 0;
    while(len>=8 && nr<max){
        int size = _get32(p+4);
        if(!_cmdid(p,"icon")){
            abydos_t *ar = abydos_create("image/vnd.microsoft.icon");
            if(ar){
                if(abydos_from_data(ar,p+8,size)<0){
                    DEBUG_p("failed to read embedded icon");
                    abydos_destroy(ar);
                    ar = NULL;
                }
            }
            frame[nr++].ar = ar;
        }else{
            DEBUG_p("[unknown]");
        }
        size+=8+(size&1);
        p+=size;
        len-=size;
    }
    return nr;
}


static int
_read_ACON(abydos_plugin_handle_t *h,const char *p, size_t len)
{
    int d_rate = 1;
    int f_rate = -1;
    int seq = 0;
    while(len >= 8){
        int size = _get32(p+4);
        if(!_cmdid(p,"LIST")){
            DEBUG_p("LIST (%d)",size);
#ifdef USE_DEBUG
            if(!_cmdid(p+8,"INFO")){
                _read_INFO(h,p+12,size);
            }else
#endif
            if(!_cmdid(p+8,"fram")){
                int i;
                int count = _read_fram(h,&h->frame[seq],h->info->frame_count-seq,p+12,size);
                if(count<h->info->frame_count){
                    /* not all promised frames were present */
                    h->info->frame_count = count;
                }
                for(i=0;i<count;++i){
                    int width;
                    int height;
                    abydos_get_size(h->frame[seq].ar,&width,&height);
                    if(width>h->info->width)
                        h->info->width = width;
                    if(height>h->info->height)
                        h->info->height = height;
                    h->frame[seq].duration = (double)(f_rate>0 ? f_rate : d_rate)/60;
                    ++seq;
                }
                h->info->variant_count = abydos_get_variant_count(h->frame[0].ar);
                h->variant = (variant_t*)malloc(h->info->variant_count*sizeof(variant_t));
                for(i=0;i<h->info->variant_count;++i)
                    abydos_get_variant_size(h->frame[0].ar,i,&h->variant[i].width,&h->variant[i].height);
                /* What if different frames have a different set of variants? */
            }
        }else if(!_cmdid(p,"anih")){
            /*  8 frames
               12 steps
               16 cx
               20 cy
               24 bitcount
               28 planes
               32 jifrate
               36 flags
             */
            int flags = _get32(p+36);
            DEBUG_p("anih flags: %x",flags);
            h->info->frame_count = _get32(p+12);
            d_rate = _get32(p+32);
            f_rate = -1;
            DEBUG_p("frames: %d, rate: %d",h->info->frame_count,d_rate);
            h->frame = (frame_t*)malloc(h->info->frame_count*sizeof(frame_t));
        }else if(!_cmdid(p,"rate")){
            f_rate = _get32(p+8);
            DEBUG_p("rate: %d",f_rate);
        }else if(!_cmdid(p,"seq ")){
            int f_seq = _get32(p+8);
            DEBUG_p("seq: %d",f_seq);
        }else{
            DEBUG_p("[unknown]");
        }
        size+=8+(size&1);
        p+=size;
        len-=size;
    }
    /* FIXME!! make sure every frame of the animation is defined */
    return 0;
}


#ifdef ENABLE_WINICON
static int
_get_size(const char *p){
    int size = *(uint8_t*)p;
    return size>0 ? size : 256;
}


static cairo_surface_t *
_read_image(abydos_plugin_handle_t *h,const char *bytes, size_t len)
{
}


static int
_read_ico(abydos_plugin_handle_t *h,const char *bytes, size_t len)
{
}


static int
_read_cur(abydos_plugin_handle_t *h,const char *bytes, size_t len)
{
    const char *p = bytes;
    int i;
    h->info->frame_count = _get16(p);
    p += 2;
    for(i=0;i<h->info->frame_count;++i){
        size_t size = _get32(p+8);
        size_t offs = _get32(p+12);
        p += 16;
    }
}
#endif


static int
_ani_create_from_data(abydos_plugin_handle_t *h,const char *bytes, size_t len)
{
    const char* p = bytes;
    size_t size;
    switch(h->type){
    case T_ANI:
        if(_cmdid(p+0,"RIFF") || _cmdid(p+8,"ACON"))
            return -1;
        size = _get32(p+4);
        if(len < size+8)
            return -1;
        return _read_ACON(h,p+12,size-4);
    case T_WINICON:
#ifdef ENABLE_WINCON
        if(_get16(p)!=0)
            return -1;
        switch(_get16(p+2)){
        case 1:return _read_ico(h,p+4,len-4);
        case 2:return _read_cur(h,p+4,len-4);
        }
#endif
        return -1;
    }
    return -1;
}


static cairo_surface_t *
_ani_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    abydos_t *ar = abydos_clone(h->frame[frame].ar);
    cairo_surface_t *surface;
    abydos_set_variant(ar,variant);
    surface = abydos_get_image_surface(ar);
    abydos_destroy(ar);
    return surface;
}


static void
_ani_get_variant_size(abydos_plugin_handle_t *h, int page, abydos_size_t *size)
{
    int i;
    for(i=0;i<h->info->variant_count;++i){
        size[i].width = h->variant[i].width;
        size[i].height = h->variant[i].height;
    }
}


static void
_ani_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->frame[i].duration;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _ani_get_mime_types,
        .new = _ani_new,
        .free = _ani_free,
        .create_from_data = _ani_create_from_data,
        .get_image_surface = _ani_get_image_surface,
        .get_variant_size = _ani_get_variant_size,
        .get_duration = _ani_get_duration
    };
    return &p;
}
