/*
 * Copyright (C) 2020 - 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <nil-amiga.h>
#include <nil-cairo.h>
#include <stdlib.h>
#include <string.h>

#include "xpk-dec.h"

/*

Header:

-------v------v--------------
offset | size | Description
-------+------+--------------
     0 |   18 | "SVG Graphics File\0"
    18 |    2 | Version (always 1)
    20 |    4 | Header size
    24 |    4 | X offset
    28 |    4 | Y offset
    32 |    4 | Width
    36 |    4 | Height
    40 |    4 | Depth
    44 |    4 | View mode
    48 |    1 | Bits
    49 |    1 | Planes
    50 |    4 | Stride
    54 |  768 | Palette

*/

#define ID_XPKF ID('X','P','K','F')
#define ID_PP20 ID('P','P','2','0')
#define ID_LZ77 ID('L','Z','7','7')

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


#define get8(a) (*(uint8_t*)(a))
#define get16(a) be16toh(*(uint16_t*)(a))
#define get32(a) be32toh(*(uint32_t*)(a))


static const char **
_sview_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-superview",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_sview_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = malloc(sizeof(abydos_plugin_handle_t));
    info->threadsafe = 1;
    h->info = info;
    h->surface = NULL;
    return h;
}


static void
_sview_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


static int
_sview_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    const uint8_t *palette;
    //nile_t ni;
    int bits;
    int planes;
    int depth;
    int src_stride;
    uint32_t bplcon0;
    uint8_t *body;
    int body_offs;
    int body_len;
    uint8_t *dst;
    int dst_stride;

    if(len<58)
        return -1;
    if(memcmp(data, "SVG Graphics File\0", 18) && memcmp(data, "SGX Graphics File\0", 18))
        return -1;
    if(get16(data+18)!=1)
        return -1;

    bits = get8(data+48);
    planes = get8(data+49);
    if(bits > 1 && planes > 1)
        return -1;

    depth = get32(data+40);
    DEBUG_p("bits: %d, planes: %d, depth: %d",bits,planes,depth);

    switch(bits){
    case 1:
        if(planes > 8 || depth > planes)
            return -1;
        break;
    case 8:
        if(depth > 8)
            return -1;
        break;
    case 24:
    case 32:
    case 48:
    case 64:
        if(bits != depth)
            return -1;
        break;
    default:
        return -1;
    }

    if(depth <= 8){
        if(54 + (1<<depth) > len)
            return -1;
        palette = (uint8_t*)data+54;
        //ni.colors = 1<<depth;
        //_expand_palette((uint8_t*)data+54,palette,1<<depth);
    }

    h->info->width = get32(data+32);
    h->info->height = get32(data+36);
    src_stride = get32(data+50);
    DEBUG_p("size: %dx%d, stride: %d",h->info->width,h->info->height, src_stride);

    body_offs = get32(data+20);
    DEBUG_p("data offset: %d",body_offs);
    body_len = src_stride * h->info->height * planes;

    bplcon0 = get32(data+44);
    body = (uint8_t*)malloc(body_len);
    if(xpk_decompress((char*)body, body_len, data + body_offs, len-body_offs) < 0){
        free(body);
        return -1;
    }

    DEBUG_p("success");
    h->surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24, h->info->width, h->info->height);
    dst = cairo_image_surface_get_data(h->surface);
    dst_stride = cairo_image_surface_get_stride(h->surface);

    /* create surface from body */
    if(1==bits){
        /* bitplanes */
        uint8_t *bitplane[8];
        int i;
        for(i=0;i<planes;++i)
            bitplane[i] = &body[src_stride*h->info->height*i];
        nil_amiga_blit_planar_from_bitplanes(dst, dst_stride,
                                             4, NIL_CAIRO_FORMAT_RGB24,
                                             bitplane, nil_amiga_plane_stride(h->info->width), depth,
                                             palette, 1<<depth, 3, NILE_FORMAT_r8g8b8,
                                             h->info->width, h->info->height,
                                             bplcon0, 0, 0);
    }else if(8==bits){
        nile_blit_planar_from_indexed_rgb(dst, dst_stride,
                                          4, NIL_CAIRO_FORMAT_RGB24,
                                          body, src_stride,
                                          palette, 1<<depth, 3, NILE_FORMAT_r8g8b8,
                                          h->info->width, h->info->height,
                                          0);
    }else{
        nile_format_t format;
        int bpc;
        switch(depth){
        case 24:
            format = NILE_FORMAT_r8g8b8;
            bpc = 3;
            break;
        case 32:
            format = NILE_FORMAT_r8g8b8a8;
            bpc = 4;
            break;
        case 48:
            format = NILE_FORMAT_r16g16b16;
            bpc = 3;
            break;
        case 64:
            format = NILE_FORMAT_r16g16b16a16;
            bpc = 4;
            break;
        }
        nile_blit_planar_rgb(dst, dst_stride, 4, NIL_CAIRO_FORMAT_RGB24,
                             body, src_stride, bpc, format,
                             h->info->width, h->info->height,
                             0);
        //h->surface = nil_cairo_surface_create_from_planar(body, h->info->width, h->info->height, src_stride, bpc, format, 0);
    }
    free(body);
    return 0;
}


static cairo_surface_t *
_sview_get_image_surface(abydos_plugin_handle_t *h,
                         int page,
                         cairo_bool_t *layer_visibility,
                         int variant,
                         double time,
                         int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _sview_get_mime_types,
        .new = _sview_new,
        .free = _sview_free,
        .create_from_data = _sview_create_from_data,
        .get_image_surface = _sview_get_image_surface
    };
    return &p;
}
