/*
 * Copyright (C) 1995 - 1997  Peter Mattis, Spencer Kimball and Josh MacDonald
 * Copyright (C) 2009         Mica, Östersunds kummun
 * Copyright (C) 2016 - 2020  Magnus Bergman
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
/*
 * Based on ghash.c from GLIB (ftp://ftp.gtk.org/pub/gtk/)
 */
#include "hash.h"

#ifndef HASH_USE_GLIB

#include <ctype.h>                      /* tolower() */
#include <limits.h>                     /* UINT_MAX */
#include <stdint.h>
#include <string.h>                     /* memset(), strcmp() */

#include "memutil.h"


#ifndef MAX
# define MAX(a,b) ((a)>(b)?(a):(b))
#endif

#define HASH_MAP_MIN_SHIFT 3            /* 1 << 3 == 8 buckets */
#define UNUSED_HASH 0
#define TOMBSTONE_HASH 1
#define FIRST_REAL_HASH 2
#define HASH_IS_REAL(a) ((a)>=FIRST_REAL_HASH)


typedef struct{
    void* key;
    void* value;
    unsigned int hash;
}HashNode;

struct _HashMap{
    int size;                           /* size of the nodes array */
    HashNode* nodes;
    int mod;
    unsigned int mask;
    int nnodes;                         /* real entries in the hash map */
    int noccupied;                      /* nnodes + tombstones */
    HashFunc hash_func;
    EqualFunc equal_func;
    HashMapNotifyFunc key_notify_func;
    HashMapNotifyFunc value_notify_func;
};

/* --- private functions ------------------------------------------------ */

static const int prime_mod [] = {
    1,          /* For 1 << 0 */
    2,
    3,
    7,
    13,
    31,
    61,
    127,
    251,
    509,
    1021,
    2039,
    4093,
    8191,
    16381,
    32749,
    65521,      /* For 1 << 16 */
    131071,
    262139,
    524287,
    1048573,
    2097143,
    4194301,
    8388593,
    16777213,
    33554393,
    67108859,
    134217689,
    268435399,
    536870909,
    1073741789,
    2147483647  /* For 1 << 31 */
};


static void hash_map_set_shift(HashMap* hm,int shift){
    unsigned int mask = 0;
    int i;
    hm->size = 1 << shift;
    hm->nodes = u_new0(HashNode,hm->size);
    hm->mod  = prime_mod[shift];
    for(i=0;i<shift;++i){
        mask <<= 1;
        mask |= 1;
    }
    hm->mask = mask;
}


static int hash_map_find_closest_shift(int n){
    int i;
    for(i=0;n;++i)
        n >>= 1;
    return i;
}


#if 1
static HashNode* hash_map_lookup_node(HashMap* hm,const void* key,unsigned int* hash_return){
    unsigned int index;
    unsigned int hash;
    HashNode* first_tombstone = NULL; /* Just to silent compiler warnings. */
    int have_tombstone = 0;
    unsigned int step = 0;

    hash = hm->hash_func(key);
    if(!HASH_IS_REAL(hash))
        hash = FIRST_REAL_HASH;

    if(hash_return)
        *hash_return = hash;

    for(index=hash%hm->mod;;){
        HashNode* node = &hm->nodes[index];
        if(UNUSED_HASH == node->hash){
            return have_tombstone?first_tombstone:node;
        }else if(node->hash == hash){
            if(hm->equal_func){
                if(hm->equal_func(node->key,key))
                    return node;
            }else if(node->key == key){
                return node;
            }
        }else if(TOMBSTONE_HASH == node->hash && !have_tombstone){
            first_tombstone = node;
            have_tombstone = 1;
        }
        index += ++step;
        index &= hm->mask;
    }
}
#else
static inline unsigned int hash_map_lookup_index(HashMap* hm,const void* key,unsigned int* hash_return){
    unsigned int index;
    unsigned int hash;
    unsigned int first_tombstone;
    int have_tombstone = 0;
    unsigned int step = 0;

    hash = hm->hash_func(key);
    if(!HASH_IS_REAL(hash))
        hash = FIRST_REAL_HASH;

    if(hash_return)
        *hash_return = hash;

    for(index=hash%hm->mod;;){
        HashNode* node = &hm->nodes[index];
        if(UNUSED_HASH == node->hash){
            break;
        }else if(node->hash == hash){
            if(hm->equal_func){
                if(hm->equal_func(node->key,key))
                    return index;
            }else if(node->key == key){
                return index;
            }
        }else if(TOMBSTONE_HASH == node->hash && !have_tombstone){
            first_tombstone = index;
            have_tombstone = 1;
        }
        index += ++step;
        index &= hm->mask;
    }
    return have_tombstone?first_tombstone:index;
}
#endif


static void hash_map_remove_all_nodes(HashMap* hm,int destruction){
    if(hm->nnodes > 0 && (hm->key_notify_func || hm->value_notify_func)){
        int i;
        for(i=0;i<hm->size;++i){
            HashNode* node = &hm->nodes[i];
            if(HASH_IS_REAL(node->hash)){
                if(hm->key_notify_func)
                    hm->key_notify_func(node->key);
                if(hm->value_notify_func)
                    hm->value_notify_func(node->value);
            }
        }
    }

    if(destruction){
        u_free(hm->nodes);
        return;
    }

    if(hm->size <= (1<<HASH_MAP_MIN_SHIFT)){
        /* No need to downsize, just clear it */
        memset(hm->nodes,0,hm->size*sizeof(HashNode));
    }else{
        u_free(hm->nodes);
        hash_map_set_shift(hm,HASH_MAP_MIN_SHIFT);
    }
    hm->nnodes = hm->noccupied = 0;
}


static void hash_map_resize(HashMap* hm){
    HashNode* old_nodes;
    int old_size;

    old_nodes = hm->nodes;
    old_size = hm->size;
    hash_map_set_shift(hm,MAX(hash_map_find_closest_shift(hm->nnodes*2),HASH_MAP_MIN_SHIFT));

    if(hm->nnodes > 0){
        int i;
        for(i=0;i<old_size;++i){
            HashNode* old_node = &old_nodes[i];
            HashNode* new_node;
            unsigned int index;
            unsigned int step = 0;

            if(!HASH_IS_REAL(old_node->hash))
                continue;

            for(index=old_node->hash%hm->mod;;){
                new_node = &hm->nodes[index];
                if(!new_node->hash)
                    break;
                index += ++step;
                index &= hm->mask;
            }
            *new_node = *old_node;
        }
    }
    u_free(old_nodes);
    hm->noccupied = hm->nnodes;
}


static inline void hash_map_maybe_resize(HashMap* hm){
    int noccupied = hm->noccupied;
    int size = hm->size;
    if((size > hm->nnodes*4 && size > 1<<HASH_MAP_MIN_SHIFT) || (size <= noccupied+(noccupied/16)))
        hash_map_resize(hm);
}

/* --- public functions ------------------------------------------------- */

HashMap* hash_map_new(
    HashFunc hash_func,
    EqualFunc equal_func,
    HashMapNotifyFunc key_notify_func,
    HashMapNotifyFunc value_notify_func
){
    HashMap* hm = u_new(HashMap,1);
    hash_map_set_shift(hm,HASH_MAP_MIN_SHIFT);
    hm->nnodes = 0;
    hm->noccupied = 0;
    hm->hash_func = hash_func ? hash_func : direct_hash;
    hm->equal_func = equal_func ? equal_func : direct_equal;
    hm->key_notify_func = key_notify_func;
    hm->value_notify_func = value_notify_func;
    return hm;
}


void hash_map_free(HashMap* hm){
    hash_map_remove_all_nodes(hm,1);
    u_free(hm);
}


void hash_map_add(HashMap* hm,void* key,void* value){
    HashNode* node;
    //unsigned int index;
    unsigned int hash;
    unsigned int old_hash;

#if 1
    node = hash_map_lookup_node(hm,key,&hash);
#else
    index = hash_map_lookup_index(hm,key,&hash);
    node = &hm->nodes[index];
#endif

    old_hash = node->hash;

    if(HASH_IS_REAL(old_hash)){
        if(hm->key_notify_func)
            hm->key_notify_func(key);
        if(hm->value_notify_func)
            hm->value_notify_func(node->value);
        node->value = value;
    }else{
        node->key = key;
        node->value = value;
        node->hash = hash;

        ++hm->nnodes;

        if(!old_hash){
            /* We replaced an empty node, and not a tombstone */
            ++hm->noccupied;
            hash_map_maybe_resize(hm);
        }
    }
}


static const HashNode* _hash_map_get_node(HashMap* hm,const void* key){
    unsigned int index;
    unsigned int hash;
    int step = 0;

    hash = hm->hash_func(key);
    if(!HASH_IS_REAL(hash))
        hash = FIRST_REAL_HASH;

    for(index=hash%hm->mod;;){
        HashNode* node = &hm->nodes[index];
        if(UNUSED_HASH == node->hash){
            return NULL;
        }else if(node->hash == hash){
            if(hm->equal_func){
                if(hm->equal_func(node->key,key))
                    return node;
            }else if(node->key == key){
                return node;
            }
        }
        index += ++step;
        index &= hm->mask;
    }
}


void* hash_map_get(HashMap* hm,const void* key){
    const HashNode* node = _hash_map_get_node(hm,key);
    return node ? node->value : NULL;
}


const void* hash_map_get_key(HashMap* hm,const void* key){
    const HashNode* node = _hash_map_get_node(hm,key);
    return node ? node->key : NULL;
}


void* hash_map_get_any(HashMap* hm){
    int i;
    for(i=0;i<hm->size;++i){
        HashNode* node = &hm->nodes[i];
        if(HASH_IS_REAL(node->hash))
            return node->value;
    }
    return NULL;
}


int hash_map_remove(HashMap* hm,const void* key){
    //unsigned int index;
    HashNode* node;
    unsigned int hash;

#if 1
    node = hash_map_lookup_node(hm,key,&hash);
#else
    index = hash_map_lookup_index(hm,key,&hash);
    node = hm->nodes[index];
#endif

    if(!HASH_IS_REAL(node->hash))
        return 0;

    /* Erect tombstone */
    node->hash = TOMBSTONE_HASH;

    --hm->nnodes;

    if(hm->key_notify_func)
        hm->key_notify_func(node->key);
    if(hm->value_notify_func)
        hm->value_notify_func(node->value);

    hash_map_maybe_resize(hm);

    return 1;
}


void hash_map_remove_all(HashMap* hm){
    hash_map_remove_all_nodes(hm,0);
}


int hash_map_size(HashMap* hm){
    return hm->nnodes;
}


int hash_map_foreach(HashMap* hm,HashMapForeachFunc func,void* userdata){
    int i;
    for(i=0;i<hm->size;++i){
        HashNode* node = &hm->nodes[i];
        if(HASH_IS_REAL(node->hash)){
            int ret = func(node->key,node->value,userdata);
            if(ret)
                return ret;
        }
    }
    return 0;
}

/* --- Hash Functions --- */

unsigned int hash_combine(unsigned int a,unsigned int b){
    a ^= b + 0x9e3779b9 + (a << 6) + (a >> 2);
    return a;
}


int direct_equal(const void* a,const void* b){
    return a==b;
}

unsigned int direct_hash(const void* a){
    return (intptr_t)a;
}


unsigned int pointer_hash(const void* a){
#if 1
    unsigned int h = (intptr_t)a;
#else
# if INTPTR_MAX == INT64_MAX
    /* simply strip the MSBs if sizeof(void*) > sizeof(int) */
    unsigned int h = ((uint64_t)a) & UINT_MAX;
# else
    unsigned int h = (uint32_t)a;
# endif
#endif
#if 1
    h = ~h + (h << 15);
    h = h ^ (h >> 12);
    h = h + (h << 2);
    h = h ^ (h >> 4);
    h = h * 2057;
    h = h ^ (h >> 16);
#endif
    return h;
}


int int_equal(const void* a,const void* b){
    return *((int*)a) == *((int*)b);
}

unsigned int int_hash(const void* a){
    int h = *((int*)a);
    h = ((h >> 16) ^ h) * 0x45d9f3b;
    h = ((h >> 16) ^ h) * 0x45d9f3b;
    h = ((h >> 16) ^ h);
    return h;
}


int str_equal(const void* a,const void* b){
    return !strcmp((char*)a,(char*)b);
}

unsigned int str_hash_update(unsigned int h,const char* s,int len){
    const signed char* p = (const signed char*)s;
    while(*p && len){
        h = (h<<5)-h+*p;
        ++p;
        --len;
    }
    return h;
}

unsigned int str_hash(const void* a){
    /* 31 bit hash function */
    const signed char* p = a;
    unsigned int h = *p;
#if 1
    return str_hash_update(h,((const char*)a)+1,-1);
#else
    if(h)
        for(p+=1;*p;++p)
            h = (h<<5)-h+*p;
    return h;
#endif
}


int strcase_equal(const void* a,const void* b){
    return !strcasecmp((char*)a,(char*)b);
}

unsigned int strcase_hash_update(unsigned int h,const char* s,int len){
    const signed char* p = (const signed char*)s;
    while(*p && len){
        h = (h<<5)-h+tolower(*p);
        ++p;
        --len;
    }
    return h;
}

unsigned int strcase_hash(const void* a){
    /* 31 bit hash function */
    const signed char* p = a;
    unsigned int h = *p;
    return strcase_hash_update(h,((const char*)a)+1,-1);
}

#endif /* !HASH_USE_GLIB */
